/*
    Copyright (C) 2000-2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: canvas-waveview.c,v 1.24 2003/02/21 15:32:28 pbd Exp $
*/

#include <stdio.h>
#include <math.h>
#include <gtk-canvas.h>
#include <string.h>
#include <limits.h>

#include <ardour/dB.h>

#include "canvas-waveview.h"
#include "rgb_macros.h"

enum {
	ARG_0,
	ARG_DATA_SRC,
	ARG_CHANNEL,
	ARG_LENGTH_FUNCTION,
	ARG_PEAK_FUNCTION,
	ARG_SAMPLES_PER_PIXEL,
	ARG_AMPLITUDE_ABOVE_AXIS,
	ARG_X,
	ARG_Y,
	ARG_HEIGHT,
	ARG_WAVE_COLOR,
	ARG_RECTIFIED,
};

static void gtk_canvas_waveview_class_init (GtkCanvasWaveViewClass *class);
static void gtk_canvas_waveview_init       (GtkCanvasWaveView      *waveview);
static void gtk_canvas_waveview_destroy    (GtkObject              *object);
static void gtk_canvas_waveview_set_arg    (GtkObject              *object,
					      GtkArg                 *arg,
					      guint                   arg_id);
static void gtk_canvas_waveview_get_arg    (GtkObject              *object,
					      GtkArg                 *arg,
					      guint                   arg_id);

static void   gtk_canvas_waveview_update      (GtkCanvasItem *item, double *affine, ArtSVP *clip_path, int flags);
static void   gtk_canvas_waveview_bounds      (GtkCanvasItem *item, double *x1, double *y1, double *x2, double *y2);
static double gtk_canvas_waveview_point (GtkCanvasItem *item, double x, double y, int cx, int cy, GtkCanvasItem **actual_item);

static void gtk_canvas_waveview_render (GtkCanvasItem *item, GtkCanvasBuf *buf);
static void gtk_canvas_waveview_draw (GtkCanvasItem *item, GdkDrawable *drawable, int x, int y, int w, int h);

static void gtk_canvas_waveview_set_data_src      (GtkCanvasWaveView *, void *);
static void gtk_canvas_waveview_set_channel      (GtkCanvasWaveView *, guint32);

static gint32 gtk_canvas_waveview_ensure_cache (GtkCanvasWaveView *waveview, gulong start_sample, gulong end_sample);

static GtkCanvasItemClass *parent_class;


static void
check_cache (GtkCanvasWaveView *wv, const char *when)
{
	int i;

	for (i = 0; i < wv->cache_size; ++i) {
		if (wv->cache[i].min > 1.0 || wv->cache[i].min < -1.0 ||
		    wv->cache[i].max > 1.0 || wv->cache[i].max < -1.0) {
			fprintf (stderr, "%s: bad cache entry at %d (%f, %f)\n", when, i,
				 wv->cache[i].min, wv->cache[i].max);
		}
	}
}

GtkType
gtk_canvas_waveview_get_type (void)
{
	static GtkType waveview_type = 0;

	if (!waveview_type) {
		GtkTypeInfo waveview_info = {
			"GtkCanvasWaveView",
			sizeof (GtkCanvasWaveView),
			sizeof (GtkCanvasWaveViewClass),
			(GtkClassInitFunc) gtk_canvas_waveview_class_init,
			(GtkObjectInitFunc) gtk_canvas_waveview_init,
			NULL, /* reserved_1 */
			NULL, /* reserved_2 */
			(GtkClassInitFunc) NULL
		};

		waveview_type = gtk_type_unique (gtk_canvas_item_get_type (), &waveview_info);
	}

	return waveview_type;
}

static void
gtk_canvas_waveview_class_init (GtkCanvasWaveViewClass *class)
{
	GtkObjectClass *object_class;
	GtkCanvasItemClass *item_class;

	object_class = (GtkObjectClass *) class;
	item_class = (GtkCanvasItemClass *) class;

	parent_class = gtk_type_class (gtk_canvas_item_get_type ());

	gtk_object_add_arg_type ("GtkCanvasWaveView::data_src", GTK_TYPE_POINTER, GTK_ARG_READWRITE, ARG_DATA_SRC);
	gtk_object_add_arg_type ("GtkCanvasWaveView::channel", GTK_TYPE_POINTER, GTK_ARG_READWRITE, ARG_CHANNEL);
	gtk_object_add_arg_type ("GtkCanvasWaveView::length_function", GTK_TYPE_POINTER, GTK_ARG_READWRITE, ARG_LENGTH_FUNCTION);
	gtk_object_add_arg_type ("GtkCanvasWaveView::peak_function", GTK_TYPE_POINTER, GTK_ARG_READWRITE, ARG_PEAK_FUNCTION);
	gtk_object_add_arg_type ("GtkCanvasWaveView::samples_per_unit", GTK_TYPE_DOUBLE, GTK_ARG_READWRITE, ARG_SAMPLES_PER_PIXEL);
	gtk_object_add_arg_type ("GtkCanvasWaveView::amplitude_above_axis", GTK_TYPE_DOUBLE, GTK_ARG_READWRITE, ARG_AMPLITUDE_ABOVE_AXIS);
	gtk_object_add_arg_type ("GtkCanvasWaveView::x", GTK_TYPE_DOUBLE, GTK_ARG_READWRITE, ARG_X);
	gtk_object_add_arg_type ("GtkCanvasWaveView::y", GTK_TYPE_DOUBLE, GTK_ARG_READWRITE, ARG_Y);
	gtk_object_add_arg_type ("GtkCanvasWaveView::height", GTK_TYPE_DOUBLE, GTK_ARG_READWRITE, ARG_HEIGHT);
	gtk_object_add_arg_type ("GtkCanvasWaveView::wave_color", GTK_TYPE_INT, GTK_ARG_READWRITE, ARG_WAVE_COLOR);
	gtk_object_add_arg_type ("GtkCanvasWaveView::rectified", GTK_TYPE_BOOL, GTK_ARG_READWRITE, ARG_RECTIFIED);

	object_class->destroy = gtk_canvas_waveview_destroy;
	object_class->set_arg = gtk_canvas_waveview_set_arg;
	object_class->get_arg = gtk_canvas_waveview_get_arg;

	item_class->update = gtk_canvas_waveview_update;
	item_class->bounds = gtk_canvas_waveview_bounds;
	item_class->point = gtk_canvas_waveview_point;
	item_class->render = gtk_canvas_waveview_render;
	item_class->draw = gtk_canvas_waveview_draw;
}

static void
gtk_canvas_waveview_init (GtkCanvasWaveView *waveview)
{
	waveview->x = 0.0;
	waveview->y = 0.0;
	waveview->cache_size = 2048;
	waveview->cache = g_malloc (sizeof (GtkCanvasWaveViewCacheEntry) * 2048);
	waveview->cache_start = 0;
	waveview->cache_end = 0;

	waveview->data_src = NULL;
	waveview->channel = 0;
	waveview->peak_function = NULL;
	waveview->length_function = NULL;
	waveview->rectified = FALSE;

	waveview->samples_per_unit = 1.0;
	waveview->amplitude_above_axis = 1.0;
	waveview->height = 100.0;
	waveview->screen_width = gdk_screen_width ();
	
 	waveview->wave_color = RGB_TO_UINT(44,35,126);
}

static void
gtk_canvas_waveview_destroy (GtkObject *object)
{
	GtkCanvasWaveView *waveview;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GTK_CANVAS_IS_CANVAS_WAVEVIEW (object));

	waveview = GTK_CANVAS_WAVEVIEW (object);

	if (waveview->cache) {
		free (waveview->cache);
	}

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gtk_canvas_waveview_reload_cache (GtkCanvasWaveView *waveview)
{
	gulong nframes;
	gulong npeaks;

	nframes = waveview->cache_end - waveview->cache_start + 1;
	nframes = MIN (waveview->length_function (waveview->data_src) - waveview->cache_start, nframes);
	npeaks = (gulong) floor (nframes / waveview->samples_per_unit);
	npeaks = MAX (1.0, npeaks);
	nframes = npeaks * waveview->samples_per_unit;

	waveview->peak_function (waveview->data_src, npeaks, waveview->cache_start, nframes, waveview->cache, waveview->channel);

	if (npeaks < waveview->cache_size) {
		memset (&waveview->cache[npeaks], 0, sizeof (GtkCanvasWaveViewCacheEntry) * (waveview->cache_size - npeaks));
	}
}

#define DEBUG_CACHE 0

static gint32
gtk_canvas_waveview_ensure_cache (GtkCanvasWaveView *waveview, gulong start_sample, gulong end_sample)
{
	gulong required_cache_entries;
	gulong rf1, rf2, required_frames;
	gulong cache_start, cache_end;
	gulong half_width;

#if DEBUG_CACHE
		printf ("=> cache range: %lu - %lu request: %lu - %lu\n", 
			waveview->cache_start, waveview->cache_end,
			start_sample, end_sample);
#endif
		
	if (waveview->cache_start > start_sample || waveview->cache_end < end_sample) {
		
		gulong npeaks;
		gulong offset;

		/* make sure the cache is at least twice as wide as the screen width, and put the start sample
		   in the middle, ensuring that we cover the end_sample.
		*/
		
		rf1 = end_sample - start_sample + 1;
		rf2 = (gulong) (waveview->screen_width * waveview->samples_per_unit * 2.0);
		
		required_frames = MAX(rf1,rf2);
		
		/* Note the assumption that we have a 1:1 units:pixel ratio for the canvas. Its everywhere ... */
		
		half_width = (gulong) floor ((waveview->screen_width * waveview->samples_per_unit)/2.0);
		
		if (start_sample < half_width) {
			cache_start = 0;
		} else {
			cache_start = start_sample - half_width;
		}
		
		cache_end = cache_start + required_frames - 1;
		
		required_cache_entries = (gulong) floor (required_frames / waveview->samples_per_unit);
		
		if (required_cache_entries > waveview->cache_size) {
			waveview->cache = g_realloc (waveview->cache, sizeof (GtkCanvasWaveViewCacheEntry) * required_cache_entries);
			waveview->cache_size = required_cache_entries;
			waveview->cache_start = 0;
			waveview->cache_end = 0;
		}
		
		/* data is not entirely in the cache, so go fetch it, making sure to fill the cache */
		
		if (cache_start < waveview->cache_end) {

#ifdef USE_MEMMOVE_TO_SPEED_UP_CACHE_REFILL
			
			/* some of the required cache entries are in the cache, but in the wrong
			   locations. use memmove to fix this.
			*/
			
			if (waveview->cache_start < cache_start) {
			
				/* case one: shared area lives in RHS of existing cache. move it back to zero */
			
				present = waveview->cache_end - cache_start;
				memmove (&waveview->cache[0],
					 &waveview->cache[cache_start - waveview->cache_start],
					 present * sizeof (GtkCanvasWaveViewCacheEntry));

			} else {

				present = cache_end - waveview->cache_start;
				memmove (&waveview->cache[0],
					 &waveview->cache[waveview->cache_start - cache_start],
					 present * sizeof (GtkCanvasWaveViewCacheEntry));
			}

			printf ("copied %lu frames from existing cache\n", present);

			required_frames -= present;
			offset += present;
#else
			offset = 0;
#endif

		} else {
			offset = 0;
		}

#if DEBUG_CACHE
		printf ("desire requesting %f to cover %lu-%lu (request was %lu-%lu)\n",
			required_frames/waveview->samples_per_unit, cache_start, cache_end,
			start_sample, end_sample);
#endif

		required_frames = MIN (waveview->length_function (waveview->data_src) - cache_start, required_frames);
		npeaks = (gulong) floor (required_frames / waveview->samples_per_unit);
		npeaks = MAX (1, npeaks);
		required_frames = npeaks * waveview->samples_per_unit;

#if DEBUG_CACHE
		printf ("cache holds %lu entries, requesting %lu to cover %lu-%lu (request was %lu-%lu)\n",
			waveview->cache_size, npeaks, cache_start, cache_end,
			start_sample, end_sample);
#endif

		waveview->peak_function (waveview->data_src, npeaks, cache_start, required_frames, waveview->cache + offset, waveview->channel);

		/* take into account any copied peaks */

		npeaks += offset;

		if (npeaks < waveview->cache_size) {
			memset (&waveview->cache[npeaks], 0, sizeof (GtkCanvasWaveViewCacheEntry) * (waveview->cache_size - npeaks));
		}

		waveview->cache_start = cache_start;
		waveview->cache_end = cache_end;

//		check_cache (waveview, "after peak read");
		
//		printf ("cache range: %lu - %lu\n", waveview->cache_start, waveview->cache_end);
	} else {

#if DEBUG_CACHE
		printf ("cache hit for %lu-%lu (cache holds: %lu-%lu\n",
			start_sample, end_sample, waveview->cache_start, waveview->cache_end);
#endif
	}

	return (gint32) floor (((double) (start_sample - waveview->cache_start)) / waveview->samples_per_unit);
}

void
gtk_canvas_waveview_set_data_src (GtkCanvasWaveView *waveview, void *data_src)
{
	if (waveview->data_src == data_src) {
		gtk_canvas_waveview_reload_cache (waveview);
		return;
	}

	waveview->cache_start  = 0;
	waveview->cache_end = 0;
	
	waveview->data_src = data_src;
}

void
gtk_canvas_waveview_set_channel (GtkCanvasWaveView *waveview, guint32 chan)
{
	if (waveview->channel == chan) {
		return;
	}
	
	waveview->channel = chan;
}

static void 
gtk_canvas_waveview_reset_bounds (GtkCanvasItem *item)

{
	double x1, x2, y1, y2;
	ArtPoint i1, i2;
	ArtPoint w1, w2;
	int Ix1, Ix2, Iy1, Iy2;
	double i2w[6];

	gtk_canvas_waveview_bounds (item, &x1, &y1, &x2, &y2);

	i1.x = x1;
	i1.y = y1;
	i2.x = x2;
	i2.y = y2;

	gtk_canvas_item_i2w_affine (item, i2w);
	art_affine_point (&w1, &i1, i2w);
	art_affine_point (&w2, &i2, i2w);

	Ix1 = (int) rint(w1.x);
	Ix2 = (int) rint(w2.x);
	Iy1 = (int) rint(w1.y);
	Iy2 = (int) rint(w2.y);

	gtk_canvas_update_bbox (item, Ix1, Iy1, Ix2, Iy2);
}

/* 
 * CANVAS CALLBACKS 
 */

static void
gtk_canvas_waveview_set_arg (GtkObject *object, GtkArg *arg, guint arg_id)
{
	GtkCanvasItem *item;
	GtkCanvasWaveView *waveview;
	int redraw;
	int calc_bounds;

	item = GTK_CANVAS_ITEM (object);
	waveview = GTK_CANVAS_WAVEVIEW (object);

	redraw = FALSE;
	calc_bounds = FALSE;

	switch (arg_id) {
	case ARG_DATA_SRC:
		gtk_canvas_waveview_set_data_src (waveview, GTK_VALUE_POINTER(*arg));
		redraw = TRUE;
		break;

	case ARG_CHANNEL:
		gtk_canvas_waveview_set_channel (waveview, GTK_VALUE_UINT(*arg));
		redraw = TRUE;
		break;

	case ARG_LENGTH_FUNCTION:
		waveview->length_function = GTK_VALUE_POINTER(*arg);
		redraw = TRUE;
		break;

	case ARG_PEAK_FUNCTION:
		waveview->peak_function = GTK_VALUE_POINTER(*arg);
		redraw = TRUE;
		break;

	case ARG_SAMPLES_PER_PIXEL:
		if ((waveview->samples_per_unit = GTK_VALUE_DOUBLE(*arg)) < 1.0) {
			waveview->samples_per_unit = 1.0;
		}
		/* invalidate cache */
		waveview->cache_start = 0;
		waveview->cache_end = 0;
		redraw = TRUE;
		calc_bounds = TRUE;
		break;

	case ARG_AMPLITUDE_ABOVE_AXIS:
		waveview->amplitude_above_axis = GTK_VALUE_DOUBLE(*arg);
		redraw = TRUE;
		break;

	case ARG_X:
	        if (waveview->x != GTK_VALUE_DOUBLE (*arg)) {
		        waveview->x = GTK_VALUE_DOUBLE (*arg);
			calc_bounds = TRUE;
		}
		break;

	case ARG_Y:
	        if (waveview->y != GTK_VALUE_DOUBLE (*arg)) {
		        waveview->y = GTK_VALUE_DOUBLE (*arg);
			calc_bounds = TRUE;
		}
		break;

	case ARG_HEIGHT:
	        if (waveview->height != fabs (GTK_VALUE_DOUBLE (*arg))) {
		        waveview->height = fabs (GTK_VALUE_DOUBLE (*arg));
			redraw = TRUE;
		}
		break;

	case ARG_WAVE_COLOR:
		if (waveview->wave_color != GTK_VALUE_INT(*arg)) {
			waveview->wave_color = GTK_VALUE_INT(*arg);
			redraw = TRUE;
		}
		break;

	case ARG_RECTIFIED:
		if (waveview->rectified != GTK_VALUE_BOOL(*arg)) {
			waveview->rectified = GTK_VALUE_BOOL(*arg);
			redraw = TRUE;
		}
		break;


	default:
		break;
	}

	if (calc_bounds) {
		gtk_canvas_waveview_reset_bounds (item);
	}

	if (redraw) {
		gtk_canvas_item_request_update (item);
	}

}

static void
gtk_canvas_waveview_get_arg (GtkObject *object, GtkArg *arg, guint arg_id)
{
	GtkCanvasWaveView *waveview;

	waveview = GTK_CANVAS_WAVEVIEW (object);

	switch (arg_id) {
	case ARG_DATA_SRC:
		GTK_VALUE_POINTER(*arg) = waveview->data_src;
		break;

	case ARG_CHANNEL:
		GTK_VALUE_UINT(*arg) = waveview->channel;
		break;

	case ARG_LENGTH_FUNCTION:
		GTK_VALUE_POINTER(*arg) = waveview->length_function;
		break;

	case ARG_PEAK_FUNCTION:
		GTK_VALUE_POINTER(*arg) = waveview->peak_function;
		break;

	case ARG_SAMPLES_PER_PIXEL:
		GTK_VALUE_DOUBLE(*arg) = waveview->samples_per_unit;
		break;

	case ARG_AMPLITUDE_ABOVE_AXIS:
		GTK_VALUE_DOUBLE(*arg) = waveview->amplitude_above_axis;
		break;

	case ARG_X:
		GTK_VALUE_DOUBLE (*arg) = waveview->x;
		break;

	case ARG_Y:
		GTK_VALUE_DOUBLE (*arg) = waveview->y;
		break;

	case ARG_HEIGHT:
		GTK_VALUE_DOUBLE (*arg) = waveview->height;
		break;

	case ARG_WAVE_COLOR:
		GTK_VALUE_INT (*arg) = waveview->wave_color;
		break;

	case ARG_RECTIFIED:
		GTK_VALUE_BOOL (*arg) = waveview->rectified;

	default:
		arg->type = GTK_TYPE_INVALID;
		break;
	}
}

static void
gtk_canvas_waveview_update (GtkCanvasItem *item, double *affine, ArtSVP *clip_path, int flags)
{
	GtkCanvasWaveView *waveview;
	double x, y;

	waveview = GTK_CANVAS_WAVEVIEW (item);

//	check_cache (waveview, "start of update");

	if (parent_class->update)
		(* parent_class->update) (item, affine, clip_path, flags);

	gtk_canvas_waveview_reset_bounds (item);

	/* get the canvas coordinates of the view. Do NOT use affines
	   for this, because they do not round to the integer units used
	   by the canvas, resulting in subtle pixel-level errors later.
	*/

	x = 0;
	y = 0;

	gtk_canvas_item_i2w (item, &x, &y);
	gtk_canvas_w2c (GTK_CANVAS(item->canvas), x, y, &waveview->bbox_ulx, &waveview->bbox_uly);

	waveview->samples = waveview->length_function (waveview->data_src);
	
	x = waveview->samples / waveview->samples_per_unit;
	y = waveview->height;

	gtk_canvas_item_i2w (item, &x, &y);
	gtk_canvas_w2c (GTK_CANVAS(item->canvas), x, y, &waveview->bbox_lrx, &waveview->bbox_lry);

	/* cache the half-height and the end point in canvas units */

	waveview->half_height = waveview->height / 2.0;

	/* parse the color */

	UINT_TO_RGB (waveview->wave_color, &waveview->wave_r, &waveview->wave_g, &waveview->wave_b);

//	check_cache (waveview, "end of update");
}

static void
gtk_canvas_waveview_render (GtkCanvasItem *item,
			    GtkCanvasBuf *buf)
{
	GtkCanvasWaveView *waveview;
	gulong s1, s2;
	int clip_length;
	int pymin, pymax;
	int cache_index;
	double half_height;
	int x, end, begin;

	waveview = GTK_CANVAS_WAVEVIEW (item);

//	check_cache (waveview, "start of render");

	if (parent_class->render) {
		(*parent_class->render) (item, buf);
	}

	if (buf->is_bg) {
		gtk_canvas_buf_ensure_buf (buf);
		buf->is_bg = FALSE;
	}

	begin = MAX(waveview->bbox_ulx,buf->rect.x0);

	if (waveview->bbox_lrx >= 0) {
		end = MIN(waveview->bbox_lrx,buf->rect.x1);
	} else {
		end = buf->rect.x1;
	}

	if (begin == end) {
		return;
	}

	s1 = floor ((begin - waveview->bbox_ulx) * waveview->samples_per_unit);

	if (end == waveview->bbox_lrx) {
		/* This avoids minor rounding errors when we have the
		   entire region visible.
		*/
		s2 = waveview->samples;
	} else {
		s2 = s1 + floor (((end - begin) * waveview->samples_per_unit));
	}

#if 0
	printf ("r (%d..%d)(%d..%d) bbox (%d..%d)(%d..%d)"
		" b/e %d..%d s= %lu..%lu\n",
		buf->rect.x0,
		buf->rect.x1,
		buf->rect.y0,
		buf->rect.y1,
		waveview->bbox_ulx,
		waveview->bbox_lrx,
		waveview->bbox_uly,
		waveview->bbox_lry,
		begin, end, s1, s2);
#endif

	/* now ensure that the cache is full and properly
	   positioned.
	*/

//	check_cache (waveview, "pre-ensure");

	cache_index = gtk_canvas_waveview_ensure_cache (waveview, s1, s2);

//	check_cache (waveview, "post-ensure");

	/* 
	   Now draw each line, clipping it appropriately. The clipping
	   is done by the macros PAINT_FOO().
	*/

	half_height = waveview->half_height;

/* this makes it slightly easier to comprehend whats going on */

#define origin half_height

	for (x = begin; x < end; x++) {

		double max, min;
		int clip_max, clip_min;
		
		clip_max = 0;
		clip_min = 0;

		max = waveview->cache[cache_index].max;
		min = waveview->cache[cache_index].min;

		if (max > 1.0) {
			max = 1.0;
			clip_max = 1;
		}

		if (min < -1.0) {
			min = -1.0;
			clip_min = 1;
		}

		/* don't rectify at single-sample zoom */

		if (waveview->rectified && waveview->samples_per_unit > 1) {
			double tmax, tmin;

			if (fabs (min) > fabs (max)) {
				max = fabs (min);
			} 

			max = max * waveview->height;

			pymax = (int) rint ((item->y1 + waveview->height - max) * item->canvas->pixels_per_unit);
			pymin = (int) rint ((item->y1 + waveview->height) * item->canvas->pixels_per_unit);

		} else {
			
			max = max * half_height;
			min = min * half_height;
			
			pymax = (int) rint ((item->y1 + origin - max) * item->canvas->pixels_per_unit);
			pymin = (int) rint ((item->y1 + origin - min) * item->canvas->pixels_per_unit);
		}

//		fprintf (stderr, "x = %d, cache_index = %d, max = %f min = %f ymax = %d ymin = %d\n", x, cache_index, max, min, pymax, pymin);

		/* OK, now fill the RGB buffer at x=i with a line between pymin and pymax,
		   or, if samples_per_unit == 1, then a dot at each location.
		*/

		if (pymax == pymin) {
			PAINT_DOT(buf, waveview->wave_r, waveview->wave_g, waveview->wave_b, x, pymin);
		} else {
			PAINT_VERT(buf, waveview->wave_r, waveview->wave_g, waveview->wave_b, x, pymax, pymin);
		}
		
		/* show clipped waveforms with small red lines */

		if (clip_max || clip_min) {
			clip_length = MIN(5,(waveview->height/4));
		}

		if (clip_max) {
			PAINT_VERT(buf, 255, 0, 0, x, pymax, pymax+clip_length);
		}

		if (clip_min) {
			PAINT_VERT(buf, 255, 0, 0, x, pymin-clip_length, pymin);
		}

		/* presto, we're done */
		
		cache_index++;
	}

#undef origin

}

static void
gtk_canvas_waveview_draw (GtkCanvasItem *item,
			  GdkDrawable *drawable,
			  int x, int y,
			  int width, int height)
{
	GtkCanvasWaveView *waveview;

	waveview = GTK_CANVAS_WAVEVIEW (item);

	if (parent_class->draw) {
		(* parent_class->draw) (item, drawable, x, y, width, height);
	}

	fprintf (stderr, "please don't use the CanvasWaveView item in a non-aa Canvas\n");
	abort ();
}

static void
gtk_canvas_waveview_bounds (GtkCanvasItem *item, double *x1, double *y1, double *x2, double *y2)
{
	GtkCanvasWaveView *waveview = GTK_CANVAS_WAVEVIEW (item);

	*x1 = waveview->x;
	*y1 = waveview->y;

	*x2 = ceil (*x1 + (waveview->length_function (waveview->data_src) / waveview->samples_per_unit));
	*y2 = *y1 + waveview->height;

#if 0
	x = 0; y = 0;
	gtk_canvas_item_i2w (item, &x, &y);
	gtk_canvas_w2c_d (GTK_CANVAS(item->canvas), x, y, &a, &b);
	x = *x2;
	y = *y2;
	gtk_canvas_item_i2w (item, &x, &y);
	gtk_canvas_w2c_d (GTK_CANVAS(item->canvas), x, y, &c, &d);
	printf ("item bounds now (%g,%g),(%g,%g)\n", a, b, c, d);
#endif		

}

static double
gtk_canvas_waveview_point (GtkCanvasItem *item, double x, double y, int cx, int cy, GtkCanvasItem **actual_item)
{
	GtkCanvasWaveView *waveview;
	double x1, y1, x2, y2;
	double dx, dy;

	/* XXX for now, point is never inside the wave */

	return DBL_MAX;

#if 0
	waveview = GTK_CANVAS_WAVEVIEW (item);

	*actual_item = item;

	/* Find the bounds for the rectangle plus its outline width */

	gtk_canvas_waveview_bounds (item, &x1, &y1, &x2, &y2);

	/* Is point inside rectangle */
	
	if ((x >= x1) && (y >= y1) && (x <= x2) && (y <= y2)) {
		return 0.0;
	}

	/* Point is outside rectangle */

	if (x < x1)
		dx = x1 - x;
	else if (x > x2)
		dx = x - x2;
	else
		dx = 0.0;

	if (y < y1)
		dy = y1 - y;
	else if (y > y2)
		dy = y - y2;
	else
		dy = 0.0;

	return sqrt (dx * dx + dy * dy);
#endif
}

