/*
    Copyright (C) 2000-2003 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: editor.h,v 1.176 2003/06/25 20:33:40 essej Exp $
*/

#ifndef __ardour_editor_h__
#define __ardour_editor_h__

#include <list>
#include <map>
#include <string>

#include <cmath>

#include <sndfile.h>

#include <gtk--.h>
#include <gtk--/layout.h>
#include <gtk-canvas.h>
#include <gtk-canvas/gtk-canvas-util.h>
#include <gtkmmext/selector.h>
#include <gtkmmext/click_box.h>

#include <ardour/stateful.h>
#include <ardour/session.h>
#include <ardour/tempo.h>
#include <ardour/overlap.h>
#include <ardour/location.h>

#include "keyboard_target.h"
#include "audio_clock.h"
#include "rgb_macros.h"
#include "gtk-custom-ruler.h"
#include "ardour_dialog.h"

namespace Gtkmmext {
	class TearOff;
}

namespace LinuxAudioSystems {
	class AudioEngine;
}

namespace ARDOUR {
	class DiskStream;
	class RouteGroup;
	class Source;
	class Playlist;
	class Region;
	class Location;
	class TempoSection;
	class NamedSelection;
}

namespace LADSPA {
	class Plugin;
}

class TimeAxisView;
class AudioTimeAxisView;
class RegionView;
class PluginManager;
class Marker;
class GroupedButtons;
class AutomationLine;
class UIExportSpecification;
class ExportDialog;
class Selection;
class TimeSelection;
class RegionSelection;
class TrackSelection;
class AutomationSelection;
class MixerStrip;

class Editor : public Gtk::Window, 
	       public Stateful, 
	       public KeyboardTarget
{
  public:
	enum SelectionMode {
		SelectData,
		SelectGain,
		SelectPan,
		SelectSolo,
		SelectMute,
		SelectRedirect,
	};

	enum SnapType {
		SnapToBar,
		SnapToBeat,
		SnapToFrame,
		SnapToMark,
		SnapToAThirdBeat,
                SnapToAQuarterBeat,
	};

	enum XFadeType {
		Pre,
		Post,
		At
	};

	enum RegionListSortType {
		ByName,
		ByLength,
		ByPosition,
		ByTimestamp,
		ByStartInFile,
		ByEndInFile,
		BySourceFileName,
		BySourceFileLength,
		BySourceFileCreationDate,
		BySourceFileFS
	};

	Editor (ARDOUR::AudioEngine&);
	~Editor ();

	void             connect_to_session (ARDOUR::Session *);
	ARDOUR::Session* current_session() const { return session; }

	void set_snap_to (SnapType);

	void undo (unsigned int n = 1);
	void redo (unsigned int n = 1);

	XMLNode& get_state (void);
	int set_state (const XMLNode& );

	/* mouse modes */

	enum MouseMode {
		MouseRange,
		MouseObject,
		MouseGain,
		MousePan,
		MouseAutomation,
		MouseZoom,
		MouseScrub,
		MouseTimeFX,
	};

	void set_mouse_mode (MouseMode);
	void step_mouse_mode (bool next);
	MouseMode current_mouse_mode () { return mouse_mode; }

	/* XXX there must be a better way than making these public */

	static gint _canvas_region_view_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_region_view_name_highlight_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_region_view_name_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_stream_view_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_marker_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_zoom_rect_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_selection_rect_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_selection_start_trim_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_selection_end_trim_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_control_point_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_line_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_tempo_marker_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_meter_marker_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_tempo_bar_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_meter_bar_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	static gint _canvas_marker_bar_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);

	SigC::Signal0<void> frames_per_unit_changed;

	void start_export_cd_markers ();

	/* public access to auditioning */

	void consider_auditioning (ARDOUR::Region*);

	/* option editor-access */

	void set_show_waveforms (bool yn);
	bool show_waveforms() const { return _show_waveforms; }

	/* things that need to be public to be used in the main menubar */

	void new_region_from_selection ();
	void separate_region_from_selection ();
	void toggle_playback (bool with_abort);

	/* crossfade */
	
	void maybe_crossfade (ARDOUR::Overlap *);
	void crossfade (ARDOUR::Overlap *);

	/* undo related */

	void set_edit_menu (Gtk::Menu&);

	jack_nframes_t unit_to_frame (double unit) {
		return (jack_nframes_t) rint (unit * frames_per_unit);
	}
	
	double frame_to_unit (jack_nframes_t frame) {
		return rint ((double) frame / (double) frames_per_unit);
	}

	double frame_to_unit (double frame) {
		return rint (frame / frames_per_unit);
	}

	/* NOTE: these functions assume that the "pixel" coordinate is
	   the result of using the world->canvas affine transform on a
	   world coordinate. These coordinates already take into
	   account any scrolling carried out by adjusting the
	   xscroll_adjustment.  
	*/

	jack_nframes_t pixel_to_frame (double pixel) {
		
		/* pixel can be less than zero when motion events
		   are processed. since we've already run the world->canvas
		   affine, that means that the location *really* is "off
		   to the right" and thus really is "before the start".
		*/

		if (pixel >= 0) {
			return (jack_nframes_t) rint (pixel * frames_per_unit * GTK_CANVAS(track_gtk_canvas)->pixels_per_unit);
		} else {
			return 0;
		}
	}

	gulong frame_to_pixel (jack_nframes_t frame) {
		return (gulong) rint ((frame / (frames_per_unit *  GTK_CANVAS(track_gtk_canvas)->pixels_per_unit)));
	}

	/* selection */

	Selection& get_selection() const { return *selection; }
	jack_nframes_t selection_start () const;
	jack_nframes_t selection_end () const;
	void play_selection ();

	/* tempo */

	void set_show_measures (bool yn);
	bool show_measures () const { return _show_measures; }

	/* export */

	/* these initiate export ... */
	
	void export_session();
	void export_selection();

	/* this is what actually does it */
	
	void export_audiofile (ARDOUR::AudioExportSpecification&);

	/* */

	void add_toplevel_controls (Gtk::Container&);

	/* ZOOM */

	enum ZoomFocus {
		ZoomFocusLeft,
		ZoomFocusRight,
		ZoomFocusCenter,
		ZoomFocusPlayhead,
		ZoomFocusEdit
	};

	void      set_zoom_focus (ZoomFocus);
	ZoomFocus get_zoom_focus () const { return zoom_focus; }
	gdouble   get_current_zoom () { return frames_per_unit; }

	SigC::Signal0<void> ZoomFocusChanged;

	/* stuff that AudioTimeAxisView and related classes use */

	PluginManager* plugin_manager;
	void route_name_changed (TimeAxisView *);
	gdouble       frames_per_unit;
	void clear_playlist (ARDOUR::Playlist&);

	void set_selected_mixer_strip (TimeAxisView&);
	void unselect_strip_in_display (TimeAxisView& tv);

	/* nudge is initiated by transport controls owned by ARDOUR_UI */

	void nudge_forward (bool next);
	void nudge_backward (bool next);

	/* floating windows/transient */

	void ensure_float (Gtk::Window&);

	void show_window ();
	
  protected:
	void map_transport_state ();
	void map_position_change ();

	void realize_impl();
	void map__impl();

  private:
	
	ARDOUR::Session   *session;
	ARDOUR::AudioEngine& engine;

	void          set_frames_per_unit (double);
	void          frames_per_unit_modified ();

	jack_nframes_t current_page_frames() const {
		return (jack_nframes_t) floor (canvas_width * frames_per_unit);
	}

	MouseMode mouse_mode;
	MouseMode last_mouse_mode;
	void      mouse_insert (GdkEventButton *);

	void pane_allocation_handler (GtkAllocation*, Gtk::Paned*);

	Gtk::HPaned   canvas_region_list_pane;
	Gtk::HPaned   track_list_canvas_pane;

	Gtk::EventBox meter_base;
	Gtk::HBox     meter_box;
	Gtk::EventBox marker_base;
	Gtk::HBox     marker_box;
	Gtk::VBox     scrollers_rulers_markers_box;

	void location_changed (ARDOUR::Location *);
	void location_flags_changed (ARDOUR::Location *, void *);
	void refresh_location_display ();
	void refresh_location_display_internal (ARDOUR::Locations::LocationList&);
	void add_new_location (ARDOUR::Location *);
	void location_gone (ARDOUR::Location *);
	void remove_marker (GtkCanvasItem*, GdkEvent*);
	gint really_remove_marker (ARDOUR::Location* loc);

	unsigned int location_marker_color;
	unsigned int location_range_color;
	unsigned int location_cd_marker_color;

	struct LocationMarkers {
	    Marker* start;
	    Marker* end;

	    ~LocationMarkers ();

	    void hide();
	    void show ();
	    void set_name (const string&);
	    void set_position (jack_nframes_t start, jack_nframes_t end = 0);
	    void set_color_rgba (unsigned int);
	};

	LocationMarkers  *find_location_markers (ARDOUR::Location *);
	ARDOUR::Location* find_location_from_marker (Marker *, bool& is_start);

	typedef map<ARDOUR::Location*,LocationMarkers *> LocationMarkerMap;
	LocationMarkerMap location_markers;

	void hide_marker (GtkCanvasItem*, GdkEvent*);
	void clear_markers ();
	void mouse_add_new_marker (GdkEvent*);

	TimeAxisView*  clicked_trackview;
	AudioTimeAxisView*     clicked_audio_trackview;
	RegionView*         clicked_regionview;
	RegionView*         latest_regionview;
	unsigned long       clicked_selection;

	void catch_vanishing_regionview (RegionView *);
	void set_selected_regionview_from_click (bool add = false);
	void collect_new_region_view (RegionView *);

	Gtk::Menu dstream_edit_menu;
	Gtk::Menu dstream_region_edit_menu;
	Gtk::Menu dstream_selection_edit_menu;
	Gtk::MenuItem* region_edit_menu_split_item;
	Gtk::MenuItem* region_edit_menu_split_multichannel_item;

	void popup_dstream_edit_menu (int, int, bool, bool);
	void build_dstream_edit_menu ();
	void build_dstream_region_edit_menu ();
	void build_dstream_selection_edit_menu ();
	void add_dstream_edit_items (Gtk::Menu_Helpers::MenuList&);
	void add_region_edit_items (Gtk::Menu_Helpers::MenuList&);
	void add_selection_edit_items (Gtk::Menu_Helpers::MenuList&);

	void fake_handle_new_route (ARDOUR::Route&);
	void handle_new_route (ARDOUR::Route&);
	void handle_new_route_p (ARDOUR::Route*);
	void remove_route (TimeAxisView *);
	bool route_removal;

	Gtk::HBox           global_hpacker;
	Gtk::VBox           global_vpacker;
	Gtk::VBox           vpacker;

	GdkCursor          *current_canvas_cursor;

	GtkWidget          *track_gtk_canvas;
	GtkWidget          *time_gtk_canvas;
	Gtk::Widget        *track_canvas;
	Gtk::Widget        *time_canvas;

	GtkCanvasItem      *no_session_message;

	GtkCanvasItem      *verbose_canvas_cursor;
	bool                verbose_cursor_visible;

	gint track_canvas_motion (GtkCanvasItem*, GdkEvent*);
	void set_verbose_canvas_cursor (string, double x, double y);
	void set_verbose_canvas_cursor_text (string);
	void show_verbose_canvas_cursor();
	void hide_verbose_canvas_cursor();

	bool verbose_cursor_on; // so far unused

	GtkCanvasItem      *region_copy_group;
	GtkCanvasItem      *region_copy_rect;
	GtkCanvasItem      *region_copy_name;
	double              region_copy_initial_affine[6];

	void flush_track_canvas ();
	void flush_time_canvas ();

	Gtk::EventBox      time_canvas_event_box;
	Gtk::EventBox      track_canvas_event_box;

	GtkCanvasItem      *minsec_group;
	GtkCanvasItem      *bbt_group;
	GtkCanvasItem      *smpte_group;
	GtkCanvasItem      *frame_group;
	GtkCanvasItem      *tempo_group;
	GtkCanvasItem      *meter_group;
	GtkCanvasItem      *marker_group;
	
	enum {
		ruler_metric_smpte = 0,
		ruler_metric_bbt = 1,
		ruler_metric_frames = 2,
		ruler_metric_minsec = 3,
	};

	static GtkCustomMetric ruler_metrics[4];

	void initialize_rulers ();
	void update_fixed_rulers ();
	void update_tempo_based_rulers (); 

	static gint _metric_get_smpte (GtkCustomRulerMark **, gulong, gulong, gint);
	static gint _metric_get_bbt (GtkCustomRulerMark **, gulong, gulong, gint);
	static gint _metric_get_frames (GtkCustomRulerMark **, gulong, gulong, gint);
	static gint _metric_get_minsec (GtkCustomRulerMark **, gulong, gulong, gint);
	
	gint metric_get_smpte (GtkCustomRulerMark **, gulong, gulong, gint);
	gint metric_get_bbt (GtkCustomRulerMark **, gulong, gulong, gint);
	gint metric_get_frames (GtkCustomRulerMark **, gulong, gulong, gint);
	gint metric_get_minsec (GtkCustomRulerMark **, gulong, gulong, gint);

	GtkWidget          *_smpte_ruler;
	GtkWidget          *_bbt_ruler;
	GtkWidget          *_frames_ruler;
	GtkWidget          *_minsec_ruler;
	Gtk::Widget        *smpte_ruler;
	Gtk::Widget        *bbt_ruler;
	Gtk::Widget        *frames_ruler;
	Gtk::Widget        *minsec_ruler;
	static Editor      *ruler_editor;

	static const double timebar_height = 15.0;
	guint32 visible_timebars;

	GtkCanvasItem      *tempo_bar;
	GtkCanvasItem      *meter_bar;
	GtkCanvasItem      *marker_bar;

	GtkCanvasItem      *tempo_line;
	GtkCanvasItem      *meter_line;
	GtkCanvasItem      *marker_line;

	GtkCanvasPoints    *tempo_line_points;
	GtkCanvasPoints    *meter_line_points;
	GtkCanvasPoints    *marker_line_points;

	Gtk::ToggleButton  minsec_button;
	Gtk::ToggleButton  bbt_button;
	Gtk::ToggleButton  smpte_button;
	Gtk::ToggleButton  frame_button;
	Gtk::ToggleButton  tempo_button;
	Gtk::ToggleButton  meter_button;
	Gtk::ToggleButton  mark_button;

	Gtk::VBox          time_button_vbox;
	Gtk::HBox          time_button_hbox;

	struct Cursor {
	    Editor&             editor;
	    GtkCanvasPoints    *points;
	    GtkCanvasItem      *canvas_item;
	    jack_nframes_t           current_frame;
	    GtkSignalFunc       callback;
	    bool                full_height;

	    Cursor (Editor&, const string& color, GtkSignalFunc callback, bool full_height = false);
	    ~Cursor ();

	    void set_position (jack_nframes_t);
	    void set_height (double units);
	};

	friend struct Cursor; /* it needs access to several private
				 fields. XXX fix me.
			      */

	Cursor* playhead_cursor;
	Cursor* edit_cursor;
	Cursor* select_left_cursor;
	Cursor* select_right_cursor;

	GtkCanvasItem* scrub_rect;
	void scrub_ready_proxy ();
	void scrub_ready ();
	SigC::Connection scrub_connection;

	Gtk::VBox           trackview_vpacker;

	Gtk::HBox           top_hbox;
	Gtk::HBox           bottom_hbox;
	
	Gtk::Table          edit_packer;
	Gtk::Frame          edit_frame;
	Gtk::VScrollbar     edit_vscrollbar;

	/* the horizontal scroller works in a rather different way
	   than a regular scrollbar, since its used for
	   zoom control/indication as well. But more importantly,
	   its different components (slider, left arrow, right arrow) 
	   have to be packed separately into the edit_packer.
	*/

	Gtk::DrawingArea    edit_hscroll_slider;
	Gtk::Arrow          edit_hscroll_left_arrow;
	Gtk::Arrow          edit_hscroll_right_arrow;
	Gtk::EventBox       edit_hscroll_left_arrow_event;
	Gtk::EventBox       edit_hscroll_right_arrow_event;
	gint                edit_hscroll_slider_width;
	gint                edit_hscroll_slider_height;
	static const gint   edit_hscroll_edge_width = 3;
	bool                edit_hscroll_dragging;
	double              edit_hscroll_drag_last;
	
	void hscroll_slider_allocate (GtkAllocation *);
	gint hscroll_slider_expose (GdkEventExpose*);
	gint hscroll_slider_button_press (GdkEventButton*);
	gint hscroll_slider_button_release (GdkEventButton*);
	gint hscroll_slider_motion (GdkEventMotion*);

	gint hscroll_trough_expose (GdkEventExpose*);
	gint hscroll_trough_button_press (GdkEventButton*);
	gint hscroll_trough_button_release (GdkEventButton*);

	void update_hscroller ();

	gint hscroll_left_arrow_button_press (GdkEventButton *);
	gint hscroll_left_arrow_button_release (GdkEventButton *);
	gint hscroll_right_arrow_button_press (GdkEventButton *);
	gint hscroll_right_arrow_button_release (GdkEventButton *);
	
	guint32             canvas_width;
	guint32             canvas_height;

	Gtk::ScrolledWindow  track_canvas_scroller;
	Gtk::ScrolledWindow  time_canvas_scroller;
	Gtk::ScrolledWindow  edit_controls_scroller;

	gint edit_controls_button_release (GdkEventButton*);
	Gtk::Menu *edit_controls_left_menu;
	Gtk::Menu *edit_controls_right_menu;

	void track_canvas_scroller_realized ();
	void time_canvas_scroller_realized ();

	Gtk::VBox           track_canvas_vbox;
	Gtk::VBox           time_canvas_vbox;
	Gtk::VBox           edit_controls_vbox;
	Gtk::HBox           edit_controls_hbox;

	jack_nframes_t leftmost_frame;

	void tie_vertical_scrolling ();
	void canvas_horizontally_scrolled ();
	void reposition_x_origin (jack_nframes_t sample);
	void reposition_and_zoom (jack_nframes_t sample, double fpu);
	bool no_zoom_repos_update;
	bool no_tempo_map_update;

	Gtk::CTree          region_list_display;
	Gtk::Menu          *region_list_menu;
	Gtk::ScrolledWindow region_list_scroller;
	Gtk::Frame          region_list_frame;

	gint region_list_display_button_press (GdkEventButton *);
	gint region_list_display_button_release (GdkEventButton *);
	gint region_list_display_enter_notify (GdkEventCrossing *);
	gint region_list_display_leave_notify (GdkEventCrossing *);
	void region_list_display_selected (gint row, gint col, GdkEvent* ev);
	void region_list_display_unselected (gint row, gint col, GdkEvent* ev);
	void region_list_column_click (gint);
	void region_list_clear ();
	void build_region_list_menu ();

	Gtk::CheckMenuItem* toggle_auto_regions_item;
	Gtk::CheckMenuItem* toggle_full_region_list_item;

	bool show_automatic_regions_in_region_list;
	RegionListSortType region_list_sort_type;

	void reset_region_list_sort_direction (bool);
	void reset_region_list_sort_type (RegionListSortType);

	void toggle_full_region_list ();
	void toggle_show_auto_regions ();

	static gint _region_list_sorter (GtkCList*, gconstpointer, gconstpointer);
	gint region_list_sorter (gconstpointer, gconstpointer);

	/* named selections */

	Gtk::VPaned         region_selection_vpane;
	Gtk::CList          named_selection_display;
	Gtk::ScrolledWindow named_selection_scroller;

	void name_selection();
	void named_selection_name_chosen ();
	void create_named_selection (string);
	void paste_named_selection (float times);

	void handle_new_named_selection ();
	void add_named_selection_to_named_selection_display (ARDOUR::NamedSelection&);
	void redisplay_named_selections ();

	gint named_selection_display_button_press (GdkEventButton *ev);
	void named_selection_display_selected (gint row, gint col, GdkEvent *ev);
	void named_selection_display_unselected (gint row, gint col, GdkEvent *ev);

	/* track views */

	pthread_mutex_t track_lock;
	typedef list<TimeAxisView *> TrackViewList;

	int track_spacing;
	TrackViewList  track_views;
	TrackViewList* figure_out_valid_views ();
	TrackViewList* get_valid_views (TimeAxisView*, ARDOUR::RouteGroup* grp = 0);
	TimeAxisView     *trackview_by_y_position (double ypos);

	static GdkCursor* cross_hair_cursor;
	static GdkCursor* trimmer_cursor;
	static GdkCursor* selector_cursor;
	static GdkCursor* grabber_cursor;
	static GdkCursor* zoom_cursor;
	static GdkCursor* time_fx_cursor;
	static GdkCursor* fader_cursor;
	static GdkCursor* null_cursor;
	static GdkCursor* wait_cursor;

	static void build_cursors ();

	SigC::Connection scroll_connection;
	jack_nframes_t last_update_frame;
	void center_screen (jack_nframes_t);
	void update_current_screen ();
	
	gint show_dstream_edit_menu (GdkEventButton *);
	void hide_dstream_edit_menu ();

	void session_going_away ();

	jack_nframes_t cut_buffer_start;
	jack_nframes_t cut_buffer_length;

	void session_transport_state_change ();
	void session_position_change (jack_nframes_t frame);

	enum ItemType {
		RegionItem,
		StreamItem,
		PlayheadCursorItem,
		EditCursorItem,
		MarkerItem,
		MarkerBarItem,
		SelectionItem,
		GainControlPointItem,
		GainLineItem,
		GainAutomationControlPointItem,
		GainAutomationLineItem,
		PanAutomationControlPointItem,
		PanAutomationLineItem,
		RedirectAutomationControlPointItem,
		RedirectAutomationLineItem,
		MeterMarkerItem,
		TempoMarkerItem,
		MeterBarItem,
		TempoBarItem,
		RegionViewNameHighlight,
		RegionViewName,
		StartSelectionTrimItem,
		EndSelectionTrimItem,
		NoItem
	};

	gint typed_event (GtkCanvasItem*, GdkEvent*, ItemType);
	gint button_press_handler (GtkCanvasItem*, GdkEvent*, ItemType);
	gint button_release_handler (GtkCanvasItem*, GdkEvent*, ItemType);
	gint motion_handler (GtkCanvasItem*, GdkEvent*, ItemType);
	gint enter_handler (GtkCanvasItem*, GdkEvent*, ItemType);
	gint leave_handler (GtkCanvasItem*, GdkEvent*, ItemType);
	
	/* KEYMAP HANDLING */

	void install_keybindings ();

	int ensure_cursor (jack_nframes_t* pos);

	void fake_handle_new_region (ARDOUR::Region *);
	void handle_new_region (ARDOUR::Region *);
	void add_region_to_region_display (ARDOUR::Region *);
	void redisplay_regions ();
	void insert_into_tmp_regionlist(ARDOUR::Region *);

	list<ARDOUR::Region *> tmp_region_list;
	list<ARDOUR::Playlist *> cut_buffer;
	void clear_cut_buffer ();

	enum CutCopyOp {
		CutOp,
		CopyOp
	};

	void cut_copy (CutCopyOp);
	void cut_copy_regions (CutCopyOp, RegionSelection&);
	void cut_copy_audio_ranges (CutCopyOp, list<ARDOUR::AudioRange>&, TrackViewList&, list<ARDOUR::Playlist*>& dest);
	void cut_copy_music_ranges (CutCopyOp, list<ARDOUR::MusicRange>&, TrackViewList&, list<ARDOUR::Playlist*>& dest);

	void paste_regions (TrackViewList& views, RegionSelection& regions, jack_nframes_t position, float times);
	void paste_playlists (TrackViewList& views, list<ARDOUR::Playlist*>& playlists, jack_nframes_t position, float times);
	void paste_internal (jack_nframes_t position, float times, bool use_cut_buffer);

	/* EDITING OPERATIONS */
	
	list<ARDOUR::Playlist *> *playlists_at_edit_cursor ();

	/* this is here because SigC++'s bind function doesn't handle
	   references very well, at least not with gcc 2.95.3.
	   the Playlist::add_region() function takes a reference,
	   and we need something to bind that takes a pointer.
	*/

	void fake_add_region (ARDOUR::Playlist *, ARDOUR::Region *, jack_nframes_t position);

	void toggle_region_mute ();
	void toggle_region_opaque ();
	void raise_region ();
	void raise_region_to_top ();
	void lower_region ();
	void lower_region_to_bottom ();
	void split_region ();
	void crop_region_to_selection ();
	void set_region_sync_from_edit_cursor ();
	void remove_region_sync();
	void mouse_align_region (ARDOUR::Region&, GdkEventButton *, jack_nframes_t position);
	void remove_some_regions ();
	void remove_clicked_region ();
	void edit_region ();
	void duplicate_some_regions (RegionSelection&, float times);
	void duplicate_selection (float times);
	void region_fill_selection ();
	void audition_playlist_region_standalone (ARDOUR::Region&);
	void audition_playlist_region_via_route (ARDOUR::Region&,ARDOUR::Route&);
	void split_multichannel_region();

	void cut ();
	void copy ();
	void paste (float times);
	void x_style_paste (jack_nframes_t pos, float times);

	int  get_prefix (float&, bool&);

	void keyboard_paste ();
	void keyboard_duplicate_region ();
	void keyboard_duplicate_selection ();
	void keyboard_nudge ();
	void keyboard_insert_region_list_selection ();

	void region_from_selection ();
	ARDOUR::Region* create_region_from_selection ();

	bool region_renamed;
	void rename_region ();
	void rename_region_finished (bool);

	void play_from_start ();
	void play_clicked_regionview ();
	void play_selected_region ();
	void play_location ();

	ZoomFocus zoom_focus;

	void temporal_zoom_step (bool coarser);
	void temporal_zoom_selection ();
	void temporal_zoom_session ();
	void temporal_zoom (gdouble scale);
	void temporal_zoom_by_frame (jack_nframes_t start, jack_nframes_t end, string op);
	void temporal_zoom_to_frame (bool coarser, jack_nframes_t frame);

	void amplitude_zoom (gdouble scale);
	void amplitude_zoom_step (bool in);

	ARDOUR::Region* region_list_display_drag_region;
	char            need_wave_cursor;
	gint            region_list_display_motion (GdkEventMotion*);

	void insert_region_list_drag (ARDOUR::Region&);
	void insert_region_list_selection (float times);

	void insert_sndfile ();

	/* all this .... */
	
	ArdourDialog* import_progress_window;
	Gtk::Label import_progress_label;
	Gtk::VBox import_progress_vbox;
	Gtk::ProgressBar import_progress_bar;
	Gtk::Button import_cancel_button;
	SigC::Connection  import_progress_connection;
	ARDOUR::Session::import_status import_status;
	static void *_import_thread (void *);
	gint import_progress_timeout (void *);
	void import_cancel_clicked ();
	void* import_thread ();
	void build_import_progress_window ();

	/* to support this ... */

	void import_audio ();

	void move_to_start ();
	void move_to_end ();
	void goto_frame ();
	void center_playhead ();
	void center_edit_cursor ();
	void edit_cursor_backward ();
	void edit_cursor_forward ();
	void playhead_backward ();
	void playhead_forward ();
	void scroll_backward ();
	void scroll_forward ();
	void scroll_tracks_down ();
	void scroll_tracks_up ();
	void delete_sample_forward ();
	void delete_sample_backward ();
	void delete_screen ();
	void search_backwards ();
	void search_forwards ();
	void set_mark ();
	void jump_forward_to_mark ();
	void jump_backward_to_mark ();
	void cursor_align (bool playhead_to_edit);

	void add_location_from_selection ();
	void set_selection_from_location ();
	void set_location_from_selection ();
	void set_route_loop_selection ();

	void add_location_from_playhead_cursor ();

	void reverse_selection ();
	void edit_envelope ();

	void start_scrolling ();
	void stop_scrolling ();

	void keyboard_selection_begin ();
	void keyboard_selection_finish ();
	bool have_pending_keyboard_selection;
	jack_nframes_t pending_keyboard_selection_start;

	ARDOUR::Region* select_region_for_operation (int dir, TimeAxisView **tv);
	void extend_selection_to_end_of_region (bool next);
	void extend_selection_to_start_of_region (bool previous);

	SnapType snap_type;

	void soundfile_chosen_for_insert (string selection, bool multichan_region);
	void soundfile_chosen_for_import (string selection, bool multichan_region);

	void handle_gui_changes (string, void *);

	void    hide_all_tracks (bool with_select);

	void route_list_selected (gint row, gint col, GdkEvent* ev);
	void route_list_unselected (gint row, gint col, GdkEvent* ev);
	void redisplay_route_list();
	gint route_list_reordered ();
	bool ignore_route_list_reorder;
	void queue_route_list_reordered (gint row, gint col);

	void auto_crossfade ();

	struct DragInfo {
	    GtkCanvasItem* item;
	    void* data;
	    jack_nframes_t last_frame_position;
	    jack_nframes_t pointer_frame_offset;
	    jack_nframes_t grab_frame;
	    jack_nframes_t last_pointer_frame;
	    jack_nframes_t current_pointer_frame;
	    double grab_x, grab_y;
	    double cumulative_x_drag;
	    double cumulative_y_drag;
	    double current_pointer_x;
	    double current_pointer_y;
	    void (Editor::*motion_callback)(GtkCanvasItem*, GdkEvent*);
	    void (Editor::*finished_callback)(GtkCanvasItem*, GdkEvent*);
	    TimeAxisView* last_trackview;
	    bool y_constrained;
	    bool x_constrained;
	    bool copy;
	    bool was_rolling;
	    bool been_moved;
	} drag_info;

	struct LineDragInfo {
	    unsigned long before;
	    unsigned long after;
	};

	LineDragInfo current_line_drag_info;

	void start_grab (GdkEvent*, GdkCursor* cursor = 0);
	void end_grab (GtkCanvasItem*, GdkEvent*);

	void cursor_drag_motion_callback (GtkCanvasItem*, GdkEvent*);
	void cursor_drag_finished_callback (GtkCanvasItem*, GdkEvent*);
	void marker_drag_motion_callback (GtkCanvasItem*, GdkEvent*);
	void marker_drag_finished_callback (GtkCanvasItem*, GdkEvent*);
	void region_drag_motion_callback (GtkCanvasItem*, GdkEvent*);
	void region_drag_finished_callback (GtkCanvasItem*, GdkEvent*);
	void region_copy_drag_finished_callback (GtkCanvasItem*, GdkEvent*);
	void control_point_drag_motion_callback (GtkCanvasItem*, GdkEvent*);
	void control_point_drag_finished_callback (GtkCanvasItem*, GdkEvent*);
	void line_drag_motion_callback (GtkCanvasItem*, GdkEvent*);
	void line_drag_finished_callback (GtkCanvasItem*, GdkEvent*);

	void tempo_marker_drag_motion_callback (GtkCanvasItem*, GdkEvent*);
	void tempo_marker_drag_finished_callback (GtkCanvasItem*, GdkEvent*);
	void meter_marker_drag_motion_callback (GtkCanvasItem*, GdkEvent*);
	void meter_marker_drag_finished_callback (GtkCanvasItem*, GdkEvent*);

	gint mouse_rename_region (GtkCanvasItem*, GdkEvent*);

	void start_region_grab (GtkCanvasItem*, GdkEvent*);
	void start_region_copy_grab (GtkCanvasItem*, GdkEvent*);
	void start_selection_grab (GtkCanvasItem*, GdkEvent*);
	void start_cursor_grab (GtkCanvasItem*, GdkEvent*);
	void start_marker_grab (GtkCanvasItem*, GdkEvent*);
	void start_control_point_grab (GtkCanvasItem*, GdkEvent*);
	void start_line_grab_from_regionview (GtkCanvasItem*, GdkEvent*);
	void start_line_grab_from_line (GtkCanvasItem*, GdkEvent*);
	void start_line_grab (AutomationLine *, GdkEvent*);
	void start_tempo_marker_grab (GtkCanvasItem*, GdkEvent*);
	void start_meter_marker_grab (GtkCanvasItem*, GdkEvent*);

	void region_view_item_click (RegionView&, GdkEventButton*);

	bool just_play;
	void start_scrubbing (GtkCanvasItem*, GdkEvent*);
	void scrub_motion_callback (GtkCanvasItem*, GdkEvent*);
	void scrub_finished_callback(GtkCanvasItem*, GdkEvent*);
	
	void remove_control_point (GtkCanvasItem*, GdkEvent*);

	bool mouse_brush_on;
	void mouse_brush_insert_region ();

	void show_verbose_time_cursor (jack_nframes_t frame, double offset = 0);

	/* static versions of these are public */
	
	gint canvas_region_view_event (GtkCanvasItem* item, GdkEvent* event, RegionView *);
	gint canvas_stream_view_event (GtkCanvasItem* item, GdkEvent* event, AudioTimeAxisView *);
	gint canvas_marker_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_zoom_rect_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_selection_rect_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_selection_start_trim_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_selection_end_trim_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_control_point_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_line_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_tempo_marker_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_meter_marker_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_tempo_bar_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_meter_bar_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_marker_bar_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_region_view_name_highlight_event (GtkCanvasItem* item, GdkEvent* event);
	gint canvas_region_view_name_event (GtkCanvasItem* item, GdkEvent* event);

	static gint _canvas_copy_region_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	gint canvas_copy_region_event (GtkCanvasItem* item, GdkEvent* event);

	static gint _canvas_playhead_cursor_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	gint canvas_playhead_cursor_event (GtkCanvasItem* item, GdkEvent* event);

	static gint _canvas_edit_cursor_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	gint canvas_edit_cursor_event (GtkCanvasItem* item, GdkEvent* event);
	
	static gint _track_canvas_event (GtkCanvasItem* item, GdkEvent* event, gpointer data);
	gint track_canvas_event (GtkCanvasItem* item, GdkEvent* event);

	gint track_canvas_button_press_event (GdkEventButton *);
	gint track_canvas_button_release_event (GdkEventButton *);
	
	void track_canvas_allocate (GtkAllocation* alloc);
	void time_canvas_allocate (GtkAllocation* alloc);

	void set_edit_cursor (GdkEvent* event);
	void set_playhead_cursor (GdkEvent* event);
	
	void fake_handle_new_duration ();
	void handle_new_duration ();
	void initialize_canvas ();
	void reset_scrolling_region (GtkAllocation* alloc = 0);
	void scroll_canvas ();

        /* sub-event loop handling */

        int sub_event_loop_status;
        void run_sub_event_loop ();
        void finish_sub_event_loop (int status);
        gint finish_sub_event_loop_on_delete (GdkEventAny*, int status);
	
	/* tempo display */
	
	bool _show_measures;
	bool _show_waveforms;

	void add_bbt_marks (ARDOUR::TempoMap::BBTPointList&);

	ARDOUR::TempoMap::BBTPointList *current_bbt_points;
	
	typedef vector<GtkCanvasItem*> TimeLineList;
	TimeLineList free_measure_lines;
	TimeLineList used_measure_lines;

	GtkCanvasItem* time_line_group;
	GtkCanvasItem* get_time_line ();
	void hide_measures ();
	void draw_measures ();
	void draw_time_bars ();

	void new_tempo_section ();

	void mouse_add_new_tempo_event (GdkEventButton*);
	void mouse_add_new_meter_event (GdkEventButton*);

	void remove_tempo_marker (GtkCanvasItem*);
	void remove_meter_marker (GtkCanvasItem*);
	gint real_remove_tempo_marker (ARDOUR::TempoSection*);
	gint real_remove_meter_marker (ARDOUR::MeterSection*);
	
	void edit_tempo_section (ARDOUR::TempoSection*);
	void edit_meter_section (ARDOUR::MeterSection*);
	void edit_tempo_marker (GtkCanvasItem*);
	void edit_meter_marker (GtkCanvasItem*);
	
	void popup_tempo_marker_menu (int button, int time, GtkCanvasItem*);
	void popup_meter_marker_menu (int button, int time, GtkCanvasItem*);

	typedef list<Marker*> Marks;
	Marks metric_marks;

	void remove_metric_marks ();
	void draw_metric_marks (const ARDOUR::TempoMap::Metrics& metrics);

	void tempo_map_changed ();
	void redisplay_tempo ();
	
	void snap_to (jack_nframes_t& first, int direction = 0, bool for_mark = false);
	unsigned long bbt_beat_subdivision;

	/* toolbar */

	AudioClock               selection_start_clock;
	Gtk::Label               selection_start_clock_label;
	AudioClock               selection_end_clock;
	Gtk::Label               selection_end_clock_label;
	AudioClock               edit_cursor_clock;
	Gtk::Label               edit_cursor_clock_label;

	Gtk::VBox                toolbar_clock_vbox;
	Gtk::VBox                toolbar_selection_clock_vbox; 
	Gtk::Table               toolbar_selection_clock_table;
	Gtk::Label               toolbar_selection_cursor_label;
	
	Gtk::Table               mouse_mode_button_table;
	Gtkmmext::TearOff*       mouse_mode_tearoff;
	Gtk::ToggleButton        mouse_select_button;
	Gtk::ToggleButton        mouse_move_button;
	Gtk::ToggleButton        mouse_gain_button;
	Gtk::ToggleButton        mouse_pan_button;
	Gtk::ToggleButton        mouse_automation_button;
	Gtk::ToggleButton        mouse_zoom_button;
	Gtk::ToggleButton        mouse_scrub_button;
	Gtk::ToggleButton        mouse_timefx_button;
	GroupedButtons          *mouse_mode_button_set;
	void                     mouse_mode_toggled (MouseMode m);

	gint                     mouse_select_button_release (GdkEventButton*);

	Gtk::VBox                automation_box;
	Gtk::Button              automation_mode_button;
	Gtk::ToggleButton        global_automation_button;

	Gtk::Combo               edit_mode_selector;
	Gtk::Label               edit_mode_label;
	Gtk::VBox                edit_mode_box;

	gint edit_mode_selection_done (GdkEventAny*);

	Gtk::Combo               snap_mode_selector;
	Gtk::Label               snap_mode_label;
	Gtk::VBox                snap_mode_box;

	gint snap_mode_selection_done (GdkEventAny*);

	Gtk::Combo               zoom_focus_selector;
	Gtk::Label               zoom_focus_label;
	Gtk::VBox                zoom_focus_box;
	
	gint zoom_focus_selection_done (GdkEventAny*);

	Gtk::Adjustment     zoom_adjustment;
	Gtkmmext::ClickBox  zoom_indicator;
	Gtk::Frame          zoom_indicator_frame;
	Gtk::Label          zoom_indicator_label;
	Gtk::HBox           zoom_indicator_box;
	Gtk::VBox           zoom_indicator_vbox;

	void                update_zoom_indicator ();
	static void         zoom_printer (char buf[32], Gtk::Adjustment& adj, void *ignored);
	void                zoom_adjustment_changed();

	void setup_toolbar ();

	Gtkmmext::TearOff*       tools_tearoff;
	Gtk::HBox                toolbar_hbox;
	Gtk::EventBox            toolbar_base;
	Gtk::Frame               toolbar_frame;

	/* selection process */

	Selection* selection;

	void time_selection_changed ();
	void region_selection_changed ();
	void audio_track_selection_changed ();
	void line_selection_changed ();

	enum SelectionOp {
		CreateSelection,
		SelectionStartTrim,
		SelectionEndTrim,
		SelectionMove
	} selection_op;

	void start_selection_op (GtkCanvasItem* item, GdkEvent* event, SelectionOp);
	void drag_selection (GtkCanvasItem* item, GdkEvent* event);
	void end_selection_op (GtkCanvasItem* item, GdkEvent* event);
	void cancel_selection ();

	Gtk::VPaned  route_group_vpane;
	Gtk::Frame   route_list_frame;
	Gtk::Frame   edit_group_list_frame;

	/* mouse zoom process */

	void start_mouse_zoom (GtkCanvasItem* item, GdkEvent* event);
	void drag_mouse_zoom (GtkCanvasItem* item, GdkEvent* event);
	void end_mouse_zoom (GtkCanvasItem* item, GdkEvent* event);

	GtkCanvasItem   *zoom_rect;
	void reposition_zoom_rect (jack_nframes_t start, jack_nframes_t end);
	
	/* diskstream/route display management */

	Gtk::CList          route_list;
	Gtk::ScrolledWindow route_list_scroller;
	Gtk::Menu          *route_list_menu;

	void route_list_column_click (gint);
	void build_route_list_menu ();
	void select_all_routes ();
	void unselect_all_routes ();
	void select_all_audiotracks ();
	void unselect_all_audiotracks ();
	void select_all_audiobus ();
	void unselect_all_audiobus ();

	/* edit group management */

	Gtk::Button         edit_group_list_button;
	Gtk::Label          edit_group_list_button_label;
	Gtk::CList          edit_group_list;
	Gtk::ScrolledWindow edit_group_list_scroller;
	Gtk::Menu          *edit_group_list_menu;
	Gtk::VBox           edit_group_vbox;


	void edit_group_list_column_click (gint);
	void build_edit_group_list_menu ();
	void select_all_edit_groups ();
	void unselect_all_edit_groups ();
	void new_edit_group ();
	void edit_group_list_button_clicked ();
	gint edit_group_list_button_press_event (GdkEventButton* ev);
	void edit_group_selected (gint row, gint col, GdkEvent* ev);
	void edit_group_unselected (gint row, gint col, GdkEvent* ev);
	void fake_add_edit_group (ARDOUR::RouteGroup* group);
	void add_edit_group (ARDOUR::RouteGroup* group);
	void group_flags_changed (void*, ARDOUR::RouteGroup*);

	Gtk::VBox           list_vpacker;

	static GdkPixmap* check_pixmap;
	static GdkBitmap* check_mask;
	static GdkPixmap* empty_pixmap;
	static GdkBitmap* empty_mask;

	/* autoscrolling */

	int autoscroll_timeout_tag;
	int autoscroll_direction;
	unsigned long autoscroll_cnt;
	jack_nframes_t autoscroll_distance;
     
	static gint _autoscroll_canvas (void *);
	gint autoscroll_canvas ();
	void start_canvas_autoscroll (int direction);
	void stop_canvas_autoscroll ();
	void maybe_autoscroll (GdkEvent*);

	/* trimming */
	enum TrimOp {
		StartTrim,
		EndTrim,
		ContentsTrim,
	} trim_op;

	void start_trim (GtkCanvasItem*, GdkEvent*);
	void point_trim (GdkEvent*);
	void trim_motion_callback (GtkCanvasItem*, GdkEvent*);
	void single_contents_trim (RegionView&, jack_nframes_t, bool, bool);
	void single_start_trim (RegionView&, jack_nframes_t, bool);
	void single_end_trim (RegionView&, jack_nframes_t, bool);

	void trim_finished_callback (GtkCanvasItem*, GdkEvent*);
	void thaw_region_after_trim (RegionView& rv);
	
	void trim_region_to_edit_cursor ();
	void trim_region_from_edit_cursor ();

	bool show_gain_after_trim;

	/* Drag-n-Drop */

	void  track_canvas_drag_data_received  (GdkDragContext     *context,
						gint                x,
						gint                y,
						GtkSelectionData   *data,
						guint               info,
						guint               time);
	
	/* cd export */

	void export_cd_markers_callback (ARDOUR::Locations::LocationList&, const string&);
	void do_export_cd_markers (const string& path);
	void finish_export_cd_markers (bool status);
	bool export_cd_marker_status;
	
	/* audio export */

	ExportDialog *export_dialog;

	struct SillyThingForPassingArgsToExportThread {
	    Editor& editor;
	    ARDOUR::AudioExportSpecification& spec;

	    SillyThingForPassingArgsToExportThread (Editor&, ARDOUR::AudioExportSpecification&);
	};

	void export_range (jack_nframes_t start, jack_nframes_t end);

	static void* _export_thread (void*);
	int run_export (ARDOUR::AudioExportSpecification&);

	int write_region_selection(RegionSelection&);
	bool write_region (string path, ARDOUR::Region&);
	void export_region ();

	int write_audio_selection (TimeSelection&);
	bool write_audio_range (ARDOUR::Playlist&, unsigned int channels, list<ARDOUR::AudioRange>&);

	void write_selection ();

	/* history */

	UndoAction get_memento() const;

	void begin_reversible_command (string cmd_name);
	void commit_reversible_command ();

	struct State {
	    Selection* selection;
	    double     frames_per_unit;

	    State();
	    ~State();
	};

	void store_state (State&) const;
	void restore_state (State *);

	ARDOUR::Region* last_audition_region;

	/* flatten operations */
	
	void flatten_route ();
	void flatten_selection ();

	/* edit-group solo + mute */

	void set_edit_group_solo (ARDOUR::Route&, bool);
	void set_edit_group_mute (ARDOUR::Route&, bool);

	/* duplication */

	void duplicate_dialog (bool for_region);
	
	/* edit menu */

	Gtk::Menu* edit_menu;
	void edit_menu_map_handler ();

	/* scrubbing */

	jack_nframes_t scrub_frame;
	jack_nframes_t last_scrub_frame;
	float          last_scrub_speed;
	bool          _scrubbing;
	gint           scrub_timeout_tag;
	static gint   _check_scrub (gpointer);
	gint           check_scrub();

	void start_scrubbing_at (jack_nframes_t);
	void stop_scrubbing_at ();
	void scrub_motion (jack_nframes_t);

	jack_nframes_t event_frame (GdkEvent*, double* px = 0, double* py = 0);

	void time_fx_motion (GtkCanvasItem*, GdkEvent*);
	void start_time_fx (GtkCanvasItem*, GdkEvent*);
	void end_time_fx (GtkCanvasItem*, GdkEvent*);

	struct TimeStretchDialog : public ArdourDialog {
	    ARDOUR::Session::TimeStretchRequest *tsr;
	    Editor&           editor;
	    Gtk::ProgressBar  progress_bar;
	    Gtk::ToggleButton quick_button;
	    Gtk::ToggleButton antialias_button;
	    Gtk::Button       cancel_button;
	    Gtk::Button       action_button;
	    Gtk::HBox         lower_button_box;
	    Gtk::HBox         upper_button_box;
	    Gtk::VBox         packer;

	    TimeStretchDialog (Editor& e, TimeAxisView&, RegionView&);
	    gint update_progress ();
	    SigC::Connection first_cancel;
	    SigC::Connection first_delete;
	    void cancel_timestretch_in_progress ();
	    gint delete_timestretch_in_progress (GdkEventAny*);
	};

	/* "whats mine is yours" */

	friend class TimeStretchDialog;

	TimeStretchDialog* current_timestretch;

	static void* timestretch_thread (void *arg);
	void start_timestretch (TimeAxisView&, RegionView&, jack_nframes_t start, jack_nframes_t end);
	void do_timestretch (TimeStretchDialog&);
	void _do_timestretch (TimeStretchDialog&);

	/* editor-mixer strip */

	MixerStrip *current_mixer_strip;
	Gtk::VBox current_mixer_strip_vbox;
	void current_mixer_strip_hidden ();
	void current_mixer_strip_removed ();
	
	/* selection choices */

	Gtk::Table         selection_mode_table;
	Gtkmmext::TearOff* selection_mode_tearoff;
	Gtk::ToggleButton  select_data_button;
	Gtk::ToggleButton  select_gain_button;
	Gtk::ToggleButton  select_pan_button;
	Gtk::ToggleButton  select_mute_button;
	Gtk::ToggleButton  select_solo_button;
	Gtk::ToggleButton  select_redirect_button;

	SelectionMode _selection_mode;
	
	void select_data_clicked ();
	void select_gain_clicked ();
	void select_pan_clicked ();
	void select_mute_clicked ();
	void select_solo_clicked ();
	void select_redirect_clicked ();

	void detach_tearoff (Gtk::Box* b, Gtk::Widget* w);
	void reattach_tearoff (Gtk::Box* b, Gtk::Widget* w, int n);

	/* nudging tracks */

	void nudge_track (bool use_edit_cursor, bool forwards);
};

#endif /* __ardour_editor_h__ */
