/*
    Copyright (C) 2001 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: regionview.cc,v 1.73 2003/06/26 02:56:41 pbd Exp $
*/

#include <gtk--.h>
#include <math.h>

#include <gtkmmext/gtk_ui.h>

#include <ardour/playlist.h>
#include <ardour/region.h>
#include <ardour/sndfilesource.h>
#include <ardour/diskstream.h>

#include "streamview.h"
#include "regionview.h"
#include "audio_time_axis.h"
#include "canvas-simplerect.h"
#include "canvas-simpleline.h"
#include "canvas-waveview.h"
#include "editor.h"
#include "region_editor.h"
#include "region_gain_line.h"
#include "utils.h"

using namespace SigC;
using namespace ARDOUR;

SigC::Signal1<void,RegionView*> RegionView::GoingAway;

const string name_font = "-*-helvetica-medium-r-normal-*-*-100-*-*-p-*-*-1";
const double name_x_offset = 4.0; 
const double name_y_offset = 15.0; /* XXX depends a lot on the font size, sigh. */
const double name_highlight_size = 15.0; /* ditto */

RegionView::RegionView (GtkCanvasGroup *parent, AudioTimeAxisView &tv, 
			Region& r, 
			double spu, 
			double amplitude_above_axis,
			GdkColor& basic_color)

	: trackview (tv), 
	  region (r)
{
	GtkCanvasPoints *shape;
	gdouble pixel_width;
	XMLNode *node;

	editor = 0;
	valid = true;
	samples_per_unit = spu;
	name_connected = false;
	current_visible_sync_position = 0;
	_amplitude_above_axis = amplitude_above_axis;
	zero_line = 0;

	no_wave_msg = 0;
	_flags = 0;

	if ((node = region.extra_xml ("GUI")) != 0) {
		set_flags (node);
	} else {
		_flags = WaveformVisible;
		store_flags ();
	}

	compute_colors (basic_color);

	group = gtk_canvas_item_new (parent, 
				     gtk_canvas_group_get_type (), 
				     NULL);
	
	pixel_width = region.length() / samples_per_unit;
	
	frame = gtk_canvas_item_new (GTK_CANVAS_GROUP(group),
				     gtk_canvas_simplerect_get_type(),
				     "x1", (gdouble) 0.0,
				     "y1", (gdouble) 0.0,
				     NULL);

	name_highlight = gtk_canvas_item_new (GTK_CANVAS_GROUP(group),
					      gtk_canvas_simplerect_get_type(),
					      "x1", (gdouble) 1.0,
					      NULL);

	gtk_object_set_data (GTK_OBJECT(name_highlight), "regionview", this);

	create_wave ();

	name = gtk_canvas_item_new (GTK_CANVAS_GROUP(group),
				    gtk_canvas_text_get_type (),
				    "x", (gdouble) 4.0, 
				    "y", (gdouble) trackview.height - name_y_offset,
				    "font", name_font.c_str(),
				    "anchor", GTK_ANCHOR_NW,
				    NULL);

	gtk_object_set_data (GTK_OBJECT(name), "regionview", this);

	gain_line = new RegionGainLine ("gain", tv.session(), *this, tv.editor, group, region.envelope(),
					Editor::_canvas_control_point_event,
					Editor::_canvas_line_event);


	if (!(_flags & EnvelopeVisible)) {
		gain_line->hide ();
	} 

	shape = gtk_canvas_points_new (4);

	/* a 6x6 equilateral triangle */

	shape->coords[0] = 0;
	shape->coords[1] = 0;

	shape->coords[2] = 5;
	shape->coords[3] = 0;

	shape->coords[4] = 2;
	shape->coords[5] = 5;

	shape->coords[6] = 0;
	shape->coords[7] = 0;

	sync_mark = gtk_canvas_item_new (GTK_CANVAS_GROUP(group),
					 gtk_canvas_polygon_get_type(),
					 "points", shape,
					 NULL);
	gtk_canvas_item_hide (sync_mark);

	gtk_canvas_points_unref (shape);

	reset_width_dependent_items (pixel_width);

	gain_line->reset ();
	set_colors ();
	set_height (trackview.height - 2);

	region_muted ();
	region_sync_changed ();
	region_resized (Region::BoundsChangeInfo(~0U), 0);

	region.BoundsChanged.connect (slot (*this, &RegionView::region_resized));
	region.MuteChanged.connect (slot (*this, &RegionView::region_muted));
	region.OpacityChanged.connect (slot (*this, &RegionView::region_opacity));
	region.NameChanged.connect (slot (*this, &RegionView::region_renamed));
	region.SyncOffsetChanged.connect (slot (*this, &RegionView::region_sync_changed));

	gtk_signal_connect (GTK_OBJECT(group), "event",
			    (GtkSignalFunc) Editor::_canvas_region_view_event,
			    this);
	gtk_signal_connect (GTK_OBJECT(name_highlight), "event",
			    (GtkSignalFunc) Editor::_canvas_region_view_name_highlight_event,
			    this);

	gtk_signal_connect (GTK_OBJECT(name), "event",
			    (GtkSignalFunc) Editor::_canvas_region_view_name_event,
			    this);

	/* XXX sync mark drag? */
}

RegionView::~RegionView ()
{
	EMIT_SIGNAL GoingAway (this);

	for (vector<GtkCanvasItem *>::iterator wave = waves.begin(); wave != waves.end() ; ++wave) {
		gtk_object_destroy (GTK_OBJECT(*wave));
	}
	
	if (no_wave_msg) {
		gtk_object_destroy (GTK_OBJECT(no_wave_msg));
	}

	if (editor) {
		delete editor;
	}

	delete gain_line;

	gtk_object_destroy (GTK_OBJECT(frame));
	gtk_object_destroy (GTK_OBJECT(name));
	gtk_object_destroy (GTK_OBJECT(name_highlight));
	gtk_object_destroy (GTK_OBJECT(group));
}

void
RegionView::reset_name (gdouble pixel_width)
{
	gint width;
	gint lbearing;
	gint rbearing;
	gint ascent;
	gint descent;
	Gdk_Font font (name_font);

	int namelen = region.name().length();
	char cstr[namelen+1];
	strcpy (cstr, region.name().c_str());
	
	while (namelen) {
		
		gdk_string_extents (font,
				    cstr,
				    &lbearing,
				    &rbearing,
				    &width,
				    &ascent,
				    &descent);

		if (width < (pixel_width - name_x_offset)) {
			break;
		}

		--namelen;
		cstr[namelen] = '\0';

	}

	if (namelen == 0) {
		
		gtk_canvas_item_hide (name);
		
	} else {
		
		/* don't use name for event handling if it leaves no room
		   for trimming to work.
		*/
		
		if (pixel_width - width < (name_x_offset * 2.0)) {
			if (name_connected) {
				name_connected = false;
			}
		} else {
			if (!name_connected) {
				name_connected = true;
			}
		}
		
		gtk_canvas_item_set (name, "text", cstr, NULL);
		gtk_canvas_item_show (name);
	}
}

void
RegionView::region_resized (Region::BoundsChangeInfo what_changed, void *src)
{
	if (what_changed & Region::PositionChange) {
		set_frame_position (region.position());
	}

	if (what_changed & (Region::StartChange|Region::LengthChange)) {

	        gdouble pixel_width = (gdouble) region.length() / samples_per_unit;

		for (unsigned int n=0; n < waves.size(); ++n) {
			// TODO: something else to let it know the channel
			gtk_canvas_item_set (waves[n], "data_src", &region, NULL);
		}

		reset_width_dependent_items (pixel_width);
	}

}
void
RegionView::reset_width_dependent_items (double pixel_width)
{
	if (pixel_width < 1.0) {
		gtk_canvas_item_hide (group);
	} else {
		gtk_canvas_item_show (group);

		reset_name (pixel_width);

		gtk_canvas_item_set (name_highlight, "x2", pixel_width - 1.0, NULL);
		gtk_canvas_item_set (frame, "x2", pixel_width, NULL);
		
		if (zero_line) {
			gtk_canvas_item_set (zero_line, "x2", pixel_width, NULL);
		}
	}
}
	
void
RegionView::region_muted ()
{
	set_frame_color ();

	for (unsigned int n=0; n < waves.size(); ++n) {
		if (region.muted()) {
			gtk_canvas_item_set (waves[n], "wave_color", muted_wave_color, NULL);
		}
		else {
			gtk_canvas_item_set (waves[n], "wave_color", wave_color, NULL);
		}
	}
}

void
RegionView::set_selected (bool yn)
{
	if (yn != (bool) (_flags & Selected)) {
		if (yn) {
			_flags = _flags | Selected;
		} else {
			_flags = _flags & ~Selected;
		}
		set_frame_color ();
	}
}

void
RegionView::region_opacity ()
{
	set_frame_color ();
}

void
RegionView::raise ()
{
	region.raise ();
}

void
RegionView::raise_to_top ()
{
	region.raise_to_top ();
}

void
RegionView::lower ()
{
	region.lower ();
}

void
RegionView::lower_to_bottom ()
{
	region.lower_to_bottom ();
}

void
RegionView::set_height (gdouble height)
{
	unsigned int wcnt = waves.size();
	
	for (unsigned int n=0; n < wcnt; ++n) {
		gdouble ht;

		if ((height/wcnt) < name_highlight_size) {
			ht = ((height) / (double) wcnt);
		} else {
			ht = ((height - name_highlight_size) / (double) wcnt);
		}
		
		gdouble yoff = n * ht;
		
		gtk_canvas_item_set (waves[n], "height", ht, NULL);
		gtk_canvas_item_set (waves[n], "y", yoff, NULL);
	}

	if ((height/wcnt) < name_highlight_size) {
		gain_line->hide ();
		gtk_canvas_item_hide (name_highlight);

		if (no_wave_msg) {
			gtk_canvas_item_hide (no_wave_msg);
		}

	} else {
		if (_flags & EnvelopeVisible) {
			gain_line->show ();
		}

		gtk_canvas_item_show (name_highlight);

		if (no_wave_msg) {
			gtk_canvas_item_show (no_wave_msg);
		}
	}

	gtk_canvas_item_set (name, "y", height - name_y_offset, NULL);
	gtk_canvas_item_set (frame, "y2", height, NULL);
	
	if (height >= 100) {
		gdouble wave_midpoint = (height - name_highlight_size) / 2.0;
		gtk_canvas_item_set (zero_line, "y1", wave_midpoint, "y2", wave_midpoint, NULL);
		gtk_canvas_item_show (zero_line);
	} else {
		gtk_canvas_item_hide (zero_line);
	}

	if (height > name_highlight_size) {
		gtk_canvas_item_set (name_highlight, 
				     "y1", (gdouble) height - name_highlight_size,
				     "y2", (gdouble) height - 1.0,
				     NULL);
	}
	else {
		gtk_canvas_item_set (name_highlight, 
				     "y1", (gdouble) 1.0,
				     "y2", (gdouble) height - 1.0,
				     NULL);
	}

	gain_line->set_height ((unsigned int) rint (height));
}

void
RegionView::set_frame_position (jack_nframes_t pos)
{
	/* This sucks. The GtkCanvas version I am using
	   doesn't correctly implement gtk_canvas_group_set_arg(),
	   so that simply setting the "x" arg of the group
	   fails to move the group. Instead, we have to
	   use gtk_canvas_item_move(), which does the right
	   thing. I see that in GNOME CVS, the current (Sept 2001)
	   version of GNOME Canvas rectifies this issue cleanly.
	*/

	GtkArg args[1];
	double old_unit_pos;
	double new_unit_pos = pos / samples_per_unit;

	args[0].name = "x";
	gtk_object_getv (GTK_OBJECT(group), 1, args);
	old_unit_pos = GTK_VALUE_DOUBLE (args[0]);

	gtk_canvas_item_move (group, new_unit_pos - old_unit_pos, 0.0);
}

void
RegionView::set_samples_per_unit (gdouble spu)
{
	gdouble pixel_width;

	samples_per_unit = spu;

	pixel_width = (gdouble) region.length() / samples_per_unit;

	set_frame_position (region.position());

	for (unsigned int n=0; n < waves.size(); ++n) {
		gtk_canvas_item_set (waves[n], "samples_per_unit", spu, NULL);
	}

	reset_width_dependent_items (pixel_width);
	gain_line->reset ();
	region_sync_changed ();
}

void
RegionView::set_amplitude_above_axis (gdouble spp)

{
	for (unsigned int n=0; n < waves.size(); ++n) {
		gtk_canvas_item_set (waves[n], "amplitude_above_axis", spp, NULL);
	}
}

void
RegionView::compute_colors (GdkColor& base_color)
{
	unsigned char radius;
	char minor_shift;
	unsigned char r, g, b;

	/* FILL: this is simple */

	r = base_color.red/256;
	g = base_color.green/256;
	b = base_color.blue/256;

	fill_color = RGBA_TO_UINT(r,g,b,255);

	/* for minor colors:

	   if the overall saturation is strong, make the minor colors light.
	   if its weak, make them dark.

	   we do this by moving an equal distance to the other side of the
	   central circle in the color wheel from where we started.
	*/

	radius = (unsigned char) rint (floor (sqrt (static_cast<double>(r*r + g*g + b+b))/3.0f));
	minor_shift = 125 - radius;

	/* WAVE: rotate around color wheel by 120 degrees clockwise */

	r = base_color.red/256;
	g = base_color.green/256;
	b = base_color.blue/256;

	if (r > b) {
		if (r > g) {
			/* red sector => blue */
			swap (r,b);
		} else {
			/* green sector => red */
			swap (g,r);
		} 
	} else {
		if (b > g) {
			/* blue sector => green */
			swap (b,g);
		} else {
			/* green sector => red */
			swap (g,r);
		}
	}

	r += minor_shift;
	b += minor_shift;
	g += minor_shift;
	
        r = (base_color.red/256)   + 127;
        g = (base_color.green/256) + 127;
        b = (base_color.blue/256)  + 127;

	wave_color = RGB_TO_UINT(110,32,155); /* used to be r,g,b */
	muted_wave_color = RGB_TO_UINT(160,82,205); /* used to be r,g,b */
	
	
	/* LABEL: rotate around color wheel by 120 degrees anti-clockwise */

	r = base_color.red/256;
	g = base_color.green/256;
	b = base_color.blue/256;

	if (r > b) {
		if (r > g) {
			/* red sector => green */
			swap (r,g);
		} else {
			/* green sector => blue */
			swap (g,b);
		} 
	} else {
		if (b > g) {
			/* blue sector => red */
			swap (b,r);
		} else {
			/* green sector => blue */
			swap (g,b);
		}
	}
	
	r += minor_shift;
	b += minor_shift;
	g += minor_shift;
	
	label_color = RGBA_TO_UINT(r,g,b,255);

        r = (base_color.red/256)   + 127;
        g = (base_color.green/256) + 127;
        b = (base_color.blue/256)  + 127;

	label_color = RGBA_TO_UINT(r,g,b,255);

	/* XXX DO SOMETHING */

	gain_color  = RGBA_TO_UINT(13,19,23,255);

	/* XXX can we do better than this ? */

	frame_color_r = 192;
	frame_color_g = 192;
	frame_color_b = 194;

	selected_frame_color_r = 243;
	selected_frame_color_g = 242;
	selected_frame_color_b = 0;
}

void
RegionView::set_colors ()
{
	set_frame_color ();

	gain_line->set_line_color (gain_color);

	gtk_canvas_item_set (name, "fill_color_rgba", label_color, NULL);
	gtk_canvas_item_set (sync_mark, "fill_color_rgba", fill_color, NULL);

	gtk_canvas_item_set (name_highlight, "fill_color_rgba", fill_color, NULL);
	gtk_canvas_item_set (name_highlight, "outline_color_rgba", RGB_TO_UINT(0,0,0), NULL);

	for (unsigned int n=0; n < waves.size(); ++n) {
		if (region.muted()) {
			gtk_canvas_item_set (waves[n], "wave_color", muted_wave_color, NULL);
		} else {
			gtk_canvas_item_set (waves[n], "wave_color", wave_color, NULL);
		}
	}
}

void
RegionView::set_color (GdkColor& base_color)
{
	compute_colors (base_color);
	set_colors ();
}

void
RegionView::set_frame_color ()
{
	int a;

	if (region.muted()) {
		a = 50;
	} else if (region.opaque()) {
		a = 255;
	} else {
		a = 70;
	}

	if (_flags & Selected) {
		gtk_canvas_item_set (frame, "fill_color_rgba", RGBA_TO_UINT(selected_frame_color_r, selected_frame_color_g, selected_frame_color_b,a), NULL);
	} else {
		gtk_canvas_item_set (frame, "fill_color_rgba", RGBA_TO_UINT(frame_color_r, frame_color_g, frame_color_b,a), NULL);
	}

}

void
RegionView::show_region_editor ()
{
	if (editor == 0) {
		editor = new RegionEditor (trackview.session(), region, *this);
		editor->realize ();
		trackview.editor.ensure_float (*editor);
	} 

	editor->show_all ();
	editor->get_window().raise();
}

void
RegionView::hide_region_editor()
{
	if (editor) {
		editor->hide_all ();
	}
}

void
RegionView::region_renamed ()
{
	gtk_canvas_item_set (name, "text", region.name().c_str(), NULL);
}

void
RegionView::region_sync_changed ()
{
	/* this has to handle both a genuine change of position and a change of samples_per_unit
	 */

	if (region.sync_offset() == 0) {
		gtk_canvas_item_hide (sync_mark);
		current_visible_sync_position = 0;
	} else {
		double new_position = region.sync_offset() / samples_per_unit;
		double delta = new_position - current_visible_sync_position;
		gtk_canvas_item_show (sync_mark);
		gtk_canvas_item_move (sync_mark, delta, 0);
		current_visible_sync_position = new_position;
	}
}

void
RegionView::set_waveform_visible (bool yn)
{
	if (((_flags & WaveformVisible) != yn)) {
		if (yn) {
			for (unsigned int n=0; n < waves.size(); ++n) {
				gtk_canvas_item_show (waves[n]);
			}
			_flags |= WaveformVisible;
		} else {
			for (unsigned int n=0; n < waves.size(); ++n) {
				gtk_canvas_item_hide (waves[n]);
			}
			_flags &= ~WaveformVisible;
		}
		store_flags ();
	}
}

void
RegionView::set_envelope_visible (bool yn)
{
	if ((_flags & EnvelopeVisible) != yn) {
		if (yn) {
			gain_line->show ();
			_flags |= EnvelopeVisible;
		} else {
			gain_line->hide ();
			_flags &= ~EnvelopeVisible;
		}
		store_flags ();
	}
}

void
RegionView::create_wave ()
{
	bool create_zero = false;

	if (waves.size() == 0 && trackview.get_diskstream()) {
		unsigned int nchans = trackview.get_diskstream()->n_channels();
		for (unsigned int n=0; n < nchans; ++n) {
			if (n >= region.n_channels()) {
				break;
			}
			
			gdouble ht;
			if (trackview.height < name_highlight_size) {
				ht = ((trackview.height) / (double) nchans);
			} else {
				ht = ((trackview.height - name_highlight_size) / (double) nchans);
			}
			gdouble yoff = n * ht;

			if (region.source(n).peaks_ready (slot (*this, &RegionView::peaks_ready_handler))) {
				
				create_zero = true;

				GtkCanvasItem *wave = gtk_canvas_item_new (GTK_CANVAS_GROUP(group),
							    gtk_canvas_waveview_get_type (),
							    "data_src", (gpointer) &region,
							    "channel", (guint32) n,
							    "length_function", (gpointer) region_length_from_c,
							    "peak_function", (gpointer) region_read_peaks_from_c,
							    "x", 0.0,
							    "y", yoff,
							    "height", (double) ht,
							    "samples_per_unit", samples_per_unit,
							    "amplitude_above_axis", _amplitude_above_axis,
							    "wave_color", (guint32) wave_color,
							    NULL);
				
				if (!(_flags & WaveformVisible)) {
					gtk_canvas_item_hide (wave);
				}

				waves.push_back(wave);
			}
		}
	}

	if (create_zero) {
		zero_line = gtk_canvas_item_new (GTK_CANVAS_GROUP(group),
						 gtk_canvas_simpleline_get_type(),
						 "x1", (gdouble) 0.0,
						 "x2", (gdouble) region.length() / samples_per_unit,
						 "color_rgba", (guint) RGBA_TO_UINT(181,181,181,255),
						 NULL);
	}
}

void
RegionView::peaks_ready_handler ()
{
	Gtkmmext::UI::instance()->call_slot (slot (*this, &RegionView::create_wave));
}

void
RegionView::add_gain_point_event (GtkCanvasItem *item, GdkEvent *ev)
{
	double x, y;

	/* don't create points that can't be seen */

	set_envelope_visible (true);
	
	x = ev->button.x;
	y = ev->button.y;

	gtk_canvas_item_w2i (item, &x, &y);

	jack_nframes_t fx = trackview.editor.pixel_to_frame (x);

	if (fx > region.last_frame()) {
		return;
	}

	gain_t g = slider_position_to_gain (1.0 - (y / trackview.height));

	g = max (0.0f, g);
	g = min (2.0f, g);

	trackview.session().begin_reversible_command ("add gain control point");
	trackview.session().add_undo (region.playlist()->get_memento());
	region.add_envelope_control_point (fx, g);
	trackview.session().add_redo_no_execute (region.playlist()->get_memento());
	trackview.session().commit_reversible_command ();
}

void
RegionView::remove_gain_point_event (GtkCanvasItem *item, GdkEvent *ev)
{
	ControlPoint *cp = reinterpret_cast<ControlPoint *> (gtk_object_get_data(GTK_OBJECT(item), "control_point"));
	region.remove_envelope_control_point (cp->model_index);
}

void
RegionView::store_flags()
{
	XMLNode *node = new XMLNode ("GUI");

	node->add_property ("waveform-visible", (_flags & WaveformVisible) ? "yes" : "no");
	node->add_property ("envelope-visible", (_flags & EnvelopeVisible) ? "yes" : "no");

	region.add_extra_xml (*node);
}

void
RegionView::set_flags (XMLNode* node)
{
	XMLProperty *prop;

	_flags = (_flags & Selected);

	if ((prop = node->property ("waveform-visible")) != 0) {
		if (prop->value() == "yes") {
			_flags |= WaveformVisible;
		}
	}

	if ((prop = node->property ("envelope-visible")) != 0) {
		if (prop->value() == "yes") {
			_flags |= EnvelopeVisible;
		}
	}
}
	
void
RegionView::set_waveform_shape (WaveformShape shape)
{
	bool yn;

	/* this slightly odd approach is to leave the door open to 
	   other "shapes" such as spectral displays, etc.
	*/

	switch (shape) {
	case Rectified:
		yn = true;
		break;

	default:
		yn = false;
		break;
	}

	if (yn != (bool) (_flags & WaveformRectified)) {
		for (vector<GtkCanvasItem *>::iterator wave = waves.begin(); wave != waves.end() ; ++wave) {
			gtk_canvas_item_set ((*wave), "rectified", (gboolean) yn, NULL);
		}

		if (yn) {
			gtk_canvas_item_hide (zero_line);
		} else {
			gtk_canvas_item_show (zero_line);
		}
		
		if (yn) {
			_flags |= WaveformRectified;
		} else {
			_flags &= ~WaveformRectified;
		}
	}
}


