/*
    Copyright (C) 2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: selection.cc,v 1.9 2003/05/13 13:48:47 pbd Exp $
*/

#include <algorithm>
#include <sigc++/bind.h>
#include <pbd/error.h>

#include <ardour/route.h>

#include "regionview.h"
#include "selection.h"
#include "selection_templates.h"
#include "time_axis_view.h"

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

struct AudioRangeComparator {
    bool operator()(AudioRange a, AudioRange b) {
	    return a.start < b.start;
    }
};

struct RegionLayerComparator {
	bool operator() (const RegionView* a, const RegionView* b) const {
		if (a->region.layer() == b->region.layer()) return false;
		return a->region.layer() < b->region.layer();
	}
};

Selection&
Selection::operator= (const Selection& other)
{
	if (&other != this) {
		regions = other.regions;
		tracks = other.tracks;
		time = other.time;
		lines = other.lines;
	}
	return *this;
}

bool
operator== (const Selection& a, const Selection& b)
{
	return a.regions == b.regions &&
		a.tracks == b.tracks &&
		a.time.track == b.time.track &&
		a.time.group == b.time.group && 
		a.time.audio == b.time.audio &&
		a.time.music == b.time.music &&
		a.lines == b.lines;
}

void
Selection::clear ()
{
	clear_tracks ();
	clear_regions ();
	clear_time ();
	clear_lines ();
}

void
Selection::dump_region_layers()
{
	cerr << "region selection layer dump" << endl;
	for (RegionSelection::iterator i = regions.begin(); i != regions.end(); ++i) {
		cerr << "layer: " << (int)(*i)->region.layer() << endl;
	}
}

void
Selection::clear_regions ()
{
	regions.clear ();
	RegionsChanged();
}

void
Selection::clear_tracks ()
{
	tracks.clear ();
	RoutesChanged();
}

void
Selection::clear_time ()
{
	time.track = 0;
	time.group = 0;
	time.audio.clear();
	time.music.clear();

	TimeChanged ();
}

void
Selection::clear_lines ()
{
	lines.clear ();
	LinesChanged();
}

void
Selection::add (TimeAxisView* track)
{
	void (Selection::*pmf)(TimeAxisView*) = &Selection::remove;
	track->GoingAway.connect (bind (slot (*this, pmf), track));
	tracks.push_back (track);
	RoutesChanged();
}

void
Selection::add (RegionView* r)
{
	RegionLayerComparator cmp;

	void (Selection::*pmf)(RegionView*) = &Selection::remove;
	r->GoingAway.connect (slot (*this, pmf));

	regions.push_back (r);
	regions.sort(cmp);

	RegionsChanged();
}

long
Selection::add (jack_nframes_t start, jack_nframes_t end)
{
	AudioRangeComparator cmp;

	time.audio.push_back (AudioRange (start, end, next_time_id++));
	time.audio.sort (cmp);
	
	TimeChanged ();

	return next_time_id - 1;
}

long
Selection::add (BBT_Time& start, BBT_Time& end)
{
	time.music.push_back (MusicRange (start, end, next_time_id++));
	/* XXX need sort */
	TimeChanged ();
	return time.music.back().id;
}

void
Selection::replace (unsigned long id, jack_nframes_t start, jack_nframes_t end)
{
	for (list<AudioRange>::iterator i = time.audio.begin(); i != time.audio.end(); ++i) {
		if ((*i).id == id) {
			time.audio.erase (i);
			time.audio.push_back (AudioRange(start,end, id));

			AudioRangeComparator cmp;
			time.audio.sort (cmp);

			TimeChanged ();
			break;
		}
	}
}

void
Selection::replace (unsigned long id, BBT_Time& start, BBT_Time& end)
{
	for (list<MusicRange>::iterator i = time.music.begin(); i != time.music.end(); ++i) {
		if ((*i).id == id) {
			time.music.erase (i);
			time.music.push_back (MusicRange(start,end, id));
			TimeChanged ();
			break;
		}
	}
}

void
Selection::add (AutomationLine* ac)
{
	lines.push_back (ac);
	LinesChanged();
}

void
Selection::remove (TimeAxisView* track)
{
	list<TimeAxisView*>::iterator i;
	if ((i = find (tracks.begin(), tracks.end(), track)) != tracks.end()) {
		tracks.erase (i);
		RoutesChanged();
	}
}

void
Selection::remove (RegionView* r)
{
	RegionSelection::iterator i;

	if ((i = find (regions.begin(), regions.end(), r)) != regions.end()) {
		regions.erase (i);
		RegionsChanged();
	}
}


void
Selection::remove (unsigned long selection_id)
{
	if (time.audio.empty()) {
		return;
	}

	for (list<AudioRange>::iterator i = time.audio.begin(); i != time.audio.end(); ++i) {
		if ((*i).id == selection_id) {
			time.audio.erase (i);
						
			TimeChanged ();
			break;
		}
	}
}

void
Selection::remove (jack_nframes_t start, jack_nframes_t end)
{
}

void
Selection::remove (BBT_Time& start, BBT_Time& end)
{
}

void
Selection::remove (AutomationLine *ac)
{
	list<AutomationLine*>::iterator i;
	if ((i = find (lines.begin(), lines.end(), ac)) != lines.end()) {
		lines.erase (i);
		LinesChanged();
	}
}

void
Selection::set (TimeAxisView* track)
{
	clear_tracks ();
	add (track);
}

void
Selection::set (RegionView* r)
{
	clear_regions ();
	add (r);
}

long
Selection::set (TimeAxisView* track, jack_nframes_t start, jack_nframes_t end)
{
	if ((start == 0 && end == 0) || end < start) {
		return 0;
	}

	if (time.audio.empty()) {
		time.audio.push_back (AudioRange (start, end, next_time_id++));
	} else {
		/* reuse the first entry, and remove all the rest */

		while (time.audio.size() > 1) {
			time.audio.pop_front();
		}
		time.audio.front().start = start;
		time.audio.front().end = end;
	}

	if (track) {
		time.track = track;
		time.group = track->edit_group();
	} else {
		time.track = 0;
		time.group = 0;
	}
	
	TimeChanged ();

	return time.audio.front().id;
}

long
Selection::set (TimeAxisView* track, BBT_Time& start, BBT_Time& end)
{
	/* XXX check for start < end etc. */

	if (time.music.empty()) {
		time.music.push_back (MusicRange (start, end, next_time_id++));
	} else {
		/* reuse the first entry, and remove all the rest */

		while (time.music.size() > 1) {
			time.music.pop_front();
		}
		time.music.front().start = start;
		time.music.front().end = end;
	}

	if (track) {
		time.track = track;
		time.group = track->edit_group();
	} else {
		time.track = 0;
		time.group = 0;
	}

	TimeChanged ();

	return time.music.front().id;
}

void
Selection::set (AutomationLine *ac)
{
	lines.clear();
	add (ac);
}

AudioRange&
TimeSelection::operator[] (unsigned long which)
{
	for (list<AudioRange>::iterator i = audio.begin(); i != audio.end(); ++i) {
		if ((*i).id == which) {
			return *i;
		}
	}
	fatal << compose (_("programming error: request for non-existent audio range (%1)!"), which) << endmsg;
	/*NOTREACHED*/
	return *(new AudioRange(0,0,0)); /* keep the compiler happy; never called */
}

jack_nframes_t 
TimeSelection::start ()
{
	if (audio.empty()) {
		return 0;
	}

	jack_nframes_t first = max_frames;

	for (list<AudioRange>::iterator i = audio.begin(); i != audio.end(); ++i) {
		if ((*i).start < first) {
			first = (*i).start;
		}
	}
	return first;
}

jack_nframes_t 
TimeSelection::end ()
{
	jack_nframes_t last = 0;

	for (list<AudioRange>::iterator i = audio.begin(); i != audio.end(); ++i) {
		if ((*i).end > last) {
			last = (*i).end;
		}
	}
	return last;
}

jack_nframes_t
TimeSelection::length()
{
	return end() - start() + 1;
}
	
