#include <gtk--.h>
#include <math.h>

#include <gtkmmext/gtk_ui.h>

#include <ardour/playlist.h>
#include <ardour/region.h>
#include <ardour/diskstream.h>

#include "streamview.h"
#include "regionview.h"
#include "audio_time_axis.h"
#include "canvas-waveview.h"
#include "canvas-simplerect.h"
#include "region_selection.h"
#include "editor.h"

#include <ardour/audio_track.h>

using namespace ARDOUR;

StreamView::StreamView (AudioTimeAxisView& tv)
	: _trackview (tv)
{
	region_color = _trackview.color();

	if (tv.is_audio_track()) {
		stream_base_color = RGBA_TO_UINT (222,223,218,255);
	} else {
		stream_base_color = RGBA_TO_UINT (230,226,238,255);
	}

	/* set_position() will position the group */
	
	canvas_group = gtk_canvas_item_new (GTK_CANVAS_GROUP(_trackview.canvas_display),
					    gtk_canvas_group_get_type (),
					    NULL);

	canvas_rect = gtk_canvas_item_new (GTK_CANVAS_GROUP(canvas_group),
					   gtk_canvas_simplerect_get_type(),
					   "x1", 0.0,
					   "y1", 0.0,
					   "x2", 1000000.0,
					   "y2", (double) tv.height,
					   "outline_color_rgba", RGBA_TO_UINT(0,0,0,255),
					   /* outline ends and bottom */
					   "outline_what", (guint32) (0x1|0x2|0x8),
					   "fill_color_rgba", stream_base_color,
					   NULL);

	gtk_signal_connect (GTK_OBJECT(canvas_rect), "event",
			    (GtkSignalFunc) Editor::_canvas_stream_view_event, &_trackview);

	_samples_per_unit = _trackview.editor.get_current_zoom();
	_amplitude_above_axis = 1.0;

	if (_trackview.is_audio_track()) {
		_trackview.audio_track()->diskstream_changed.connect (slot (this, &StreamView::diskstream_changed));
		display_diskstream (_trackview.get_diskstream());
	} 
}

StreamView::~StreamView ()
{
	undisplay_diskstream ();
	gtk_object_destroy (GTK_OBJECT(canvas_group));
}

int
StreamView::set_position (gdouble x, gdouble y)

{
	gtk_canvas_item_set (canvas_group, "x", x, "y", y, NULL);
	return 0;
}

int
StreamView::set_height (gdouble h)
{
	RegionViewList::iterator i;

	/* limit the values to something sane-ish */

	if (h < 10.0 || h > 1000.0) {
		return -1;
	}

	gtk_object_set (GTK_OBJECT(canvas_rect), "y2", h, NULL);

	for (i = region_views.begin(); i != region_views.end(); ++i) {
		(*i)->set_height (h - 2);
	}

	return 0;
}

int 
StreamView::set_samples_per_unit (gdouble spp)

{
	RegionViewList::iterator i;

	if (spp < 1.0) {
		return -1;
	}

	_samples_per_unit = spp;

	for (i = region_views.begin(); i != region_views.end(); ++i) {
		(*i)->set_samples_per_unit (spp);
	}

	return 0;
}

int 
StreamView::set_amplitude_above_axis (gdouble app)

{
	RegionViewList::iterator i;

	if (app < 1.0) {
		return -1;
	}

	_amplitude_above_axis = app;

	for (i = region_views.begin(); i != region_views.end(); ++i) {
		(*i)->set_amplitude_above_axis (app);
	}

	return 0;
}

void
StreamView::add_region_view (Region *region)
{
	if (!Gtkmmext::UI::instance()->caller_is_gui_thread()) {
		Gtkmmext::UI::instance()->call_slot (bind (slot (*this, &StreamView::add_region_view), region));
		return;
	}

	RegionView *region_view;
	list<RegionView *>::iterator i;

	for (i = region_views.begin(); i != region_views.end(); ++i) {
		if (&(*i)->region == region) {
			
			/* great. we already have a RegionView for this Region. use it again.
			 */

			(*i)->valid = true;
			return;
		}
	}

	region_view = new RegionView (GTK_CANVAS_GROUP(canvas_group),
				      _trackview,
				      *region,
				      _samples_per_unit,
				      _amplitude_above_axis, 
				      region_color);
	region_views.push_front (region_view);

	region->LayerChanged.connect (bind (slot (*this, &StreamView::region_layered), region_view));

	RegionViewAdded (region_view);
}

void
StreamView::remove_region_view (Region *region)
{
	if (!Gtkmmext::UI::instance()->caller_is_gui_thread()) {
		Gtkmmext::UI::instance()->call_slot (bind (slot (*this, &StreamView::remove_region_view), region));
		return;
	}

	for (list<RegionView *>::iterator i = region_views.begin(); i != region_views.end(); ++i) {
		if (&((*i)->region) == region) {
			delete *i;
			region_views.erase (i);
			break;
		}
	}
}

void
StreamView::undisplay_diskstream ()
{
	list<RegionView *>::iterator i;

	for (i = region_views.begin(); i != region_views.end(); ++i) {
		delete *i;
	}
	region_views.clear();
}

void
StreamView::display_diskstream (DiskStream *ds)
{
	playlist_change_connection.disconnect();
	playlist_changed (ds);
	playlist_change_connection = ds->PlaylistChanged.connect (bind (slot (*this, &StreamView::playlist_changed), ds));
}

void
StreamView::playlist_changed (DiskStream *ds)
{
	/* disconnect from old playlist */

	for (vector<SigC::Connection>::iterator i = playlist_connections.begin(); i != playlist_connections.end(); ++i) {
		(*i).disconnect();
	}
	
	playlist_connections.clear();
	undisplay_diskstream ();

	/* draw it */

	ds->playlist()->foreach_region (this, &StreamView::add_region_view);
	ds->playlist()->foreach_overlap (this, &StreamView::add_overlap);

	/* catch changes */

	playlist_connections.push_back (ds->playlist()->RegionAdded.connect (slot (*this, &StreamView::add_region_view)));
	playlist_connections.push_back (ds->playlist()->RegionRemoved.connect (slot (*this, &StreamView::remove_region_view)));
	playlist_connections.push_back (ds->playlist()->StateChanged.connect (slot (*this, &StreamView::redisplay_diskstream)));
	playlist_connections.push_back (ds->playlist()->NewRegionOverlap.connect (slot (*this, &StreamView::add_overlap)));
}

void
StreamView::add_overlap (Overlap *overlap)
{
	_trackview.editor.maybe_crossfade (overlap);
	overlap->Changed.connect (bind (slot (*this, &StreamView::update_overlap), overlap));
}

void
StreamView::update_overlap (Overlap *overlap)
{
	_trackview.editor.maybe_crossfade (overlap);
}

void
StreamView::redisplay_diskstream ()
{
	list<RegionView *>::iterator i, tmp;

	for (i = region_views.begin(); i != region_views.end(); ++i) {
		(*i)->valid = false;
	}

	if (_trackview.is_audio_track()) {
		_trackview.get_diskstream()->playlist()->foreach_region (this, &StreamView::add_region_view);
	}

	for (i = region_views.begin(); i != region_views.end(); ) {
		tmp = i;
		tmp++;

		if (!(*i)->valid) {
			delete *i;
			region_views.erase (i);
		}

		i = tmp;
	}
}

void
StreamView::diskstream_changed (void *src_ignored)
{
	AudioTrack *at;

	if ((at = _trackview.audio_track()) != 0) {
		DiskStream& ds = at->disk_stream();
		/* XXX grrr: when will SigC++ allow me to bind references? */
		Gtkmmext::UI::instance()->call_slot (bind (slot (*this, &StreamView::display_diskstream), &ds));
	} else {
		Gtkmmext::UI::instance()->call_slot (slot (*this, &StreamView::undisplay_diskstream));
	}
}

void
StreamView::apply_color (GdkColor& color, ColorTarget target)

{
	list<RegionView *>::iterator i;

	switch (target) {
	case RegionColor:
		region_color = color;
		for (i = region_views.begin(); i != region_views.end(); ++i) {
			(*i)->set_color (region_color);
		}
		// stream_base_color = RGBA_TO_UINT (color.red/256, color.green/256, color.blue/256, 255);
		// gtk_canvas_item_set (canvas_rect, "fill_color_rgba", stream_base_color, NULL);
		break;
		
	case StreamBaseColor:
		// stream_base_color = RGBA_TO_UINT (color.red/256, color.green/256, color.blue/256, 255);
		// gtk_canvas_item_set (canvas_rect, "fill_color_rgba", stream_base_color, NULL);
		break;
	}
}

void
StreamView::set_show_waveforms (bool yn)
{
	for (list<RegionView *>::iterator i = region_views.begin(); i != region_views.end(); ++i) {
			(*i)->set_waveform_visible (yn);
	}
}

void
StreamView::set_selected_regionviews (RegionSelection& regions)
{
	bool selected;
	
	for (list<RegionView*>::iterator i = region_views.begin(); i != region_views.end(); ++i) {
		
		selected = false;
		
		for (RegionSelection::iterator ii = regions.begin(); ii != regions.end(); ++ii) {
			if (*i == *ii) {
				selected = true;
			}
		}
		
		(*i)->set_selected (selected);
	}
}

void
StreamView::set_waveform_shape (WaveformShape shape)
{
	for (RegionViewList::iterator i = region_views.begin(); i != region_views.end(); ++i) {
		(*i)->set_waveform_shape (shape);
	}
}		
		
void
StreamView::region_layered (RegionView* rv)
{
	gtk_canvas_item_lower_to_bottom (rv->group);

	/* don't ever leave it at the bottom, since then it doesn't
	   get events - the parent group does instead ...
	*/

	gtk_canvas_item_raise (rv->group, rv->region.layer() + 1);
}
