/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: import.cc,v 1.19 2003/06/18 17:50:56 pbd Exp $
*/

#include <cstdio>
#include <string>
#include <climits>
#include <cerrno>
#include <unistd.h>
#include <sys/stat.h>
#include <time.h>

#include <sndfile.h>

#include <ardour/ardour.h>
#include <ardour/session.h>
#include <ardour/diskstream.h>
#include <ardour/filesource.h>
#include <ardour/sndfile_helpers.h>

#include "i18n.h"

using namespace ARDOUR;

#define BLOCKSIZE 2048U

int
Session::import_audiofile (import_status& status)
{
	SNDFILE *in;
	FileSource **newfiles;
	ARDOUR::Region::SourceList sources;
	SF_INFO info;
	double *data;
	Sample **channel_data;
	long nfiles;
	long n;
	string basepath;
	string sounds_dir;
	string::size_type pos;
	jack_nframes_t so_far;
	char buf[PATH_MAX+1];
	int ret = -1;
	vector<Region *> new_regions;
	vector<string> new_paths;
	struct tm* now;
	
	if ((in = sf_open (status.pathname.c_str(), SFM_READ, &info)) == 0) {
		error << compose(_("Import: cannot open input sound file \"%1\""), status.pathname) << endmsg;
		return -1;
	}

	nfiles = 0;
	channel_data = 0;
	data = 0;

	newfiles = new FileSource *[info.channels];
	for (n = 0; n < info.channels; ++n) {
		newfiles[n] = 0;
	}
	
	sounds_dir = discover_best_sound_dir ();

	if ((pos = status.pathname.find_last_of ('/')) != string::npos) {
		basepath = status.pathname.substr (pos+1);
	} else {
		basepath = status.pathname;
	}

	if ((pos = basepath.find_last_of ('.')) != string::npos) {
		basepath = basepath.substr (0, pos);
	}

	for (n = 0; n < info.channels; ++n) {
		bool goodfile = false;
		do {
			if (info.channels == 2) {
				if (n == 0) {
					snprintf (buf, sizeof(buf), "%s%s-L.wav", sounds_dir.c_str(), basepath.c_str());
				} else {
					snprintf (buf, sizeof(buf), "%s%s-R.wav", sounds_dir.c_str(), basepath.c_str());
				}
			} else if (info.channels > 1) {
				snprintf (buf, sizeof(buf), "%s%s-c%lu.wav", sounds_dir.c_str(), basepath.c_str(), n+1);
			} else {
				snprintf (buf, sizeof(buf), "%s%s.wav", sounds_dir.c_str(), basepath.c_str());
			}

			if (::access (buf, F_OK) == 0) {
				/* if the file already exists, we must come up with
				 *  a new name for it.  for now we just keep appending
				 *  _ to basepath
				 */
				
				basepath += "_";
			}
			else {
				goodfile = true;
			}
		} while ( !goodfile);

			
		try { 
			newfiles[n] = new FileSource (buf, frame_rate());
		}

		catch (failed_constructor& err) {
			error << compose(_("Session::import_audiofile: cannot open new file source for channel %1"), n+1) << endmsg;
			goto out;
		}

		new_paths.push_back (buf);
		nfiles++;
	}
	
	
	data = new double[BLOCKSIZE * info.channels];
	channel_data = new Sample * [ info.channels ];

	for (n = 0; n < info.channels; ++n) {
		channel_data[n] = new Sample[BLOCKSIZE];
	}

	so_far = 0;

	status.doing_what = "converting audio";

	while (1 && !status.cancel) {

		long nread;
		long x;
		long chn;

		if ((nread = sf_readf_double (in, data, BLOCKSIZE)) == 0) {
			break;
		}

		/* de-interleave */
				
		for (chn = 0; chn < info.channels; ++chn) {
			for (x = chn, n = 0; n < nread; x += info.channels, ++n) {
				channel_data[chn][n] = (Sample) data[x];
			}
		}

		/* flush to disk */

		for (chn = 0; chn < info.channels; ++chn) {
			newfiles[chn]->write (channel_data[chn], nread);
		}

		so_far += nread;
		status.percent_done = so_far / (float) (info.frames * info.channels);
	}

	if (status.multichan) {
		status.doing_what = "building region";
	} else {
		status.doing_what = "building regions";
	}

	time_t xnow;
	time (&xnow);
	now = localtime (&xnow);

	if (status.multichan) {
		/* all sources are used in a single multichannel region */
		for (n = 0; n < nfiles && !status.cancel; ++n) {
			/* flush the final length to the header */
			newfiles[n]->update_header(0, *now, xnow);
			sources.push_back(newfiles[n]);
		}

		Region *r = new Region (sources, 0, newfiles[0]->length(), region_name_from_path (basepath),
					0, Region::Flag (Region::DefaultFlags | Region::WholeFile));
		
		new_regions.push_back (r);

	}
	else {
		for (n = 0; n < nfiles && !status.cancel; ++n) {

			/* flush the final length to the header */

			newfiles[n]->update_header(0, *now, xnow);

			/* The sources had zero-length when created, which means that the Session
			   did not bother to create whole-file Regions for them. Do it now.
			*/
		
			Region *r = new Region (*newfiles[n], 0, newfiles[n]->length(), region_name_from_path (newfiles[n]->name()),
						0, Region::Flag (Region::DefaultFlags | Region::WholeFile));

			new_regions.push_back (r);
		}
	}
	
	/* save state so that we don't lose these new Sources */

	if (!status.cancel) {
		save_state (N_("default"));
	}

	ret = 0;

  out:

	if (data) {
		delete [] data;
	}

	if (channel_data) {
		for (n = 0; n < info.channels; ++n) {
			delete [] channel_data[n];
		}
		delete [] channel_data;
	}

	if (status.cancel) {
		for (vector<Region *>::iterator i = new_regions.begin(); i != new_regions.end(); ++i) {
			delete *i;
		}

		for (vector<string>::iterator i = new_paths.begin(); i != new_paths.end(); ++i) {
			unlink ((*i).c_str());
		}
	}


	if (newfiles) {
		for (n = 0; n < nfiles; ++n) {
			if (newfiles[n]) newfiles[n]->unref();
		}
		delete [] newfiles;
	}

	sf_close (in);
	status.done = true;
	return ret;
}
