/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: io.cc,v 1.62 2003/06/14 04:22:47 pbd Exp $
*/

#include <algorithm>
#include <unistd.h>

#include <sigc++/bind.h>

#include <pbd/lockmonitor.h>
#include <pbd/xml++.h>

#include <ardour/audioengine.h>
#include <ardour/io.h>
#include <ardour/port.h>
#include <ardour/connection.h>
#include <ardour/session.h>
#include <ardour/cycle_timer.h>

#include "i18n.h"

using namespace std;
using namespace ARDOUR;
using namespace SigC;

const string IO::state_node_name = "IO";
bool         IO::connecting_legal = false;
bool         IO::ports_legal = false;
Signal0<void>                IO::GrabPeakPower;
Signal0<int>                 IO::ConnectingLegal;
Signal0<int>                 IO::PortsLegal;
Signal1<void,unsigned long>  IO::MoreOutputs;
Signal0<int>                 IO::PortsCreated;

/* this is a default mapper of MIDI control values to a gain coefficient.
   others can be imagined. see IO::set_midi_to_gain_function().
*/

static gain_t direct_midi_to_gain (double fract) { 
	/* XXX Marcus writes: this doesn't seem right to me. but i don't have a better answer ... */
	return pow (2.0,(sqrt(sqrt(sqrt(fract)))*198.0-192.0)/6.0);
}

IO::IO (Session& s, const string& name,
	int input_min, int input_max, int output_min, int output_max)
	: _session (s),
	  _name (name),
	  _midi_gain_control (*this, (MIDI::Port *) 0),
	  _midi_pan_control (*this, (MIDI::Port *) 0),
	  _input_minimum (input_min),
	  _input_maximum (input_max),
	  _output_minimum (output_min),
	  _output_maximum (output_max)
{
	_id = new_id();
	_gain = 1.0;
	_desired_gain = 1.0;
	_input_connection = 0;
	_output_connection = 0;
	pending_state_node = 0;
	_ninputs = 0;
	_noutputs = 0;

	_midi_gain_control.midi_to_gain = direct_midi_to_gain;
	
	GrabPeakPower.connect (slot (*this, &IO::grab_peak_power));
}

IO::~IO ()
{
	LockMonitor lm (io_lock, __LINE__, __FILE__);
	vector<Port *>::iterator i;

	for (i = _inputs.begin(); i != _inputs.end(); ++i) {
		_session.engine().unregister_port (*i);
	}

	for (i = _outputs.begin(); i != _outputs.end(); ++i) {
		_session.engine().unregister_port (*i);
	}
}

void
IO::silence (jack_nframes_t nframes, jack_nframes_t offset)
{
	/* io_lock, not taken: function must be called from Session::process() calltree */

	for (vector<Port *>::iterator i = _outputs.begin(); i != _outputs.end(); ++i) {
		(*i)->silence (nframes, offset);
	}
}

void
IO::apply_declick (vector<Sample *>& bufs, jack_nframes_t nframes, gain_t initial, gain_t target, bool invert_polarity)
{
	jack_nframes_t declick = min (64UL, nframes);
	gain_t delta;
	Sample *buffer;
	double fractional_shift;
	double fractional_pos;

	fractional_shift = -1.0/declick;

	if (target < initial) {
		/* fade out: remove more and more of delta from initial */
		delta = -(initial - target);
	} else {
		/* fade in: add more and more of delta from initial */
		delta = target - initial;
	}

	for (vector<Sample*>::iterator i = bufs.begin(); i != bufs.end(); ++i) {

		buffer = *i;
		fractional_pos = 1.0;

		if (invert_polarity) {
			for (jack_nframes_t nx = 0; nx < declick; ++nx) {
				buffer[nx] *= -(initial + (delta * (0.5 + 0.5 * cos (M_PI * fractional_pos))));
				fractional_pos += fractional_shift;
			}
		} else {
			for (jack_nframes_t nx = 0; nx < declick; ++nx) {
				buffer[nx] *= (initial + (delta * (0.5 + 0.5 * cos (M_PI * fractional_pos))));
				fractional_pos += fractional_shift;
			}
		}
		
		/* now ensure the rest of the buffer has the target value
		   applied, if necessary.
		*/
		
		if (declick != nframes) {
			if (target == 0.0) {
				memset (&buffer[declick], 0, sizeof (Sample) * (nframes - declick));
			} else if (target != 1.0) {
				for (jack_nframes_t nx = declick; nx < nframes; ++nx) {
					buffer[nx] *= target;
				}
			}
		}
	}
}

void
IO::pan (vector<Sample*>& bufs, jack_nframes_t nframes, jack_nframes_t offset, pan_t* panbuffer)
{
	/* io_lock, not taken: function must be called from Session::process() calltree */

	unsigned long i;
	Sample* dst;
	Sample* src;
	vector<Port *>::iterator o;
	pan_t pan;

	for (o = _outputs.begin(), i = 0; o != _outputs.end(); ++i, ++o) {
		
		src = bufs[i];
		dst = (*o)->get_buffer (nframes) + offset;

		if (panbuffer) {

			/* XXX stereo pan ONLY */

			if (i == 0) {
				for (jack_nframes_t n = 0; n < nframes; ++n) {
					dst[n] = src[n] * panbuffer[n];
				}	
			} else {
				for (jack_nframes_t n = 0; n < nframes; ++n) {
					dst[n] = src[n] * (1.0 - panbuffer[n]);
				}	
			}

		} else {

			if ((pan = _pans[i]) != 1.0) {
				
				if (pan == 0.0) {
					
					memset (dst, 0, sizeof (Sample) * nframes);
					
				} else {
					
					for (jack_nframes_t n = 0; n < nframes; ++n) {
						dst[n] = src[n] * pan;
					}	
				}
				
			} else {

				if (dst != src) {
					memcpy (dst, src, sizeof (Sample) * nframes);
				}
				
			}
		}

		(*o)->mark_silence (false);
	}
}

void
IO::deliver_output (vector<Sample *>& bufs, jack_nframes_t nframes, jack_nframes_t offset)
{
	/* io_lock, not taken: function must be called from Session::process() calltree */

	Sample *dst;
	Sample *src;
	vector<Port*>::iterator d;
	vector<Sample*>::iterator s;
	vector<pan_t>::iterator p;
	float coeff;
	gain_t dg;

	{
		TentativeLockMonitor dm (declick_lock, __LINE__, __FILE__);
		
		if (dm.locked()) {
			dg = _desired_gain;
		} else {
			dg = _gain;
		}
	}

	if (dg != _gain) {
		apply_declick (bufs, nframes, _gain, dg, false);
		_gain = dg;
	} 

	/* note that _pans is *always* as long as _outputs, and that we require bufs.size() >= 1 */
	
	for (d = _outputs.begin(), s = bufs.begin(), p = _pans.begin(); d != _outputs.end(); ++d, ++s, ++p) {

		if (s == bufs.end()) {
			break;
		}

		dst = (*d)->get_buffer (nframes) + offset;
		src = *s;

		if ((coeff = (*p) * _gain) == 1.0) {
			memcpy (dst, src, sizeof (Sample) * nframes);
		} else {
			for (jack_nframes_t n = 0; n < nframes; ++n) {
				dst[n] = src[n] * coeff;
			}
		}

		(*d)->mark_silence (false);
	}

	/* if we didn't fill all outputs, do so now using the last input */

	if (d != _outputs.end()) {

		--s;
		--p;
		src = *s;

		while (d != _outputs.end()) {

			dst = (*d)->get_buffer (nframes) + offset;
			
			if ((coeff = (*p) * _gain) == 1.0) {
				memcpy (dst, src, sizeof (Sample) * nframes);
			} else {
				for (jack_nframes_t n = 0; n < nframes; ++n) {
					*dst++ = *src++ * coeff;
				}
			}

			(*d)->mark_silence (false);
			++d;
		}
	}
}

void
IO::collect_input (vector<Sample *>& bufs, jack_nframes_t nframes, jack_nframes_t offset)
{
	/* io_lock, not taken: function must be called from Session::process() calltree */

	vector<Port *>::iterator i;
	vector<Sample *>::iterator b;
	Sample *last = 0;
	
	/* we require that bufs.size() >= 1 */

	for (b = bufs.begin(), i = _inputs.begin(); b != bufs.end(); ++i, ++b) {
		if (i == _inputs.end()) {
			break;
		}
		last = (*i)->get_buffer (nframes) + offset;
		memcpy (*b + offset, last, sizeof (Sample) * nframes);
	}

	/* fill any excess outputs with the last input */
	
	while (b != bufs.end() && last) {
		memcpy (*b + offset, last, sizeof (Sample) * nframes);
		++b;
	}
}

void
IO::drop_input_connection ()
{
	_input_connection = 0;
	input_connection_configuration_connection.disconnect();
	input_connection_connection_connection.disconnect();
}

void
IO::drop_output_connection ()
{
	_output_connection = 0;
	output_connection_configuration_connection.disconnect();
	output_connection_connection_connection.disconnect();
}

int
IO::disconnect_input (Port* our_port, string other_port, void* src)
{
	if (other_port.length() == 0 || our_port == 0) {
		return 0;
	}

	{ 
		LockMonitor em (_session.engine().process_lock(), __LINE__, __FILE__);
		
		{
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			/* check that our_port is really one of ours */
			
			if (find (_inputs.begin(), _inputs.end(), our_port) == _inputs.end()) {
				return -1;
			}
			
			/* disconnect it from the source */
			
			if (_session.engine().disconnect (other_port, our_port->name())) {
				error << compose(_("IO: cannot disconnect input port %1 from %2"), our_port->name(), other_port) << endmsg;
				return -1;
			}

			drop_input_connection();
		}
	}

	EMIT_SIGNAL input_changed (src);

	return 0;
}

int
IO::connect_input (Port* our_port, string other_port, void* src)
{
	if (other_port.length() == 0 || our_port == 0) {
		return 0;
	}

	{
		LockMonitor em(_session.engine().process_lock(), __LINE__, __FILE__);
		
		{
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			/* check that our_port is really one of ours */
			
			if (find (_inputs.begin(), _inputs.end(), our_port) == _inputs.end()) {
				return -1;
			}
			
			/* connect it to the source */
			
			if (_session.engine().connect (other_port, our_port->name())) {
				error << compose(_("IO: cannot connect input port %1 to %2"), our_port->name(), other_port) << endmsg;
				return -1;
			}
			
			drop_input_connection ();
		}
	}

	EMIT_SIGNAL input_changed (src);

	return 0;
}

int
IO::disconnect_output (Port* our_port, string other_port, void* src)
{
	if (other_port.length() == 0 || our_port == 0) {
		return 0;
	}

	{
		LockMonitor em(_session.engine().process_lock(), __LINE__, __FILE__);
		
		{
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			if (find (_outputs.begin(), _outputs.end(), our_port) == _outputs.end()) {
				return -1;
			}
			
			/* disconnect it from the destination */
			
			if (_session.engine().disconnect (our_port->name(), other_port)) {
				error << compose(_("IO: cannot disconnect output port %1 from %2"), our_port->name(), other_port) << endmsg;
				return -1;
			}

			drop_output_connection ();
		}
	}

	EMIT_SIGNAL output_changed (src);
	return 0;
}

int
IO::connect_output (Port* our_port, string other_port, void* src)
{
	if (other_port.length() == 0 || our_port == 0) {
		return 0;
	}

	{
		LockMonitor em(_session.engine().process_lock(), __LINE__, __FILE__);
		
		{
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			/* check that our_port is really one of ours */
			
			if (find (_outputs.begin(), _outputs.end(), our_port) == _outputs.end()) {
				return -1;
			}
			
			/* connect it to the destination */
			
			if (_session.engine().connect (our_port->name(), other_port)) {
				error << compose(_("IO: cannot connect output port %1 to %2"), our_port->name(), other_port) << endmsg;
				return -1;
			}

			drop_output_connection ();
		}
	}

	EMIT_SIGNAL output_changed (src);
	return 0;
}

int
IO::set_input (Port* other_port, void* src)
{
	/* this removes all but one ports, and connects that one port
	   to the specified source.
	*/

	if (_input_minimum > 1 || _input_minimum == 0) {
		/* sorry, you can't do this */
		return -1;
	}

	if (other_port == 0) {
		if (_input_minimum < 0) {
			return ensure_inputs (0, false, src);
		} else {
			return -1;
		}
	}

	if (ensure_inputs (1, true, src)) {
		return -1;
	}

	return connect_input (_inputs.front(), other_port->name(), src);
}

int
IO::remove_output_port (Port* port, void* src)
{
	bool send_change = false;
	bool send_config = false;

	{
		LockMonitor em(_session.engine().process_lock(), __LINE__, __FILE__);
		
		{
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			if (_noutputs - 1 == (unsigned long) _output_minimum) {
				/* sorry, you can't do this */
				return -1;
			}
			
			for (vector<Port *>::iterator i = _outputs.begin(); i != _outputs.end(); ++i) {
				if (*i == port) {
					if (port->connected()) {
						send_change = true;
					} 
					send_config  = true;
					_session.engine().unregister_port (*i);
					_outputs.erase (i);
					_noutputs--;
					drop_output_connection ();
					setup_peak_meters ();
					break;
				}
			}
			reset_pans (src);

		}
	}

	if (send_change && !send_config) {
		EMIT_SIGNAL output_changed (src);
	}

	if (send_config) {
		EMIT_SIGNAL output_configuration_changed (src);
		return 0;
	}
	return -1;
}

int
IO::add_output_port (string destination, void* src)
{
	Port* our_port;
	char buf[64];
	unsigned long old;

	{
		LockMonitor em(_session.engine().process_lock(), __LINE__, __FILE__);
		
		{ 
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			if (_output_maximum >= 0 && (int) _noutputs == _output_maximum) {
				return -1;
			}
			old = _noutputs;
		}
		
		/* Create a new output port */
		
		if (_output_maximum == 1) {
			snprintf (buf, sizeof (buf), "%s/out", _name.c_str());
		} else {
			snprintf (buf, sizeof (buf), "%s/out %lu", _name.c_str(), old + 1);
		}

		if ((our_port = _session.engine().register_audio_output_port (buf)) == 0) {
			error << compose(_("IO: cannot register output port %1"), buf) << endmsg;
			return -1;
		}
		
		{ 
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			_outputs.push_back (our_port);
			++_noutputs;
			drop_output_connection ();
			setup_peak_meters ();
		}

		EMIT_SIGNAL MoreOutputs (_noutputs);

		reset_pans (src);
	}

	if (destination.length()) {
		if (_session.engine().connect (our_port->name(), destination)) {
			error << compose(_("IO: cannot connect output port %1 to %2"), our_port->name(), destination) << endmsg;
			return -1;
		}
	}

	EMIT_SIGNAL output_configuration_changed (src);

	return 0;
}

int
IO::remove_input_port (Port* port, void* src)
{
	bool send_change = false;
	bool send_config = false;

	{
		LockMonitor em(_session.engine().process_lock(), __LINE__, __FILE__);
		
		{
			LockMonitor lm (io_lock, __LINE__, __FILE__);

			if (((int)_ninputs - 1) < _input_minimum) {
				/* sorry, you can't do this */
				return -1;
			}
			for (vector<Port *>::iterator i = _inputs.begin(); i != _inputs.end(); ++i) {

				if (*i == port) {
					if (port->connected()) {
						send_change = true;
					} 
					send_config  = true;
					_session.engine().unregister_port (*i);
					_inputs.erase (i);
					_ninputs--;
					drop_input_connection ();
					setup_peak_meters ();
					break;
				}
			}
			reset_pans (src);

		}
	}

	if (send_change && !send_config) {
		EMIT_SIGNAL input_changed (src);
	}

	if (send_config) {
		EMIT_SIGNAL input_configuration_changed (src);
		return 0;
	} 

	return -1;
}

int
IO::add_input_port (string source, void* src)
{
	Port* our_port;
	char buf[64];
	unsigned long old;

	{
		LockMonitor em (_session.engine().process_lock(), __LINE__, __FILE__);
		
		{ 
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			if (_input_maximum >= 0 && (int) _ninputs == _input_maximum) {
				return -1;
			}
			old = _ninputs;
		}
		
		/* Create a new input port */
		
		if (_input_maximum == 1) {
			snprintf (buf, sizeof (buf), "%s/in", _name.c_str());
		} else {
			snprintf (buf, sizeof (buf), "%s/in %lu", _name.c_str(), old + 1);
		}

		if ((our_port = _session.engine().register_audio_input_port (buf)) == 0) {
			error << compose(_("IO: cannot register input port %1"), buf) << endmsg;
			return -1;
		}
		
		{ 
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			_inputs.push_back (our_port);
			++_ninputs;
			drop_input_connection ();
			setup_peak_meters ();
		}

		EMIT_SIGNAL MoreOutputs (_ninputs);
	}

	if (source.length()) {

		if (_session.engine().connect (source, our_port->name())) {
			error << compose(_("IO: cannot connect input port %1 to %2"), our_port->name(), source) << endmsg;
			return -1;
		}
	} 

	EMIT_SIGNAL input_configuration_changed (src);

	return 0;
}

int
IO::disconnect_inputs (void* src)
{
	{ 
		LockMonitor em (_session.engine().process_lock(), __LINE__, __FILE__);
		
		{
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			for (vector<Port *>::iterator i = _inputs.begin(); i != _inputs.end(); ++i) {
				_session.engine().disconnect (*i);
			}

			drop_input_connection ();
		}
	}
	EMIT_SIGNAL input_changed (src);
	return 0;
}

int
IO::disconnect_outputs (void* src)
{
	{
		LockMonitor em (_session.engine().process_lock(), __LINE__, __FILE__);
		
		{
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			for (vector<Port *>::iterator i = _outputs.begin(); i != _outputs.end(); ++i) {
				_session.engine().disconnect (*i);
			}

			drop_output_connection ();
		}
	}

	EMIT_SIGNAL output_changed (src);
	return 0;
}

int
IO::ensure_inputs (unsigned int n, bool clear, void* src)
{
	Port* input_port;
	bool changed = false;

	if (_input_maximum >= 0) {
		n = min (_input_maximum, (int) n);
		
		if (n == _ninputs && !clear) {
			return 0;
		}
	}
	
	/* XXX caller should hold io_lock, but generally doesn't */

	{
		LockMonitor em (_session.engine().process_lock(), __LINE__, __FILE__);
		
		/* remove unused ports */
		
		while (_ninputs > n) {
			_session.engine().unregister_port (_inputs.back());
			_inputs.pop_back();
			_ninputs--;
			changed = true;
		}
		
		/* create any necessary new ports */
		
		while (_ninputs < n) {
			
			char buf[64];
			
			/* Create a new input port */
			
			if (_input_maximum == 1) {
				snprintf (buf, sizeof (buf), "%s/in", _name.c_str());
			}
			else {
				snprintf (buf, sizeof (buf), "%s/in %lu", _name.c_str(), _ninputs + 1);
			}
			
			if ((input_port = _session.engine().register_audio_input_port (buf)) == 0) {
				error << compose(_("IO: cannot register input port %1"), buf) << endmsg;
				return -1;
			}
			
			_inputs.push_back (input_port);
			++_ninputs;
			changed = true;
		}

		if (changed) {
			drop_input_connection ();
			setup_peak_meters ();
			EMIT_SIGNAL MoreOutputs (_ninputs);
		}
		
		if (clear) {
			/* disconnect all existing ports so that we get a fresh start */
			
			for (vector<Port *>::iterator i = _inputs.begin(); i != _inputs.end(); ++i) {
				_session.engine().disconnect (*i);
			}
		}
	}

	if (changed) {
		EMIT_SIGNAL input_configuration_changed (src);
	}

	return 0;
}

int
IO::ensure_outputs (unsigned int n, bool clear, void* src)
{
	Port* output_port;
	bool need_pan_reset;
	bool changed = false;

	if (_output_maximum >= 0) {
		n = min (_output_maximum, (int) n);
		if (n == _noutputs && !clear) {
			return 0;
		}
	}

	/* XXX caller should hold io_lock, but generally doesn't */

	{
		LockMonitor em (_session.engine().process_lock(), __LINE__, __FILE__);
		
		if (_noutputs == n) {
			need_pan_reset = false;
		} else {
			need_pan_reset = true;
		}
		
		/* remove unused ports */
		
		while (_noutputs > n) {
			
			_session.engine().unregister_port (_outputs.back());
			_outputs.pop_back();
			_noutputs--;
			changed = true;
		}
		
		/* create any necessary new ports */
		
		while (_noutputs < n) {
			
			char buf[64];
			
			/* Create a new output port */
			
			if (_output_maximum == 1) {
				snprintf (buf, sizeof (buf), "%s/out", _name.c_str());
			} else {
				snprintf (buf, sizeof (buf), "%s/out %lu", _name.c_str(), _noutputs + 1);
			}
			
			if ((output_port = _session.engine().register_audio_output_port (buf)) == 0) {
				error << compose(_("IO: cannot register output port %1"), buf) << endmsg;
				return -1;
			}
			
			_outputs.push_back (output_port);
			++_noutputs;
			changed = true;
		}

		if (changed) {
			drop_output_connection ();
			setup_peak_meters ();
			EMIT_SIGNAL MoreOutputs (_noutputs);
		}
		
		if (need_pan_reset) {
			reset_pans (this);
		}
		
		if (clear) {
			/* disconnect all existing ports so that we get a fresh start */
			
			for (vector<Port *>::iterator i = _outputs.begin(); i != _outputs.end(); ++i) {
				_session.engine().disconnect (*i);
			}
		}
	}

	if (changed) {
		EMIT_SIGNAL output_configuration_changed (src);
	}

	return 0;
}

pan_t
IO::effective_stereo_pan() const 
{
	if (_noutputs < 2) {
		return 0.0;
	} else {
		return _pans.front();
	}
}
		

void
IO::reset_pans (void* src)

{
	/* caller must hold io_lock */

	pan_t pan = 1.0 / _noutputs;

	/* default no panning if ninputs > 1 */
	if (_ninputs > 1) {
		pan = 1.0;
	}
	
	_pans.clear ();

	for (unsigned int i = 0; i < _noutputs; ++i) {
		_pans.push_back (pan);
	}

	EMIT_SIGNAL pan_changed (src);
}

XMLNode&
IO::get_state (void)
{
	XMLNode* node = new XMLNode (state_node_name);
	char buf[32];
	string str;
	bool need_ins = true;
	bool need_outs = true;
	LockMonitor lm (io_lock, __LINE__, __FILE__);

	node->add_property("name", _name);	
	snprintf (buf, sizeof(buf), "%Lu", id());
	node->add_property("id", buf);

	str = "";

	if (_input_connection) {
		node->add_property ("input-connection", _input_connection->name());
		need_ins = false;
	}

	if (_output_connection) {
		node->add_property ("output-connection", _output_connection->name());
		need_outs = false;
	}

	if (need_ins) {
		for (vector<Port *>::iterator i = _inputs.begin(); i != _inputs.end(); ++i) {
			
			const char **connections = (*i)->get_connections();
			
			if (connections && connections[0]) {
				str += '{';
				
				for (int n = 0; connections && connections[n]; ++n) {
					if (n) {
						str += ',';
					}
					str += connections[n];
				}
				str += '}';
				
				free (connections);
			}
			else {
				str += "{}";
			}
		}
		
		node->add_property ("inputs", str);
	}

	if (need_outs) {
		str = "";
		
		for (vector<Port *>::iterator i = _outputs.begin(); i != _outputs.end(); ++i) {
			
			const char **connections = (*i)->get_connections();
			
			if (connections && connections[0]) {
				
				str += '{';
				
				for (int n = 0; connections[n]; ++n) {
					if (n) {
						str += ',';
					}
					str += connections[n];
				}
				str += '}';
				
				free (connections);
			}
			else {
				str += "{}";
			}
		}
		
		node->add_property ("outputs", str);
	}

	str = "";

	for (unsigned long n = 0; n < n_outputs(); ++n) {
		if (n) str += ',';
		snprintf (buf, sizeof(buf), "%g", _pans[n]);
		str += buf;
	}

	node->add_property ("pans", str);

	snprintf (buf, sizeof(buf), "%g", gain());
	node->add_property ("gain", buf);

	snprintf (buf, sizeof(buf)-1, "%d,%d,%d,%d",
		  _input_minimum,
		  _input_maximum,
		  _output_minimum,
		  _output_maximum);
	node->add_property ("iolimits", buf);

	/* MIDI control */

	MIDI::channel_t chn;
	MIDI::eventType ev;
	MIDI::byte      additional;
	XMLNode*        midi_node = 0;
	XMLNode*        child;

	if (_midi_gain_control.get_control_info (chn, ev, additional)) {

		midi_node = node->add_child ("MIDI");
		child = midi_node->add_child ("gain");
		
		snprintf (buf, sizeof(buf), "0x%x", ev);
		child->add_property ("event", buf);
		snprintf (buf, sizeof(buf), "%d", chn);
		child->add_property ("channel", buf);
		snprintf (buf, sizeof(buf), "0x%x", additional);
		child->add_property ("additional", buf);
	}

	/* XXX same for pan etc. */

	return *node;
}

int
IO::connecting_became_legal ()
{
	int ret;

	if (pending_state_node == 0) {
		fatal << _("IO::connecting_became_legal() called without a pending state node") << endmsg;
		/*NOTREACHED*/
		return -1;
	}

	connection_legal_c.disconnect ();

	ret = make_connections (*pending_state_node);
	delete pending_state_node;
	pending_state_node = 0;
	return ret;
}

int
IO::ports_became_legal ()
{
	if (pending_state_node == 0) {
		fatal << _("IO::ports_became_legal() called without a pending state node") << endmsg;
		/*NOTREACHED*/
		return -1;
	}

	port_legal_c.disconnect ();

	return create_ports (*pending_state_node);
}

int
IO::set_state (const XMLNode& node)
{
	const XMLProperty* prop;

	if (node.name() != state_node_name) {
		error << compose(_("incorrect XML node \"%1\" passed to IO object"), node.name()) << endmsg;
		return -1;
	}

	if ((prop = node.property ("name")) != 0) {
		_name = prop->value();
	} 

	if ((prop = node.property ("id")) != 0) {
		sscanf (prop->value().c_str(), "%Lu", &_id);
	}

	if ((prop = node.property ("iolimits")) != 0) {
		sscanf (prop->value().c_str(), "%d,%d,%d,%d", 
			&_input_minimum,
			&_input_maximum,
			&_output_minimum,
			&_output_maximum);
	}
	
	if ((prop = node.property ("gain")) != 0) {
		set_gain (atof (prop->value().c_str()), this);
		_gain = _desired_gain;
	}

	XMLNodeList midi_kids;
	XMLNodeConstIterator iter;

	midi_kids = node.children ("MIDI");

	for (iter = midi_kids.begin(); iter != midi_kids.end(); ++iter) {
		
		XMLNodeList kids;
		XMLNodeConstIterator miter;
		XMLNode*    child;

		kids = (*iter)->children ();

		for (miter = kids.begin(); miter != kids.end(); ++miter) {

			child =* miter;

			if (child->name() == "gain") {
			
				MIDI::eventType ev = MIDI::on; /* initialize to keep gcc happy */
				MIDI::byte additional = 0;  /* ditto */
				MIDI::channel_t chn = 0;    /* ditto */
				bool ok = true;
				int xx;

				if ((prop = child->property ("event")) != 0) {
					sscanf (prop->value().c_str(), "0x%x", &xx);
					ev = (MIDI::eventType) xx;
				} else {
					ok = false;
				}

				if (ok && ((prop = child->property ("channel")) != 0)) {
					sscanf (prop->value().c_str(), "%d", &xx);
					chn = (MIDI::channel_t) xx;
				} else {
					ok = false;
				}

				if (ok && ((prop = child->property ("additional")) != 0)) {
					sscanf (prop->value().c_str(), "0x%x", &xx);
					additional = (MIDI::byte) xx;
				}

				if (ok) {
					_midi_gain_control.set_control_type (chn, ev, additional);
				} else {
					error << compose(_("MIDI gain control specification for %1 is incomplete, so it has been ignored"), _name) << endl;
				}
			}
		}
	}
			
	if (ports_legal) {

		if (create_ports (node)) {
			return -1;
		}

	} else {
		
		port_legal_c = PortsLegal.connect (slot (*this, &IO::ports_became_legal));
	}

	if (connecting_legal) {

		if (make_connections (node)) {
			return -1;
		}

	} else {
		
		connection_legal_c = ConnectingLegal.connect (slot (*this, &IO::connecting_became_legal));
	}

	if (!ports_legal || !connecting_legal) {
		pending_state_node = new XMLNode (node);
	}

	return 0;
}

int
IO::create_ports (const XMLNode& node)
{
	const XMLProperty* prop;
	int num_inputs = 0;
	int num_outputs = 0;

	if ((prop = node.property ("input-connection")) != 0) {
		Connection* c = _session.connection_by_name (prop->value());
		
		if (c == 0) {
			error << compose(_("Unknown connection \"%1\" listed for input of %2"), prop, _name) << endl;
			return -1;
		} else {
			num_inputs = c->nports();
		}

	} else if ((prop = node.property ("inputs")) != 0) {

		num_inputs = count (prop->value().begin(), prop->value().end(), '{');
	}
	
	if ((prop = node.property ("output-connection")) != 0) {
		Connection* c = _session.connection_by_name (prop->value());

		if (c == 0) {
			error << compose(_("Unknown connection \"%1\" listed for output of %2"), prop, _name) << endl;
			return -1;
		} else {
			num_outputs = c->nports ();
		}
		
	} else if ((prop = node.property ("outputs")) != 0) {
		num_outputs = count (prop->value().begin(), prop->value().end(), '{');
	}

	if (ensure_inputs (num_inputs, true, this) || ensure_outputs (num_outputs, true, this)) {
		error << compose(_("%1: cannot create I/O ports"), _name) << endmsg;
		return -1;
	}
	
	if ((prop = node.property ("pans")) != 0) {
		if (set_pans (prop->value())) {
			error << compose(_("CP: improper pan list in XML node (%1)"), prop->value()) << endmsg;
			return -1;
		} 
	}

	PortsCreated();
	return 0;
}

int
IO::make_connections (const XMLNode& node)
{
	const XMLProperty* prop;

	if ((prop = node.property ("input-connection")) != 0) {
		Connection* c = _session.connection_by_name (prop->value());
		
		if (c == 0) {
			error << compose(_("Unknown connection \"%1\" listed for input of %2"), prop, _name) << endl;
			return -1;
		} else {
			use_input_connection (*c, this);
		}

	} else if ((prop = node.property ("inputs")) != 0) {
		if (set_inputs (prop->value())) {
			error << compose(_("improper input channel list in XML node (%1)"), prop->value()) << endmsg;
			return -1;
		}
	}
	
	if ((prop = node.property ("output-connection")) != 0) {
		Connection* c = _session.connection_by_name (prop->value());
		
		if (c == 0) {
			error << compose(_("Unknown connection \"%1\" listed for output of %2"), prop, _name) << endl;
			return -1;
		} else {
			use_output_connection (*c, this);
		}
		
	} else if ((prop = node.property ("outputs")) != 0) {
		if (set_outputs (prop->value())) {
			error << compose(_("improper output channel list in XML node (%1)"), prop->value()) << endmsg;
			return -1;
		}
	}
	
	if ((prop = node.property ("pans")) != 0) {
		if (set_pans (prop->value())) {
			error << compose(_("MC: improper pan list in XML node (%1)"), prop->value()) << endmsg;
			return -1;
		}
	} 

	return 0;
}

int
IO::set_inputs (const string& str)
{
	vector<string> ports;
	int i;
	int n;
	int nports;
	
	if ((nports = count (str.begin(), str.end(), '{')) == 0) {
		return 0;
	}

	if (ensure_inputs (nports, true, this)) {
		return -1;
	}

	string::size_type start, end, ostart;

	ostart = 0;
	start = 0;
	end = 0;
	i = 0;

	while ((start = str.find_first_of ('{', ostart)) != string::npos) {
		start += 1;

		if ((end = str.find_first_of ('}', start)) == string::npos) {
			error << compose(_("IO: badly formed string in XML node for inputs \"%1\""), str) << endmsg;
			return -1;
		}

		if ((n = parse_io_string (str.substr (start, end - start), ports)) < 0) {
			error << compose(_("bad input string in XML node \"%1\""), str) << endmsg;

			return -1;
			
		} else if (n > 0) {

			for (int x = 0; x < n; ++x) {
				connect_input (input (i), ports[x], this);
			}
		}

		ostart = end+1;
		i++;
	}

	return 0;
}

int
IO::set_outputs (const string& str)
{
	vector<string> ports;
	int i;
	int n;
	int nports;
	
	if ((nports = count (str.begin(), str.end(), '{')) == 0) {
		return 0;
	}

	if (ensure_outputs (nports, true, this)) {
		return -1;
	}

	string::size_type start, end, ostart;

	ostart = 0;
	start = 0;
	end = 0;
	i = 0;

	while ((start = str.find_first_of ('{', ostart)) != string::npos) {
		start += 1;

		if ((end = str.find_first_of ('}', start)) == string::npos) {
			error << compose(_("IO: badly formed string in XML node for outputs \"%1\""), str) << endmsg;
			return -1;
		}

		if ((n = parse_io_string (str.substr (start, end - start), ports)) < 0) {
			error << compose(_("IO: bad output string in XML node \"%1\""), str) << endmsg;

			return -1;
			
		} else if (n > 0) {

			for (int x = 0; x < n; ++x) {
				connect_output (output (i), ports[x], this);
			}
		}

		ostart = end+1;
		i++;
	}

	return 0;
}

int
IO::set_pans (const string& str)
{
	vector<pan_t> p;
	unsigned long n;

	if ((n = parse_pan_string (str, p)) < 0) {
		return -1;
	}
	
	if (n != _noutputs) {
		error << compose(_("incorrect number of pans specified in XML node (%1 vs. %2"), n, _noutputs) << endmsg;
		return -1;
	}

	set_pans(p, this);

	return 0;
}

int
IO::parse_io_string (const string& str, vector<string>& ports)
{
	string::size_type pos, opos;

	if (str.length() == 0) {
		return 0;
	}

	pos = 0;
	opos = 0;

	ports.clear ();

	while ((pos = str.find_first_of (',', opos)) != string::npos) {
		ports.push_back (str.substr (opos, pos - opos));
		opos = pos + 1;
	}
	
	if (opos < str.length()) {
		ports.push_back (str.substr(opos));
	}

	return ports.size();
}

int
IO::parse_gain_string (const string& str, vector<string>& ports)
{
	string::size_type pos, opos;

	pos = 0;
	opos = 0;
	ports.clear ();

	while ((pos = str.find_first_of (',', opos)) != string::npos) {
		ports.push_back (str.substr (opos, pos - opos));
		opos = pos + 1;
	}
	
	if (opos < str.length()) {
		ports.push_back (str.substr(opos));
	}

	return ports.size();
}

int
IO::parse_pan_string (const string& str, vector<pan_t>& pans)
{
	string::size_type pos, opos;
	
	pos = 0;
	opos = 0;
	pans.clear ();

	while ((pos = str.find_first_of (',', opos)) != string::npos) {
		pans.push_back (atof (str.substr (opos, pos - opos).c_str()));
		opos = pos + 1;
	}
	
	if (opos < str.length()) {
		pans.push_back ( atof (str.substr(opos).c_str()));
	}

	return pans.size();
}

void
IO::set_name (string name, void* src)
{
	if (name == _name) {
		return;
	}

	for (vector<Port *>::iterator i = _inputs.begin(); i != _inputs.end(); ++i) {
		string current_name = (*i)->short_name();
		current_name.replace (current_name.find (_name), _name.length(), name);
		(*i)->set_name (current_name);
	}

	for (vector<Port *>::iterator i = _outputs.begin(); i != _outputs.end(); ++i) {
		string current_name = (*i)->short_name();
		current_name.replace (current_name.find (_name), _name.length(), name);
		(*i)->set_name (current_name);
	}

	_name = name;
	EMIT_SIGNAL name_changed (src);
}

int
IO::set_pans (const vector<pan_t>& pans, void* src)
{
	if (pans.size() != _noutputs) {
		error << compose(_("IO: incorrect number of pans used (%1 vs. %2 outputs"), pans.size(), _noutputs) << endmsg;
		return -1;
	}

	_pans = pans;
	EMIT_SIGNAL pan_changed (src);
	return 0;
}

void
IO::set_input_minimum (int n)
{
	_input_minimum = n;
}

void
IO::set_input_maximum (int n)
{
	_input_maximum = n;
}

void
IO::set_output_minimum (int n)
{
	_output_minimum = n;
}

void
IO::set_output_maximum (int n)
{
	_output_maximum = n;
}

jack_nframes_t
IO::output_latency () const
{
	jack_nframes_t max_latency;
	jack_nframes_t latency;

	max_latency = 0;

	/* io lock not taken - must be protected by other means */

	for (vector<Port *>::const_iterator i = _outputs.begin(); i != _outputs.end(); ++i) {
		if ((latency = _session.engine().get_port_total_latency (*(*i))) > max_latency) {
			max_latency = latency;
		}
	}

	return max_latency;
}

jack_nframes_t
IO::input_latency () const
{
	jack_nframes_t max_latency;
	jack_nframes_t latency;

	max_latency = 0;

	/* io lock not taken - must be protected by other means */

	for (vector<Port *>::const_iterator i = _inputs.begin(); i != _inputs.end(); ++i) {
		if ((latency = _session.engine().get_port_total_latency (*(*i))) > max_latency) {
			max_latency = latency;
		}
	}

	return max_latency;
}

int
IO::use_input_connection (Connection& c, void* src)
{
	unsigned int limit = c.nports();

	drop_input_connection ();
	if (ensure_inputs (limit, false, src)) {
		return -1;
	}

	{
		LockMonitor lm (_session.engine().process_lock(), __LINE__, __FILE__);
		{ 
			LockMonitor lm (io_lock, __LINE__, __FILE__);
			
			/* first pass: check the current state to see what's correctly
			   connected, and drop anything that we don't want.
			*/
			
			for (unsigned int n = 0; n < limit; ++n) {
				const Connection::PortList& pl = c.port_connections (n);
				
				for (Connection::PortList::const_iterator i = pl.begin(); i != pl.end(); ++i) {
					
					if (!_inputs[n]->connected_to ((*i))) {

						/* clear any existing connections */

						_session.engine().disconnect (_inputs[n]);

					} else if (_inputs[n]->connected() > 1) {

						/* OK, it is connected to the port we want,
						   but its also connected to other ports.
						   Change that situation.
						*/

						/* XXX could be optimized to not drop
						   the one we want.
						*/

						_session.engine().disconnect (_inputs[n]);
						
					}
				}
			}

			/* second pass: connect all requested ports where necessary */

			for (unsigned int n = 0; n < limit; ++n) {
				const Connection::PortList& pl = c.port_connections (n);
				
				for (Connection::PortList::const_iterator i = pl.begin(); i != pl.end(); ++i) {

					if (!_inputs[n]->connected_to ((*i))) {

						if (_session.engine().connect (*i, _inputs[n]->name())) {
							error << compose(_("IO: cannot connect input port %1 to %2"), _inputs[n]->name(), * i) << endmsg;
							return -1;
						}
					}
					
				}
			}
			
			_input_connection = &c;

			input_connection_configuration_connection = c.ConfigurationChanged.connect
				(slot (*this, &IO::input_connection_configuration_changed));
			input_connection_connection_connection = c.ConnectionsChanged.connect
				(slot (*this, &IO::input_connection_connection_changed));
		}
	}

	EMIT_SIGNAL input_changed (src);
	return 0;
}

int
IO::use_output_connection (Connection& c, void* src)
{
	unsigned int limit = c.nports();

	drop_output_connection ();

	if (ensure_outputs (limit, false, src)) {
		return -1;
	}

	{
		LockMonitor lm (_session.engine().process_lock(), __LINE__, __FILE__);
		{ 
			LockMonitor lm (io_lock, __LINE__, __FILE__);

			/* first pass: check the current state to see what's correctly
			   connected, and drop anything that we don't want.
			*/
			
			for (unsigned int n = 0; n < limit; ++n) {

				const Connection::PortList& pl = c.port_connections (n);
				
				for (Connection::PortList::const_iterator i = pl.begin(); i != pl.end(); ++i) {
					
					if (!_outputs[n]->connected_to ((*i))) {

						/* clear any existing connections */

						_session.engine().disconnect (_outputs[n]);

					} else if (_outputs[n]->connected() > 1) {

						/* OK, it is connected to the port we want,
						   but its also connected to other ports.
						   Change that situation.
						*/

						/* XXX could be optimized to not drop
						   the one we want.
						*/
						
						_session.engine().disconnect (_outputs[n]);
					}
				}
			}

			/* second pass: connect all requested ports where necessary */

			for (unsigned int n = 0; n < limit; ++n) {

				const Connection::PortList& pl = c.port_connections (n);
				
				for (Connection::PortList::const_iterator i = pl.begin(); i != pl.end(); ++i) {
					
					if (!_outputs[n]->connected_to ((*i))) {
						
						if (_session.engine().connect (_outputs[n]->name(), *i)) {
							error << compose(_("IO: cannot connect output port %1 to %2"), _outputs[n]->name(), * i) << endmsg;
							return -1;
						}
					}
				}
			}

			_output_connection = &c;

			output_connection_configuration_connection = c.ConfigurationChanged.connect
				(slot (*this, &IO::output_connection_configuration_changed));
			output_connection_connection_connection = c.ConnectionsChanged.connect
				(slot (*this, &IO::output_connection_connection_changed));
		}
	}

	EMIT_SIGNAL output_changed (src);

	return 0;
}

int
IO::disable_connecting ()
{
	connecting_legal = false;
	return 0;
}

int
IO::enable_connecting ()
{
	connecting_legal = true;
	return ConnectingLegal ();
}

int
IO::disable_ports ()
{
	ports_legal = false;
	return 0;
}

int
IO::enable_ports ()
{
	ports_legal = true;
	return PortsLegal ();
}

void
IO::input_connection_connection_changed (int ignored)
{
	use_input_connection (*_input_connection, this);
}

void
IO::input_connection_configuration_changed ()
{
	use_input_connection (*_input_connection, this);
}

void
IO::output_connection_connection_changed (int ignored)
{
	use_output_connection (*_output_connection, this);
}

void
IO::output_connection_configuration_changed ()
{
	use_output_connection (*_output_connection, this);
}

IO::MIDIGainControl::MIDIGainControl (IO& i, MIDI::Port* port)
	: MIDI::Controllable (port, 0), io (i)
{
	
}

void
IO::MIDIGainControl::set_value (float val)
{
	io.set_gain (midi_to_gain (val), this);
}

IO::MIDIPanControl::MIDIPanControl (IO& i, MIDI::Port* port)
	: MIDI::Controllable (port, 0), io (i)
{
}

void
IO::MIDIPanControl::set_value (float val)
{
	// what?
}

void
IO::reset_peak_meters ()
{
	unsigned long limit = max (_ninputs, _noutputs);

	for (unsigned long i = 0; i < limit; ++i) {
		_peak_power[i] = 0;
	}
}

void
IO::setup_peak_meters ()
{
	unsigned long limit = max (_ninputs, _noutputs);

	while (_peak_power.size() < limit) {
		_peak_power.push_back (0);
		_stored_peak_power.push_back (0);
		_pans.push_back (1.0);
	}
}

UndoAction
IO::get_memento() const
{
	void (IO::*pmf)(state_id_t) = &IO::set_state;
	return bind (slot (*(const_cast<IO *>(this)), pmf), _current_state_id);
}

void
IO::set_state (state_id_t id)
{
	return;
}

void
IO::grab_peak_power ()
{
	LockMonitor lm (io_lock, __LINE__, __FILE__);

	unsigned long limit = max (_ninputs, _noutputs);

	for (unsigned long n = 0; n < limit; ++n) {
		/* XXX should we use atomic exchange here ? */
		_stored_peak_power[n] = _peak_power[n];
		_peak_power[n] = 0;
	}
}

void
IO::reset_midi_control (MIDI::Port* port, bool on)
{
	MIDI::channel_t chn;
	MIDI::eventType ev;
	MIDI::byte extra;

	if (!on) {
		port = 0;
	}

	_midi_gain_control.get_control_info (chn, ev, extra);
	_midi_gain_control.midi_rebind (port, chn);
	
	_midi_pan_control.get_control_info (chn, ev, extra);
	_midi_pan_control.midi_rebind (port, chn);
}
