/*
    Copyright (C) 2000-2001 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: region.cc,v 1.49 2003/06/16 03:10:36 pbd Exp $
*/

#include <cmath>
#include <limits.h>

#include <sigc++/bind.h>
#include <sigc++/class_slot.h>

#include <pbd/lockmonitor.h>
#include <pbd/xml++.h>

#include <ardour/region.h>
#include <ardour/session.h>
#include <ardour/gain.h>
#include <ardour/dB.h>
#include <ardour/playlist.h>

#include "i18n.h"

using namespace ARDOUR;

SigC::Signal1<void,Region *>       Region::RegionCreated;

/* a Session will reset these to its chosen defaults by calling Region::set_default_fade() */

gain_t  *Region::default_fade_in_data = 0;
gain_t  *Region::default_fade_out_data = 0;
LogCurve Region::default_fade_in;
LogCurve Region::default_fade_out;

Region::Region (Source& src, jack_nframes_t start, jack_nframes_t length, const string& name, layer_t layer, Flag flags)
{
	/* basic Region constructor */

	sources.push_back(&src.ref());
	master_sources.push_back(&src.ref());

	_id = ARDOUR::new_id();
	_flags = flags;
	_sync_offset = 0;
	_playlist = 0;
	_read_data_count = 0;

	memset (_frozen, 0, sizeof (_frozen));

	_name = name;
	_start = start; 
	_length = length; 
	_position = 0; 
	_layer = layer;
	_current_state_id = 0;

	_fade_in_data = 0;
	_fade_out_data = 0;

	set_default_envelope ();
	save_state ("initial state");
	EMIT_SIGNAL RegionCreated (this);
}

Region::Region (SourceList& srcs, jack_nframes_t start, jack_nframes_t length, const string& name, layer_t layer, Flag flags)
{
	/* basic Region constructor */

	for (SourceList::iterator i=srcs.begin(); i != srcs.end(); ++i) {
		sources.push_back (&(*i)->ref());
		master_sources.push_back (&(*i)->ref());
	}

	_id = ARDOUR::new_id();
	_flags = flags;
	_sync_offset = 0;
	_playlist = 0;
	_read_data_count = 0;

	memset (_frozen, 0, sizeof (_frozen));

	_name = name;
	_start = start; 
	_length = length; 
	_position = 0; 
	_layer = layer;
	_current_state_id = 0;
	_read_data_count = 0;

	_fade_in_data = 0;
	_fade_out_data = 0;

	set_default_envelope ();
	save_state ("initial state");
	EMIT_SIGNAL RegionCreated (this);
}


Region::Region (const Region& other, jack_nframes_t offset, jack_nframes_t length, const string& name, layer_t layer, Flag flags)
{
	/* create a new Region, that is part of an existing one */

	for (SourceList::const_iterator i= other.sources.begin(); i != other.sources.end(); ++i) {
		sources.push_back (&((*i)->ref()));
	}

	for (SourceList::const_iterator i = other.master_sources.begin(); i != other.master_sources.end(); ++i) {
		master_sources.push_back (&((*i)->ref()));
	}

	_id = ARDOUR::new_id();
	memset (_frozen, 0, sizeof (_frozen));
	_sync_offset = 0;
	_playlist = 0;
	_read_data_count = 0;

	_start = other._start + offset; 
	_length = length; 
	_name = name;
	_position = 0; 
	_layer = layer; 
	_flags = flags;
	_current_state_id = 0;

	_fade_in_data = 0;
	_fade_out_data = 0;
	_fade_in = other._fade_in;
	_fade_out = other._fade_out;
	build_fade (&_fade_in_data, _fade_in, true);
	build_fade (&_fade_out_data, _fade_out, false);

	set_default_envelope (); /* XXX should really take part of the curve of `other' */
	save_state ("initial state");
	EMIT_SIGNAL RegionCreated (this);
}

Region::Region (const Region &other, const string& name)
{
	/* create a new Region, with a new name, by copying */

	for (SourceList::const_iterator i = other.sources.begin(); i != other.sources.end(); ++i) {
		sources.push_back (&((*i)->ref()));
	}

	for (SourceList::const_iterator i = other.master_sources.begin(); i != other.master_sources.end(); ++i) {
		master_sources.push_back (&((*i)->ref()));
	}

	_id = ARDOUR::new_id();
	memset (_frozen, 0, sizeof (_frozen));
	_sync_offset = 0;
	_playlist = 0;
	_read_data_count = 0;

	_name = name;

	_start = other._start; 
	_length = other._length; 
	_position = other._position; 
	_layer = other._layer; 
	_envelope = other._envelope; 
	_flags = other._flags;
	_current_state_id = 0;

	_fade_in_data = 0;
	_fade_out_data = 0;
	_fade_in = other._fade_in;
	_fade_out = other._fade_out;
	build_fade (&_fade_in_data, _fade_in, true);
	build_fade (&_fade_out_data, _fade_out, false); 

	_envelope = other._envelope;

	save_state ("initial state");
	EMIT_SIGNAL RegionCreated (this);
}

Region::Region (const Region &other)
{
	/* Pure copy constructor */

	for (SourceList::const_iterator i = other.sources.begin(); i != other.sources.end(); ++i) {
		sources.push_back (&((*i)->ref()));
	}

	for (SourceList::const_iterator i = other.master_sources.begin(); i != other.master_sources.end(); ++i) {
		master_sources.push_back (&((*i)->ref()));
	}

	_id = other.id();
	memset (_frozen, 0, sizeof (_frozen));
	_sync_offset = 0;
	_playlist = 0;
	_read_data_count = 0;

	if (other._extra_xml) {
		_extra_xml = new XMLNode (*other._extra_xml);
	} else {
		_extra_xml = 0;
	}

	_start = other._start;
	_length = other._length; 
	_name = other._name;
	_position = other._position; 
	_layer = other._layer; 
	_envelope = other._envelope; 

	/* This is the only place the Copied flag is set.
	   It is unset by ::maybe_uncopy(), which renames
	   the region when certain operations are carried
	   out that require us to make it clear that
	   the region is no longer an exact copy.
	*/

	//	_flags = other._flags | Copied;

	_flags = other._flags;
	_current_state_id = 0;

	_fade_in_data = 0;
	_fade_out_data = 0;
	_fade_in = other._fade_in;
	_fade_out = other._fade_out;
	build_fade (&_fade_in_data, _fade_in, true);
	build_fade (&_fade_out_data, _fade_out, false);

	_envelope = other._envelope;

	save_state ("initial state");
	/* NOTE: no RegionCreated signal emitted here. This is the copy constructor */
}

Region::Region (Source& src, const XMLNode& node)
{
	sources.push_back(&src.ref());
	master_sources.push_back(&src.ref());

	_id = 0;
	memset (_frozen, 0, sizeof (_frozen));
	_sync_offset = 0;
	_playlist = 0;
	_read_data_count = 0;
	_start = 0; 
	_length = 0;
	_name = "error: XML did not reset this";
	_position = 0; 
	_layer = 0;
	_flags = 0;
	_current_state_id = 0;
	_fade_in_data = 0;
	_fade_out_data = 0;
	build_fade (&_fade_in_data, _fade_in, true);
	build_fade (&_fade_out_data, _fade_out, false);

	if (set_state (node)) {
		throw failed_constructor();
	}

	save_state ("initial state");
	EMIT_SIGNAL RegionCreated (this);
}

Region::Region (SourceList& srcs, const XMLNode& node)
{
	/* basic Region constructor */
	for (SourceList::iterator i=srcs.begin(); i != srcs.end(); ++i) {
		sources.push_back (&(*i)->ref());
	}

	for (SourceList::iterator i = srcs.begin(); i != srcs.end(); ++i) {
		master_sources.push_back (&((*i)->ref()));
	}

	_id = 0;
	memset (_frozen, 0, sizeof (_frozen));
	_sync_offset = 0;
	_playlist = 0;
	_read_data_count = 0;
	_start = 0; 
	_length = 0;
	_name = "error: XML did not reset this";
	_position = 0; 
	_layer = 0;
	_flags = 0;
	_current_state_id = 0;
	_fade_in_data = 0;
	_fade_out_data = 0;
	build_fade (&_fade_in_data, _fade_in, true);
	build_fade (&_fade_out_data, _fade_out, false);

	if (set_state (node)) {
		throw failed_constructor();
	}

	save_state ("initial state");
	EMIT_SIGNAL RegionCreated (this);
}

Region::~Region ()
{
	GoingAway (this);

	for (SourceList::iterator i = sources.begin(); i != sources.end(); ++i) {
		(*i)->unref();
	}

	for (SourceList::iterator i = master_sources.begin(); i != master_sources.end(); ++i) {
		(*i)->unref();
	}
}

void
Region::set_playlist (Playlist* pl)
{
	_playlist = pl;
}

void
Region::store_state (RegionState& state) const
{
	state._start = _start;
	state._length = _length;
	state._position = _position;
	state._flags = _flags;
	state._fade_in = _fade_in;
	state._fade_out = _fade_out;
	state._sync_offset = _sync_offset;
	state._layer = _layer;
	state._envelope = _envelope;
	state._name = _name;
}	

void
Region::restore_state (RegionState& state)
{
	FreezeFlag freeze_flags = FreezeFlag (0);

	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		bool bounds_change = false;
		
		if (_start != state._start) {
			bounds_change = true;
			_start = state._start;
		}
		if (_length != state._length) {
			bounds_change = true;
			_length = state._length;
		}
		if (_position != state._position) {
			bounds_change = true;
			_position = state._position;
		} 
		
		if (bounds_change) {
			freeze_flags = FreezeFlag (freeze_flags|BoundsNotification);
		}
		
		if (_flags != state._flags) {
			unsigned int old_flags = _flags;
			
			_flags = state._flags;
			
			if ((old_flags ^ state._flags) & Muted) {
				EMIT_SIGNAL MuteChanged();
			}
			if ((old_flags ^ state._flags) & Opaque) {
				EMIT_SIGNAL OpacityChanged ();
			}
			if ((old_flags ^ state._flags) & Locked) {
				EMIT_SIGNAL LockChanged ();
			}
			if ((old_flags ^ state._flags) & EnvelopeActive) {
				freeze_flags = FreezeFlag (freeze_flags|EnvelopeNotification);
			}
		}
		
		if (_fade_in != state._fade_in) {
			_fade_in = state._fade_in;
			build_fade (&_fade_in_data, _fade_in, true);
			freeze_flags = FreezeFlag (freeze_flags|FadeNotification);
		}
		if (_fade_out != state._fade_out) {
			_fade_out = state._fade_out;
			build_fade (&_fade_out_data, _fade_out, false);
			freeze_flags = FreezeFlag (freeze_flags|FadeNotification);
		}
		
		if (_sync_offset != state._sync_offset) {
			_sync_offset = state._sync_offset;
			freeze_flags = FreezeFlag (freeze_flags|SyncNotification);
		}
		
		if (_layer != state._layer) {
			_layer = state._layer;
			EMIT_SIGNAL LayerChanged ();
		}

		_envelope = state._envelope;
		_name = state._name;

		freeze_flags = FreezeFlag (freeze_flags|EnvelopeNotification);
	}
	
	freeze (freeze_flags);
	thaw (freeze_flags, "state restored", true, this);
}

UndoAction
Region::get_memento() const
{
	void (Region::*pmf)(state_id_t) = &Region::set_state;
	return bind (slot (*(const_cast<Region *> (this)), pmf), _current_state_id);
}

void
Region::save_state (const string& why)
{
	LockMonitor lm (lock, __LINE__, __FILE__);
	
	State *state_now = new State (why);
	store_state (state_now->state);
	states.push_back (state_now);
	_current_state_id = states.size() - 1;
}
 
void
Region::set_state (state_id_t id)
{
	if (id >= states.size()) {
		fatal << _("programming error: illegal state ID passed to Playlist::set_state()") << endmsg;
		/*NOTREACHED*/
		return;
	}

	restore_state (states[id]->state);
	_current_state_id = id;
}

void
Region::set_name (string str)

{
	if (_name != str) {
		_name = str; 
		EMIT_SIGNAL NameChanged();
	}
}

void
Region::add_envelope_control_point (jack_nframes_t frame, gain_t val)
{
	_envelope.add_control_point (frame, val);

	if (!hold_envelope_notification()) {
		char buf[64];
		sprintf (buf, "added gain control point at %lu,%f", frame, val);
		save_state (buf);
		EnvelopeChanged();
	}
}

void
Region::remove_envelope_control_point (size_t n)
{
	_envelope.remove_control_point (n);
	
	if (!hold_envelope_notification()) {
		char buf[64];
		snprintf (buf, sizeof(buf), "removed gain control");
		save_state (buf);
		EnvelopeChanged ();
	}
}

void
Region::remove_envelope_control_points (size_t start, size_t end)
{
	_envelope.remove_control_points (start, end);
	
	if (!hold_envelope_notification()) {
		char buf[64];
		snprintf (buf, sizeof(buf), "removed gain controls");
		save_state (buf);
		EnvelopeChanged ();
	}
}

void
Region::modify_envelope_control_point (size_t n, jack_nframes_t frame, gain_t val)
{
	if (n >= _envelope.n_control_points()) {
		return;
	}

	/* clamp frame position to the adjacent points, or the ends of the region */
	
	if (n == 0) {
		frame = max (frame, 0UL);
	} else {
		double x, y;
		_envelope.nth (n-1, x, y);
		frame = max (frame, (jack_nframes_t) x);
	}
	
	if (n < _envelope.n_control_points() - 1) {
		double x, y;
		_envelope.nth (n+1, x, y);
		frame = min (frame, (jack_nframes_t) x);
	} else {
		frame = min (frame, _length);
	}

	/* clamp height to the standard normalized range */
	
	val = min (val, 2.0f);
	val = max (val, 0.0f);
	
	_envelope.reset_control_point (n, frame, val);
	
	if (!hold_envelope_notification()) {
		char buf[64];
		sprintf (buf, "moved gain control point #%dZ", (int)n);
		save_state (buf);
		EnvelopeChanged();
	}
}

void
Region::modify_envelope_control_points (size_t start, size_t end, gain_t delta)
{
	_envelope.reset_control_points (start, end, delta);

	if (!hold_envelope_notification()) {
		char buf[64];
		snprintf (buf, sizeof (buf), "moved gain curve segment");
		save_state (buf);
		EnvelopeChanged();
	}
}

void
Region::set_length (jack_nframes_t len, void *src)
{
	if (_flags & Locked) {
		return;
	}

	if (_length != len && len != 0) {
		for (unsigned int n=0; n < sources.size(); ++n) {
			if (_start > sources[n]->length() - len) {
				return;
			}
		}
		
		_length = len;

		recompute_gain_at_end ();
		
		maybe_uncopy ();

		char buf[64];
		sprintf (buf, "length set to %lu", len);
		save_state (buf);

		if (!hold_bounds_notification()) {
			EMIT_SIGNAL BoundsChanged (LengthChange, src);
		}
	}
}

void
Region::maybe_uncopy ()
{
	return;

#if 0
	if (!copied()) {
		return;
	}

	/* rename */
	
	string::size_type pos;
	pos = _name.length() - 1;
	
	while (isdigit (_name[pos]) && pos > 0) {
		--pos;
	}
	
	if (pos == 0) {
		fatal << compose (_("programmer error: illegal region name: %1"), _name) << endmsg;
				/*NOTREACHED*/
	}
	
	/* pos now indexes the start of the last numerical
	   part of the region name.
	*/
	
	++pos;
	
	unsigned long x = atoi (_name.substr(pos).c_str());
	char buf[32];
	snprintf (buf, sizeof(buf), "%lu", x + 1);
	_name.replace (pos, _name.substr(pos).length(), buf);
	
	EMIT_SIGNAL NameChanged ();
	
	_flags = (_flags & ~Copied);
#endif
}
	
void
Region::set_position (jack_nframes_t pos, void *src)
{
	if (_flags & Locked) {
		return;
	}

	if (_position != pos) {
		_position = pos;

		char buf[64];
		sprintf (buf, "position set to %lu", pos);
		save_state (buf);
	}

	/* do this even if the position is the same. this helps out
	   a GUI that has moved its representation already.
	*/

	if (!hold_bounds_notification()) {
		EMIT_SIGNAL BoundsChanged (PositionChange, src);
	}
}

void
Region::set_position_on_top (jack_nframes_t pos, void *src)
{
	bool changed = false;

	if (_flags & Locked) {
		return;
	}

	if (_position != pos) {
		_position = pos;
		changed = true;
	}
	
	if (changed) {
		char buf[64];
		sprintf (buf, "position set to %lu", pos);
		save_state (buf);
	}

	_playlist->raise_region_to_top (*this);
	
	/* do this even if the position is the same. this helps out
	   a GUI that has moved its representation already.
	*/
	
	if (!hold_bounds_notification()) {
		EMIT_SIGNAL BoundsChanged (PositionChange, src);
	}
}

void
Region::nudge_position (long n, void *src)
{
	if (_flags & Locked) {
		return;
	}

	if (n == 0) {
		return;
	}
	
	if (n > 0) {
		if (_position > max_frames - n) {
			_position = max_frames;
		} else {
			_position += n;
		}
	} else {
		if (_position < (jack_nframes_t) -n) {
			_position = 0;
		} else {
			_position += n;
		}
	}

	char buf[64];
	sprintf (buf, "position set to %lu", _position);
	save_state (buf);

	if (!hold_bounds_notification()) {
		EMIT_SIGNAL BoundsChanged (PositionChange, src);
	}
}

void
Region::set_start (jack_nframes_t pos, void *src)
{
	if (_flags & Locked) {
		return;
	}
	/* This just sets the start, nothing else. It effectively shifts
	   the contents of the Region within the overall extent of the Source,
	   without changing the Region's position or length
	*/

	if (_start != pos) {
		for (unsigned int n=0; n < sources.size(); ++n) {
			if (pos > sources[n]->length() - _length) {
				return;
			}
		}


		_start = pos;

		maybe_uncopy ();

 		char buf[64];
		sprintf (buf, "start set to %lu", pos);
		save_state (buf);

		if (!hold_bounds_notification()) {
			EMIT_SIGNAL BoundsChanged (StartChange, src);
		}
	}
}

void
Region::trim_start (jack_nframes_t new_position, void *src)
{
	if (_flags & Locked) {
		return;
	}
	jack_nframes_t new_start;
	long start_shift;
	
	if (new_position > _position) {
		start_shift = new_position - _position;
	} else {
		start_shift = -(_position - new_position);
	}

	if (start_shift > 0) {

		if (_start > max_frames - start_shift) {
			new_start = max_frames;
		} else {
			new_start = _start + start_shift;
		}

		for (unsigned int n=0; n < sources.size(); ++n) {
			if (new_start > sources[n]->length() - _length) {
				new_start = sources[n]->length() - _length;
			}
		}
		

	} else if (start_shift < 0) {

		if (_start < (jack_nframes_t) -start_shift) {
			new_start = 0;
		} else {
			new_start = _start + start_shift;
		}
	} else {
		return;
	}

	if (new_start == _start) {
		return;
	}
	
	_start = new_start;

	maybe_uncopy ();

	char buf[64];
	sprintf (buf, "slipped start to %lu", _start);
	save_state (buf);

	if (!hold_bounds_notification()) {
		EMIT_SIGNAL BoundsChanged (StartChange, src);
	}
}

void
Region::trim_front (jack_nframes_t new_position, void *src)
{
	if (_flags & Locked) {
		return;
	}

	jack_nframes_t end = _position + _length - 1;

	if ((new_position < end) && ((new_position + _start) >= _position)) {
		jack_nframes_t newlen;

		if (new_position > _position) {
			newlen = _length - (new_position - _position);
		} else {
			newlen = _length + (_position - new_position);
		}

		trim_to_internal (new_position, newlen, src);
		recompute_gain_at_start ();
	}
}

void
Region::trim_end (jack_nframes_t new_position, void *src)
{
	if (_flags & Locked) {
		return;
	}
	if (new_position > _position) {
		trim_to_internal (_position, new_position - _position, this);
		recompute_gain_at_end ();
	}
}

void
Region::trim_to (jack_nframes_t position, jack_nframes_t length, void *src)
{
	if (_flags & Locked) {
		return;
	}
	trim_to_internal (position, length, src);
	recompute_gain_at_start ();
	recompute_gain_at_end ();
}

void
Region::trim_to_internal (jack_nframes_t position, jack_nframes_t length, void *src)

{
	long start_shift;
	jack_nframes_t new_start;

	if (_flags & Locked) {
		return;
	}

	if (position > _position) {
		start_shift = position - _position;
	} else {
		start_shift = -(_position - position);
	}

	if (start_shift > 0) {

		if (_start > max_frames - start_shift) {
			new_start = max_frames;
		} else {
			new_start = _start + start_shift;
		}

		for (unsigned int n=0; n < sources.size(); ++n) {
			if (new_start > sources[n]->length() - length) {
				new_start = sources[n]->length() - length;
			}
		}
		

	} else if (start_shift < 0) {

		if (_start < (jack_nframes_t) -start_shift) {
			new_start = 0;
		} else {
			new_start = _start + start_shift;
		}
	} else {
		new_start = _start;
	}

	for (unsigned int n=0; n < sources.size(); ++n) {
		if (length > sources[n]->length() - new_start) {
			return;
		}
	}
	
	BoundsChangeInfo bounds_bits = BoundsChangeInfo (0);

	if (_start != new_start) {
		_start = new_start;
		bounds_bits = BoundsChangeInfo (bounds_bits|StartChange);
	}
	if (_length != length) {
		_length = length;
		bounds_bits = BoundsChangeInfo (bounds_bits|LengthChange);
	}
	if (_position != position) {
		_position = position;
		bounds_bits = BoundsChangeInfo (bounds_bits|PositionChange);
	}

	if (!hold_bounds_notification() && bounds_bits) {
		char buf[64];
		maybe_uncopy ();
		sprintf (buf, "trimmed to %lu-%lu", _position, _position+_length-1);
		save_state (buf);
		EMIT_SIGNAL BoundsChanged (bounds_bits, src);
	}
}	

void
Region::set_muted (bool yn)
{
	if (muted() != yn) {
		char buf[64];
		if (yn) {
			sprintf (buf, "muted");
			_flags |= Muted;
		} else {
			sprintf (buf, "unmuted");
			_flags &= ~Muted;
		}
		save_state (buf);
		EMIT_SIGNAL MuteChanged ();
	}
}

void
Region::set_opaque (bool yn)
{
	if (opaque() != yn) {
		char buf[64];
		if (yn) {
			sprintf (buf, "opaque");
			_flags |= Opaque;
		} else {
			sprintf (buf, "translucent");
			_flags &= ~Opaque;
		}
		save_state (buf);
		EMIT_SIGNAL OpacityChanged ();
	}
}

void
Region::set_envelope_active (bool yn)
{
	if (envelope_active() != yn) {
		char buf[64];
		if (yn) {
			sprintf (buf, "envelope active");
			_flags |= EnvelopeActive;
		} else {
			sprintf (buf, "envelope off");
			_flags &= ~EnvelopeActive;
		}
		save_state (buf);
		EMIT_SIGNAL EnvelopeActiveChanged ();
	}
}

void
Region::set_locked (bool yn)
{
	if (locked() != yn) {
		char buf[64];
		if (yn) {
			sprintf (buf, "locked");
			_flags |= Locked;
		} else {
			sprintf (buf, "unlocked");
			_flags &= ~Locked;
		}
		save_state (buf);
		EMIT_SIGNAL LockChanged ();
	}
}

void
Region::set_sync_offset (jack_nframes_t n)
{
	if (_sync_offset != n && n < _length) {
		_sync_offset = n;

		if (!hold_sync_notification()) {
			char buf[64];
			maybe_uncopy ();
			sprintf (buf, "sync offset set to %lu", _sync_offset);
			save_state (buf);
			EMIT_SIGNAL SyncOffsetChanged ();
		}
	}
}

void
Region::raise ()
{
	if (_playlist == 0) {
		return;
	}

	_playlist->raise_region (*this);
}

void
Region::lower ()
{
	if (_playlist == 0) {
		return;
	}

	_playlist->lower_region (*this);
}

void
Region::raise_to_top ()
{
	if (_playlist == 0) {
		return;
	}

	_playlist->raise_region_to_top (*this);
}

void
Region::lower_to_bottom ()
{
	if (_playlist == 0) {
		return;
	}

	_playlist->lower_region_to_bottom (*this);
}

void
Region::set_layer (layer_t l)
{
	if (_layer != l) {
		_layer = l;
		char buf[64];
		sprintf (buf, "layer set to %lu", _layer);
		save_state (buf);
		if (!hold_layer_notification()) {
			EMIT_SIGNAL LayerChanged ();
		}
	}
}


void
Region::use_default_fade_out ()
{
	if (_flags & DefaultFadeOut) {
		return;
	}
	
	_flags |= DefaultFadeOut;

	/* reset the non-default one to match the default, but
	   be lazy by not constructing the data set. we won't
	   need it till someone resets one of the fade params,
	   when we'll recalculate it anyway.
	*/

	_fade_out.set_steepness (default_fade_out.steepness());
	_fade_out.set_length (default_fade_out.length());
	build_fade (&_fade_out_data, _fade_out, false);

	EMIT_SIGNAL FadeOutChanged ();
}

void
Region::use_default_fade_in ()
{
	if (_flags & DefaultFadeIn) {
		return;
	}
	
	_flags |= DefaultFadeIn;

	/* reset the non-default one to match the default, but
	   be lazy by not constructing the data set. we won't
	   need it till someone resets one of the fade params,
	   when we'll recalculate it anyway.
	*/

	_fade_in.set_steepness(default_fade_in.steepness());
	_fade_in.set_length (default_fade_in.length());

	EMIT_SIGNAL FadeInChanged ();
}

void
Region::set_fade_in (float steepness, jack_nframes_t length)
{
	if (steepness == _fade_in.steepness() && length == _fade_in.length()) {
		return;
	}

	_fade_in.set_steepness(steepness);
	_fade_in.set_length (length);
	
	_flags &= ~DefaultFadeIn;

	char buf[64];
	sprintf (buf, "fade in set to %lu/%f", _fade_in.length(), _fade_in.steepness());
	save_state (buf);

	build_fade (&_fade_in_data, _fade_in, true);
	EMIT_SIGNAL FadeInChanged ();
}

void
Region::set_fade_out (float steepness, jack_nframes_t length)
{
	if (steepness == _fade_out.steepness() && length == _fade_out.length()) {
		return;
	}

	_fade_out.set_steepness(steepness);
	_fade_out.set_length (length);

	_flags &= ~DefaultFadeOut;

	char buf[64];
	sprintf (buf, "fade out set to %lu/%f", _fade_out.length(), _fade_out.steepness());
	save_state (buf);
	build_fade (&_fade_out_data, _fade_out, false);
	EMIT_SIGNAL FadeOutChanged ();
}

jack_nframes_t
Region::read_peaks (PeakData *buf, jack_nframes_t npeaks, jack_nframes_t offset, jack_nframes_t cnt, unsigned int chan_n) const
{
	if (chan_n >= sources.size()) {
		return 0;
	}
	
	if (sources[chan_n]->read_peaks (buf, npeaks, _start + offset, cnt)) {
		return 0;
	} else {
		return cnt;
	}
}

jack_nframes_t
Region::read_at (Sample *buf, Sample *mixdown_buffer, float *gain_buffer, jack_nframes_t position, jack_nframes_t cnt, unsigned int chan_n) const
{
	return _read_at (sources, buf, mixdown_buffer, gain_buffer, position, cnt, chan_n);
}

jack_nframes_t
Region::master_read_at (Sample *buf, Sample *mixdown_buffer, float *gain_buffer, jack_nframes_t position, jack_nframes_t cnt, unsigned int chan_n) const
{
	return _read_at (master_sources, buf, mixdown_buffer, gain_buffer, position, cnt, chan_n);
}

jack_nframes_t
Region::_read_at (const SourceList& srcs, Sample *buf, Sample *mixdown_buffer, float *gain_buffer, jack_nframes_t position, jack_nframes_t cnt, unsigned int chan_n) const
{
	jack_nframes_t internal_offset;
	jack_nframes_t buf_offset;
	jack_nframes_t to_read;
	
	/* precondition: caller has verified that we cover the desired section */
	if (chan_n >= sources.size()) {
		return 0; /* read nothing */
	}
	
	if (muted()) {
		return 0; /* read nothing */
	}

	if (position < _position) {
		internal_offset = 0;
		buf_offset = _position - position;
		cnt -= buf_offset;
	} else {
		internal_offset = position - _position;
		buf_offset = 0;
	}

	if (internal_offset >= _length) {
		return 0; /* read nothing */
	}

	if ((to_read = min (cnt, _length - internal_offset)) == 0) {
		return 0; /* read nothing */
	}

	if (opaque()) {
		/* overwrite whatever is there */
		mixdown_buffer = buf + buf_offset;
	} else {
		mixdown_buffer += buf_offset;
	}

	_read_data_count = 0;

	if (srcs[chan_n]->read (mixdown_buffer, _start + internal_offset, to_read) != to_read) {
		return 0; /* "read nothing" */
	}

	_read_data_count += srcs[chan_n]->read_data_count();

	/* Apply fades if appropriate */

	const LogCurve *fin;
	const LogCurve *fout;
	gain_t   *din;
	gain_t   *dout;

	if (_flags & DefaultFadeIn) {
		fin = &default_fade_in;
		din = default_fade_in_data;
	} else {
		fin = &_fade_in;
		din = _fade_in_data;
	}

	if (_flags & DefaultFadeOut) {
		fout = &default_fade_out;
		dout = default_fade_out_data;
	} else {
		fout = &_fade_out;
		dout = _fade_out_data;
	}

	/* fade in */

	if (din && (_flags & FadeIn) && internal_offset < fin->length()) {

		jack_nframes_t gain = internal_offset;
		jack_nframes_t limit = min (fin->length() - gain, to_read);

		for (jack_nframes_t n = 0; n < limit; ++n, ++gain) {
			mixdown_buffer[n] *= din[gain];
		}
	}
	
	/* fade out */

	if (_flags & FadeOut) {
		jack_nframes_t fade_out_start = _length - fout->length();
		
		if (dout && internal_offset+to_read >= fade_out_start) {
			
			jack_nframes_t limit;
			jack_nframes_t frame;
			jack_nframes_t gain;

	
			if (internal_offset < fade_out_start) {
				jack_nframes_t fade_out_offset = fade_out_start - internal_offset;
				limit = min (fout->length(), to_read - fade_out_offset);
				frame = fade_out_offset;
				gain = 0;
			} else {
				jack_nframes_t fade_out_offset = internal_offset - fade_out_start;
				limit = min (fout->length() - fade_out_offset, to_read);
				frame = 0;
				gain = fade_out_offset;
			}

			for (jack_nframes_t n = 0; n < limit; ++n, ++frame, ++gain) {
				mixdown_buffer[frame] *= dout[gain];
			}
		}
	}

	/* Regular gain curves */

	if (envelope_active())  {
		_envelope.get_vector (internal_offset, internal_offset + to_read, gain_buffer, to_read);
		
		for (jack_nframes_t n = 0; n < to_read; ++n) {
			mixdown_buffer[n] *= gain_buffer[n];
		}
	}

	if (!opaque()) {

		/* gack. the things we do for users.
		 */

		buf += buf_offset;

		for (jack_nframes_t n = 0; n < to_read; ++n) {
			*buf++ += *mixdown_buffer++;
		}
	} 
	
	return to_read;
}
	
XMLNode&
Region::get_state ()
{
	XMLNode *node = new XMLNode ("Region");
	XMLNode *child;
	char buf[64];
	char buf2[64];
	
	sprintf (buf, "%Lu", _id);
	node->add_property ("id", buf);
	node->add_property ("name", _name);
	sprintf (buf, "%lu", _start);
	node->add_property ("start", buf);
	sprintf (buf, "%lu", _length);
	node->add_property ("length", buf);
	sprintf (buf, "%lu", _position);
	node->add_property ("position", buf);
	sprintf (buf, "0x%x", (int) _flags);
	node->add_property ("flags", buf);

	/* backward compatibility */
	sprintf (buf, "%Lu", sources[0]->id());
	node->add_property ("source", buf);

	for (unsigned int n=0; n < sources.size(); ++n) {
		snprintf (buf2, sizeof(buf2), "source-%d", n);
		snprintf (buf, sizeof(buf), "%Lu", sources[n]->id());
		node->add_property (buf2, buf);
	}

	sprintf (buf, "%u", (unsigned int) sources.size());
	node->add_property ("channels", buf);
	
	sprintf (buf, "%d", (int) _layer);
	node->add_property ("layer", buf);
	sprintf (buf, "%lu", _sync_offset);
	node->add_property ("sync-offset", buf);
	
	if ((_flags & DefaultFadeIn) == 0) {
		child = node->add_child ("FadeIn");
		sprintf (buf, "%f", _fade_in.steepness());
		child->add_property ("steepness", buf);
		sprintf (buf, "%lu", _fade_in.length());
		child->add_property ("length", buf);
	} 

	if ((_flags & DefaultFadeOut) == 0) {
		child = node->add_child ("FadeOut");
		sprintf (buf, "%f", _fade_out.steepness());
		child->add_property ("steepness", buf);
		sprintf (buf, "%lu", _fade_out.length());
		child->add_property ("length", buf);
	}

	child = node->add_child ("Envelope");

	bool default_env = false;

	if (_envelope.n_control_points() == 2) {

		double x, y;
		_envelope.nth (1, x, y);

		if (x == _length) {
			default_env = true;
			child->add_property ("default", "yes");
			
		}
	} 

	if (default_env) {
		child->add_property ("default", "yes");
	} else {
		double x, y;
		for (size_t n = 0; n < _envelope.n_control_points(); ++n) {
			XMLNode *pointnode = new XMLNode ("point");
			
			_envelope.nth (n, x, y);

			sprintf (buf, "%g", x);
			pointnode->add_property ("x", buf);
			sprintf (buf, "%g", y);
			pointnode->add_property ("y", buf);
			child->add_child_nocopy (*pointnode);
		}
	}

	if (_extra_xml) {
		node->add_child_copy (*_extra_xml);
	}

	return *node;
}

XMLNode&
Region::get_short_state ()
{
	XMLNode *node = new XMLNode ("Region");
	XMLNode *child;
	char buf[64];
	char buf2[64];
	
	sprintf (buf, "%Lu", _id);
	node->add_property ("id", buf);
	node->add_property ("name", _name);
	sprintf (buf, "%lu", _start);
	node->add_property ("start", buf);
	sprintf (buf, "%lu", _length);
	node->add_property ("length", buf);
	sprintf (buf, "0x%x", (int) _flags);
	node->add_property ("flags", buf);

	/* backward compatibility */
	sprintf (buf, "%Lu", sources[0]->id());
	node->add_property ("source", buf);

	for (unsigned int n=0; n < sources.size(); ++n) {
		snprintf (buf2, sizeof(buf2), "source-%d", n);
		snprintf (buf, sizeof(buf), "%Lu", sources[n]->id());
		node->add_property (buf2, buf);
	}

	sprintf (buf, "%u", (unsigned int) sources.size());
	node->add_property ("channels", buf);

	sprintf (buf, "%lu", _sync_offset);
	node->add_property ("sync-offset", buf);
	
	if ((_flags & DefaultFadeIn) == 0) {
		child = node->add_child ("FadeIn");
		sprintf (buf, "%f", _fade_in.steepness());
		child->add_property ("steepness", buf);
		sprintf (buf, "%lu", _fade_in.length());
		child->add_property ("length", buf);
	} 

	if ((_flags & DefaultFadeOut) == 0) {
		child = node->add_child ("FadeOut");
		sprintf (buf, "%f", _fade_out.steepness());
		child->add_property ("steepness", buf);
		sprintf (buf, "%lu", _fade_out.length());
		child->add_property ("length", buf);
	}

	child = node->add_child ("Envelope");
	child->add_property ("default", "yes");

	return *node;
}

int
Region::set_state (const XMLNode& node)
{
	const XMLNodeList& nlist = node.children();
	const XMLProperty *prop;

	if (_extra_xml) {
		delete _extra_xml;
		_extra_xml = 0;
	}

	if ((prop = node.property ("id")) == 0) {
		error << _("Session: XMLNode describing a Region is incomplete (no id)") << endmsg;
		return -1;
	}

	sscanf (prop->value().c_str(), "%Lu", &_id);

	if ((prop = node.property ("name")) == 0) {
		error << _("Session: XMLNode describing a Region is incomplete (no name)") << endmsg;
		return -1;
	}

	_name = prop->value();

	if ((prop = node.property ("start")) != 0) {
		_start = (jack_nframes_t) atoi (prop->value().c_str());
	}

	if ((prop = node.property ("length")) != 0) {
		_length = (jack_nframes_t) atoi (prop->value().c_str());
	}

	if ((prop = node.property ("position")) != 0) {
		_position = (jack_nframes_t) atoi (prop->value().c_str());
	}

	if ((prop = node.property ("layer")) != 0) {
		_layer = (layer_t) atoi (prop->value().c_str());
	}

	if ((prop = node.property ("flags")) != 0) {
		_flags = strtol (prop->value().c_str(), (char **) 0, 16);
	}

	if ((prop = node.property ("sync-offset")) != 0) {
		_sync_offset = (jack_nframes_t) atoi (prop->value().c_str());
	}

	
	if ((_flags & DefaultFadeIn) == 0) {

		/* Check for children that describe various curves */
		
		for (XMLNodeConstIterator niter = nlist.begin(); niter != nlist.end(); ++niter) {
			
			XMLNode *child;
			XMLProperty *childprop;
			
			child = (*niter);
			
			if (child->name() == "FadeIn") {
				
				if ((childprop = child->property ("steepness")) == 0) {
					error << _("Region XML node contains FadeIn, but no steepness") << endmsg;
					return -1;
				}
				_fade_in.set_steepness(atof (childprop->value().c_str()));
				
				if ((childprop = child->property ("length")) == 0) {
					error << _("Region XML node contains FadeIn, but no length") << endmsg;
					return -1;
				}
				_fade_in.set_length((jack_nframes_t) atoi (childprop->value().c_str()));
				
				build_fade (&_fade_in_data, _fade_in, true);
			}
		}
	}
	
	if ((_flags & DefaultFadeOut) == 0) {

		for (XMLNodeConstIterator niter = nlist.begin(); niter != nlist.end(); ++niter) {
			
			XMLNode *child;
			XMLProperty *childprop;
			
			child = (*niter);
			
			if (child->name() == "FadeOut") {
				
				if ((childprop = child->property ("steepness")) == 0) {
					error << _("Region XML node contains FadeOut, but no steepness") << endmsg;
					return -1;
				}
				_fade_out.set_steepness(atof (childprop->value().c_str()));
				
				if ((childprop = child->property ("length")) == 0) {
					error << _("Region XML node contains FadeOut, but no length") << endmsg;
					return -1;
				}
				_fade_out.set_length((jack_nframes_t) atoi (childprop->value().c_str()));

				build_fade (&_fade_out_data, _fade_out, false);
			}
		}
	}

	/* Now find envelope description and other misc child items */

	for (XMLNodeConstIterator niter = nlist.begin(); niter != nlist.end(); ++niter) {
		
		XMLNode *child;
		
		child = (*niter);
		
		if (child->name() == "Envelope") {
			
			XMLProperty *prop;
			double x, y;

			_envelope.clear ();

			if ((prop = child->property ("default")) != 0) {
				set_default_envelope ();
			} else {
			
				const XMLNodeList& elist = child->children();
				XMLNodeConstIterator i;
				
				

				for (i = elist.begin(); i != elist.end(); ++i) {
					
					if ((prop = (*i)->property ("x")) == 0) {
						error << compose(_("%1: no x-coordinate stored for control point (point ignored)"), _name) << endmsg;
						continue;
					}
					x = atof (prop->value().c_str());
					
					if ((prop = (*i)->property ("y")) == 0) {
						error << compose(_("%1: no y-coordinate stored for control point (point ignored)"), _name) << endmsg;
						continue;
					}
					y = atof (prop->value().c_str());

					_envelope.add_control_point (x, y);
				}

			}

			_envelope.set_max_xval (_length);
			
		} else if (child->name () == "extra") {
			_extra_xml = new XMLNode (*child);
			break;
		}
	}

	return 0;
}

void
Region::build_fade (gain_t **gain, LogCurve& curve, bool in)
{
	if (*gain) {
		delete [] *gain;
		*gain = 0;
	}

	if (curve.length() == 0 || curve.steepness() == 0) {
		return;
	}

	*gain = new gain_t[curve.length()];
	curve.fill (*gain, curve.length(), !in);
}

void
Region::set_default_fade (float steepness, jack_nframes_t len)
{
	default_fade_in.set_steepness (steepness);
	default_fade_in.set_length (len);
	default_fade_out.set_steepness (steepness);
	default_fade_out.set_length (len);

	build_fade (&default_fade_in_data, default_fade_in, true);
	build_fade (&default_fade_out_data, default_fade_out, false);
}

void
Region::freeze (FreezeFlag flag)
{
	LockMonitor lm (lock, __LINE__, __FILE__);
	if (flag & BoundsNotification) {
		_frozen[BoundsNotificationIndex]++;
	}
	if (flag & FadeNotification) {
		_frozen[FadeNotificationIndex]++;
	}
	if (flag & SyncNotification) {
		_frozen[SyncNotificationIndex]++;
	}
	if (flag & EnvelopeNotification) {
		_frozen[EnvelopeNotificationIndex]++;
	}
	if (flag & LayerNotification) {
		_frozen[LayerNotificationIndex]++;
	}
}

void
Region::thaw (FreezeFlag flag, const string& why, bool with_sig, void *src)
{
	FreezeFlag signals = FreezeFlag (0);

	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		
		if (flag & BoundsNotification && _frozen[BoundsNotificationIndex] && --_frozen[BoundsNotificationIndex] == 0 && with_sig) {
			signals = FreezeFlag (signals|BoundsNotification);
		}

		if (flag & FadeNotification && _frozen[FadeNotificationIndex] && --_frozen[FadeNotificationIndex] == 0 && with_sig) {
			signals = FreezeFlag (signals|FadeNotification);
		}
		if (flag & SyncNotification && _frozen[SyncNotificationIndex] && --_frozen[SyncNotificationIndex] == 0 && with_sig) {
			signals = FreezeFlag (signals|SyncNotification);
		}
		if (flag & EnvelopeNotification && _frozen[EnvelopeNotificationIndex] && --_frozen[EnvelopeNotificationIndex] == 0 && with_sig) {
			signals = FreezeFlag (signals|EnvelopeNotification);
		}
		if (flag & LayerNotification && _frozen[LayerNotificationIndex] && --_frozen[LayerNotificationIndex] == 0 && with_sig) {
			signals = FreezeFlag (signals|LayerNotification);
		}
	}

	if (signals == FreezeFlag (0)) {
		return;
	}
		
	if (src != this) {
		save_state (why);
	}

	if (!with_sig) {
		return;
	}

	if (signals & EnvelopeNotification) {
		EMIT_SIGNAL EnvelopeChanged();
	}

	if (signals & BoundsNotification) {
		maybe_uncopy ();
		EMIT_SIGNAL BoundsChanged (BoundsChangeInfo (StartChange|LengthChange|PositionChange), src);	
	}

	if (signals & FadeNotification) {
		/* XXX should really differentiate which one */
		EMIT_SIGNAL FadeInChanged ();
		EMIT_SIGNAL FadeOutChanged ();
	}

	if (signals & SyncNotification) {
		maybe_uncopy ();
		EMIT_SIGNAL SyncOffsetChanged ();
	}

	if (signals & LayerNotification) {
		EMIT_SIGNAL LayerChanged ();
	}
       
}

void
Region::set_default_envelope ()
{
	_envelope.clear ();
	_envelope.add_control_point (0.0, 1.0f);
	_envelope.add_control_point (_length, 1.0f);

	if (!hold_envelope_notification()) {
		EMIT_SIGNAL EnvelopeChanged ();
	}
}

void
Region::recompute_gain_at_end ()
{
	/* our length has changed. recompute a new final point by interpolating 
	   based on the the existing curve.
	*/
	
	_envelope.truncate_end (_length);
	
	_envelope.set_max_xval (_length);

	if (!hold_envelope_notification()) {
		EMIT_SIGNAL EnvelopeChanged ();
	}

}	

void
Region::recompute_gain_at_start ()
{
	/* as above, but the shift was from the front */

	_envelope.truncate_start (_length);

	if (!hold_envelope_notification()) {
		EMIT_SIGNAL EnvelopeChanged ();
	}
}

void
Region::set_fade_in_length (jack_nframes_t len)
{
	if (len == _fade_in.length()) {
		return;
	}

	_fade_in.set_length (len);
	if (_flags & DefaultFadeIn) {
		_fade_in.set_steepness (default_fade_in.steepness());
	}
	build_fade (&_fade_in_data, _fade_in, true);
	_flags &= ~DefaultFadeIn;

	if (!hold_fade_notification()) {
		char buf[64];
		sprintf (buf, "fade in length changed to %lu", len);
		save_state (buf);
		FadeInChanged ();
	}
}

void
Region::set_fade_out_length (jack_nframes_t len)
{
	if (len == _fade_out.length()) {
		return;
	}

	_fade_out.set_length (len);
	if (_flags & DefaultFadeOut) {
		_fade_out.set_steepness (default_fade_in.steepness());
	}
	build_fade (&_fade_out_data, _fade_out, false);
	_flags &= ~DefaultFadeOut;

	if (!hold_fade_notification()) {
		char buf[64];
		sprintf (buf, "fade out length changed to %lu", len);
		save_state (buf);
		FadeOutChanged ();
	}
}

void
Region::set_fade_in_steepness (float steep)
{
	if (steep == _fade_in.steepness()) {
		return;
	}

	_fade_in.set_steepness (steep);
	if (_flags & DefaultFadeIn) {
		_fade_in.set_length (default_fade_in.length());
	}
	build_fade (&_fade_in_data, _fade_in, true);
	_flags &= ~DefaultFadeIn;

	if (!hold_fade_notification()) {
		char buf[64];
		sprintf (buf, "fade in steepness changed to %f", steep);
		save_state (buf);
		FadeInChanged ();
	}
}

void
Region::set_fade_out_steepness (float steep)
{
	if (steep == _fade_out.steepness()) {
		return;
	}

	_fade_out.set_steepness (steep);
	if (_flags & DefaultFadeOut) {
		_fade_out.set_length (default_fade_out.length());
	}
	build_fade (&_fade_out_data, _fade_out, false);
	_flags &= ~DefaultFadeOut;

	if (!hold_fade_notification()) {
		char buf[64];
		sprintf (buf, "fade out steepness changed to %f", steep);
		save_state (buf);
		FadeOutChanged ();
	}
}

const LogCurve& 
Region::fade_in() const
{
	if (_flags & DefaultFadeIn) {
		return default_fade_in;
	} else {
		return _fade_in;
	}
}

const LogCurve& 
Region::fade_out() const
{
	if (_flags & DefaultFadeOut) {
		return default_fade_out;
	} else {
		return _fade_out;
	}
}

void
Region::set_fade_in_active (bool yn)
{
	if (yn == (_flags & FadeIn)) {
		return;
	}
	if (yn) {
		_flags |= FadeIn;
	} else {
		_flags &= ~FadeIn;
	}

	EMIT_SIGNAL FadeInActiveChanged();
}

void
Region::set_fade_out_active (bool yn)
{
	if (yn == (_flags & FadeOut)) {
		return;
	}
	if (yn) {
		_flags |= FadeOut;
	} else {
		_flags &= ~FadeOut;
	}

	EMIT_SIGNAL FadeOutActiveChanged();
}

int
Region::separate_by_channel (Session& session, vector<Region*>& v) const
{
	SourceList s, m;
	Region r (*this); // invoke pure copy constructor
	string new_name;

	/* now mess with its source list */
	
	s = r.sources;
	m = r.master_sources;

	for (SourceList::iterator i = s.begin(); i != s.end(); ++i) {

		r.sources.clear ();
		r.master_sources.clear();
		r.sources.push_back (*i);
		r.master_sources.push_back (*i);

		/* generate a new name */
		
		if (session.region_name (new_name, _name)) {
			return -1;
		}

		/* create a copy with just one source */

		v.push_back (new Region (r, new_name));
	}

	return 0;
}

extern "C" {

int region_read_peaks_from_c (void *arg, gulong npeaks, gulong start, gulong cnt, gpointer data, unsigned int n_chan) 
{
	return ((Region *) arg)->read_peaks ((PeakData *) data, (jack_nframes_t) npeaks, (jack_nframes_t) start, (jack_nframes_t) cnt, n_chan);
}

gulong region_length_from_c (void *arg)
{

	return ((Region *) arg)->length();
}

} /* extern "C" */
