/*
    Copyright (C) 1999-2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: session.cc,v 1.142 2003/06/26 02:56:41 pbd Exp $
*/

#include <algorithm>
#include <string>
#include <vector>
#include <sstream>
#include <fstream>

#include <sigc++/bind.h>

#include <cstdio> /* sprintf(3) ... grrr */
#include <cmath>
#include <cerrno>
#include <unistd.h>
#include <limits.h>
#include <libgen.h> // for dirname(3)

#include <ardour/timestamps.h>

#include <pbd/error.h>
#include <pbd/lockmonitor.h>
#include <pbd/pathscanner.h>
#include <pbd/stl_delete.h>

#include <ardour/audioengine.h>
#include <ardour/configuration.h>
#include <ardour/session.h>
#include <ardour/diskstream.h>
#include <ardour/utils.h>
#include <ardour/playlist.h>
#include <ardour/source.h>
#include <ardour/filesource.h>
#include <ardour/sndfilesource.h>
#include <ardour/auditioner.h>
#include <ardour/redirect.h>
#include <ardour/send.h>
#include <ardour/insert.h>
#include <ardour/connection.h>
#include <ardour/slave.h>
#include <ardour/tempo.h>
#include <ardour/audio_track.h>
#include <ardour/cycle_timer.h>
#include <ardour/named_selection.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

const char* Session::_template_suffix = ".template";
const char* Session::_statefile_suffix = ".ardour";

string
Session::find_session (string name, bool& isnew)
{
	PathScanner scanner;
	vector<string*>* results;
	string ret;

	/* handle absolute paths without any attempt to find them */

	if (name[0] == '/') {
		return name;
	} 

	/* handle explicit relative paths by canonicalizing the initial portion */

	if (name.length() > 2 && name[0] == '.' && name[1] == '/') {
		char buf[PATH_MAX+1];
		getcwd (buf, sizeof (buf));
		ret = buf;
		ret += name.substr (1);

		return ret;
	}

	if (name.length() > 3 && name[0] == '.' && name[1] == '.' && name[2] == '/') {
		char buf[PATH_MAX+1];
		getcwd (buf, sizeof (buf));
		ret = dirname (buf);
		ret += name.substr (2);

		return ret;
	}

	/* everything else is just a name and we have to use the Config
	   search path to look it up. 
	*/

	string regexp = "^";
	regexp += name;
	regexp += '$';

	results = scanner (Config->get_session_path(), regexp, false, true, 0);
	
	if (results) {

		if (!results->empty()) {

			ret = *results->front();
			isnew = false;

		} 

		vector_delete (results);
		delete results;
	}

	if (ret.length() == 0) {

		/* its a new file. just put it into cwd */

		ret = legalize_for_path (name);
		isnew = true;
	}

	return ret;
}

Session::Session (AudioEngine &eng,
		  string fullpath,
		  string name,
		  string snapshot_name,
		  string* mix_template)

	: _engine (eng),
	  _mmc_port (default_mmc_port),
	  _mtc_port (default_mtc_port),
	  pending_events (2048),
	  midi_requests (16)
{
	bool new_session;
	
	first_stage_init (fullpath, name, snapshot_name);

	if (create (new_session, mix_template)) {
		throw failed_constructor ();
	}

	if (second_stage_init (new_session)) {
		throw failed_constructor ();
	}

	store_in_recent_sessions ();

	_state_of_the_state = StateOfTheState (_state_of_the_state & ~Dirty);
}

Session::~Session ()
{
	_state_of_the_state = StateOfTheState (CannotSave|Deletion);
	_engine.remove_session ();

	EMIT_SIGNAL going_away ();

	terminate_butler_thread ();
	terminate_midi_thread ();

	if (click_data && click_data != default_click) {
		delete [] click_data;
	}

	if (click_emphasis_data && click_emphasis_data != default_click_emphasis) {
		delete [] click_emphasis_data;
	}

	clear_clicks ();

	for (vector<Sample*>::iterator i = _passthru_buffers.begin(); i != _passthru_buffers.end(); ++i) {
		delete [] *i;
	}

	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		(*i)->unref();
	}

	for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
		delete *i;
	}

	for (NamedSelectionList::iterator i = named_selections.begin(); i != named_selections.end(); ++i) {
		delete *i;
	}

	for (PlaylistList::iterator i = playlists.begin(); i != playlists.end(); ++i) {
		delete *i;
	}

	for (RegionList::iterator i = regions.begin(); i != regions.end(); ++i) {
		delete (*i).second;
	}
	
	for (SourceList::iterator i = sources.begin(); i != sources.end(); ++i) {
		delete (*i).second;
	}

	for (list<RouteGroup *>::iterator i = mix_groups.begin(); i != mix_groups.end(); ++i) {
		delete *i;
	}

	for (list<RouteGroup *>::iterator i = edit_groups.begin(); i != edit_groups.end(); ++i) {
		delete *i;
	}
	
	for (ConnectionList::iterator i = _connections.begin(); i != _connections.end(); ++i) {
		delete *i;
	}

	if (butler_mixdown_buffer) {
		delete [] butler_mixdown_buffer;
	}

	if (butler_gain_buffer) {
		delete [] butler_gain_buffer;
	}

	if (mmc) {
		delete mmc;
	}

	if (state_tree) {
		delete state_tree;
	}
}

void
Session::set_worst_physical_output_latency (bool take_lock)
{
	_worst_physical_output_latency = 0;

	if (!_engine.running()) {
		return;
	}

	if (take_lock) {
		route_lock.lock ();
	}
	
	for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
		for (unsigned long n = 0; n < (*i)->n_outputs(); ++n) {
			_worst_physical_output_latency = max (_worst_physical_output_latency, 
							      _engine.get_port_total_latency (*((*i)->output (n))));
		}
	}

	if (take_lock) {
		route_lock.unlock ();
	}
}

void
Session::when_engine_running ()
{
	string first_physical_output;

	_clicking = false;

	try {
		_click_io = new IO (*this, "click", 0, 0, -1, -1);

		first_physical_output = _engine.get_nth_physical_output (0);

		if (first_physical_output.length()) {
			if (_click_io->add_output_port (first_physical_output, this)) {
				error << compose (_("cannot connect click track to %1"), first_physical_output) << endmsg;
			} else {
				_clicking = click_requested;
			}
		}
	}

	catch (failed_constructor& err) {
		error << _("cannot register click track output port") << endmsg;
	}

	set_worst_physical_output_latency (true);

	if (_clicking) {
		EMIT_SIGNAL ControlChanged (Clicking);
	}

	if (auditioner == 0) {

		/* we delay creating the auditioner till now because
		   it makes its own connections to ports named
		   in the ARDOUR_RC config file. the engine has
		   to be running for this to work.
		*/

		try {
			auditioner = new Auditioner (*this);
		}

		catch (failed_constructor& err) {
			warning << _("cannot create Auditioner: no auditioning of regions possible") << endmsg;
		}
	}

	/* Create a set of Connection objects that map
	   to the physical outputs currently available
	*/

	int np_out = _engine.n_physical_outputs();
	int np_in = _engine.n_physical_inputs();

	for (int np = 0; np < np_out; ++np) {
		char buf[32];
		snprintf (buf, sizeof (buf), "out %d", np+1);

		Connection* c = new OutputConnection (buf, true);

		c->add_port ();
		c->add_connection (0, _engine.get_nth_physical_output (np));

		add_connection (c);
	}

	for (int np = 0; np < np_in; ++np) {
		char buf[32];
		snprintf (buf, sizeof (buf), "in %d", np+1);

		Connection* c = new InputConnection (buf, true);

		c->add_port ();
		c->add_connection (0, _engine.get_nth_physical_input (np));

		add_connection (c);
	}

	hookup_io ();

	_state_of_the_state = StateOfTheState (_state_of_the_state & ~(CannotSave|Dirty));

	/* hook us up to the engine */

	_engine.set_session (this);

	_state_of_the_state = Clean;
}

void
Session::hookup_io ()
{
	/* Tell all IO objects to create their ports */

	IO::enable_ports ();

	/* stop graph reordering notifications from
	   causing resorts, etc.
	*/

	_state_of_the_state = StateOfTheState (_state_of_the_state | InitialConnecting);

	/* Tell all IO objects to connect themselves together */

	IO::enable_connecting ();

	/* now handle the whole enchilada as if it was one
	   graph reorder event.
	*/

	_state_of_the_state = StateOfTheState (_state_of_the_state & ~InitialConnecting);

	graph_reordered ();

	/* update mixer solo state */
	catch_up_on_solo();
}

void
Session::playlist_length_changed (Playlist* pl)
{
	/* we can't just increase _current_end if pl->get_maximum_extent() 
	   if larger. if the playlist used to be the longest playlist,
	   and its now shorter, we have to decrease _current_end. hence,
	   we have to iterate over all diskstreams and check the 
	   playlists currently in use.
	*/
	find_current_end ();
}

void
Session::diskstream_playlist_changed (DiskStream* dstream)
{
	Playlist *playlist;

	if ((playlist = dstream->playlist()) != 0) {
		playlist->LengthChanged.connect (bind (slot (this, &Session::playlist_length_changed), playlist));
	}
	
	/* see comment in playlist_length_changed () */
	find_current_end ();
}

bool
Session::record_enabling_legal () const
{
	if (record_status() == Recording) {
		return false;
	}

	if (all_safe) {
		return false;
	}
	return true;
}

void
Session::set_auto_input (bool yn)
{
	if (auto_input != yn) {
		auto_input = yn;
		set_dirty();
		ControlChanged (AutoInput);
	}
}

void
Session::auto_punch_start_changed (Location* location)
{
	replace_event (Event::PunchIn, location->start());
}	

void
Session::auto_punch_end_changed (Location* location)
{
	jack_nframes_t when_to_stop = location->end();

	if (_worst_physical_output_latency > current_block_size) {
		when_to_stop += _worst_physical_output_latency;
	}

	replace_event (Event::PunchOut, when_to_stop);
}	

void
Session::auto_loop_start_changed (Location* location)
{
	replace_event (Event::AutoLoop, location->end(), location->start());
}	

void
Session::auto_loop_end_changed (Location* location)
{
	replace_event (Event::AutoLoop, location->end(), location->start());
}	

void
Session::set_auto_punch_location (Location* location)
{
	Location* existing;

	if ((existing = _locations.auto_punch_location()) != 0 && existing != location) {
		auto_punch_start_changed_connection.disconnect();
		auto_punch_end_changed_connection.disconnect();
		existing->set_auto_punch (false, this);
		remove_event (existing->start(), Event::PunchIn);
		remove_event (existing->end(), Event::PunchOut);
		auto_punch_location_changed (0);
	}

	set_dirty();

	if (location == 0) {
		return;
	}
	
	if (location->end() <= location->start()) {
		error << _("Session: you can't use that location for auto punch (start <= end)") << endmsg;
		return;
	}

	auto_punch_start_changed_connection.disconnect();
	auto_punch_end_changed_connection.disconnect();

	auto_punch_start_changed_connection = location->start_changed.connect (slot (this, &Session::auto_punch_start_changed));
	auto_punch_end_changed_connection = location->end_changed.connect (slot (this, &Session::auto_punch_end_changed));

	location->set_auto_punch (true, this);
	auto_punch_location_changed (location);
}

void
Session::set_punch_in (bool yn)
{
	if (punch_in == yn) {
		return;
	}

	Location* location;

	if ((location = _locations.auto_punch_location()) != 0) {
		if ((punch_in = yn) == true) {
			add_event (location->start(), Event::PunchIn);
		} else {
			remove_event (location->start(), Event::PunchIn);
		}
	}

	set_dirty();
	EMIT_SIGNAL ControlChanged (PunchIn); 
}

void
Session::set_punch_out (bool yn)
{
	if (punch_out == yn) {
		return;
	}

	Location* location;

	if ((location = _locations.auto_punch_location()) != 0) {
		if ((punch_out = yn) == true) {
			add_event (location->end(), Event::PunchOut);
		} else {
			remove_event (location->end(), Event::PunchOut);
		}
	}

	set_dirty();
	EMIT_SIGNAL ControlChanged (PunchOut); 
}

void
Session::set_auto_loop_location (Location* location)
{
	Location* existing;

	if ((existing = _locations.auto_loop_location()) != 0 && existing != location) {
		auto_loop_start_changed_connection.disconnect();
		auto_loop_end_changed_connection.disconnect();
		existing->set_auto_loop (false, this);
		remove_event (existing->end(), Event::AutoLoop);
		auto_loop_location_changed (0);
	}

	set_dirty();

	if (location == 0) {
		return;
	}

	if (location->end() <= location->start()) {
		error << _("Session: you can't use a mark for auto loop") << endmsg;
		return;
	}

	auto_loop_start_changed_connection.disconnect();
	auto_loop_end_changed_connection.disconnect();

	auto_loop_start_changed_connection = location->start_changed.connect (slot (this, &Session::auto_loop_start_changed));
	auto_loop_end_changed_connection = location->end_changed.connect (slot (this, &Session::auto_loop_end_changed));

	location->set_auto_loop (true, this);
	auto_loop_location_changed (location);
}


void
Session::locations_changed ()
{
	_locations.apply (*this, &Session::handle_locations_changed);
}

void
Session::handle_locations_changed (Locations::LocationList& locations)
{
	Locations::LocationList::iterator i;
	Location* location;
	bool set_loop = false;
	bool set_punch = false;

	for (i = locations.begin(); i != locations.end(); ++i) {

		location =* i;

		if (location->is_auto_punch()) {
			set_auto_punch_location (location);
			set_punch = true;
		}
		if (location->is_auto_loop()) {
			set_auto_loop_location (location);
			set_loop = true;
		}
		
	}

	if (!set_loop) {
		set_auto_loop_location (0);
	}
	if (!set_punch) {
		set_auto_punch_location (0);
	}

	set_dirty();
}						     

void
Session::maybe_enable_record ()
{
	atomic_set (&_record_status, Enabled);

	if (_transport_speed) {
		if (!punch_in) {
			enable_record ();
		} 
	} else {
		send_mmc_in_another_thread (MIDI::MachineControl::cmdRecordPause);
		EMIT_SIGNAL RecordEnabled ();
	}

	set_dirty();
}
jack_nframes_t
Session::audible_frame () const
{
	jack_nframes_t ret;
	extern jack_nframes_t tf_at_cycle_start;

	if (tf_at_cycle_start < _worst_physical_output_latency) {
		return 0;
	}

	ret = tf_at_cycle_start;

	if (_transport_speed && !non_realtime_work_pending()) {

		/* take latency into account */
		
		ret -= _worst_physical_output_latency;
	
		/* account for frames since the start of the last process() callback
		*/
				
		ret += (jack_nframes_t) floor (_engine.frames_since_cycle_start () * _transport_speed);

#if 0
		if (auto_loop) {
			
			Location* location;
			
			/* if we're not 1 block beyond the loop start,
			   we still playing the end of the loop.
			*/
				
			if ((location = _locations.auto_loop_location()) != 0) {

				if (ret >= location->start() && ret <= location->end()) {
					
					long delta = ret - location->start();
						
						
					if (delta <= 0) { /* haven't looped yet in terms of audible frames */
							
						ret = location->end() + delta;
							
					} 
				}

			} 
		} else 
#endif

		if (process_function == &Session::process_scrub) {

			/* scrubbing: last_stop_frame is irrelevant */

		} else {

			if (ret < last_stop_frame) {
				ret = last_stop_frame;
			}
		}
			
	}
		
	return ret;
}

void
Session::set_frame_rate (jack_nframes_t frames_per_second)
{
	/** \fn void Session::set_frame_size(jack_nframes_t)
		the AudioEngine object that calls this guarantees 
		that it will not be called while we are also in
		::process(). Its fine to do things that block
		here.
	*/

	_current_frame_rate = frames_per_second;
	_frames_per_smpte_frame = (jack_nframes_t) ceil ((double) _current_frame_rate / (double) smpte_frames_per_second);

	Route::set_automation_interval ((jack_nframes_t) ceil ((double) frames_per_second * 0.25));

	set_dirty();

	/* XXX need to reset/reinstantiate all LADSPA plugins */
}

void
Session::set_block_size (jack_nframes_t nframes)

{
	/** \fn void Session::set_block_size(jack_nframes_t)
		the AudioEngine object that calls this guarantees 
		that it will not be called while we are also in
		::process(). Its also fine to do things that block
		here.
	*/

	{ 
		LockMonitor lm (route_lock, __LINE__, __FILE__);
		
		current_block_size = nframes;

		for (vector<Sample*>::iterator i = _passthru_buffers.begin(); i != _passthru_buffers.end(); ++i) {
			delete [] *i;
			*i = new Sample[current_block_size];
		}
		
		for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
			(*i)->set_block_size (nframes);
		}

		for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
			(*i)->set_block_size (nframes);
		}

		set_worst_physical_output_latency (false);
	}

	set_dirty();
}

void
Session::set_default_fade (float steepness, float fade_msecs)
{
	jack_nframes_t fade_frames;
	
	/* Don't allow fade of less 1 frame */
	
	if (fade_msecs < (1000.0 * (1.0/_current_frame_rate))) {

		fade_msecs = 0;
		fade_frames = 0;

	} else {
		
		fade_frames = (jack_nframes_t) floor (fade_msecs * _current_frame_rate * 0.001);
		
	}

	default_fade_msecs = fade_msecs;
	default_fade_steepness = steepness;

	{
		LockMonitor lm (route_lock, __LINE__, __FILE__);
		Region::set_default_fade (steepness, fade_frames);
	}

	set_dirty();

	/* XXX have to do this at some point */
	/* foreach region using default fade, reset, then 
	   refill_all_diskstream_buffers ();
	*/
}

struct RouteSorter {
    bool operator() (Route* r1, Route* r2) {
	    if (r1->fed_by.find (r2) != r1->fed_by.end()) {
		    return false;
	    } else if (r2->fed_by.find (r1) != r2->fed_by.end()) {
		    return true;
	    } else {
		    return r1->order_key(N_("signal")) < r2->order_key(N_("signal"));
	    }
    }
};

static void
trace_terminal (Route* r1, Route* rbase)
{
	Route* r2;

	if ((r1->fed_by.find (rbase) != r1->fed_by.end()) && (rbase->fed_by.find (r1) != rbase->fed_by.end())) {
		error << compose(_("feedback loop setup between %1 and %2"), r1->name(), rbase->name()) << endmsg;
		return;
	}

	/* make a copy of the existing list of routes that feed r1 */

	set<Route *> existing = r1->fed_by;

	/* for each route that feeds r1, recurse, marking it as feeding
	   rbase as well.
	*/

	for (set<Route *>::iterator i = existing.begin(); i != existing.end(); ++i) {
		r2 =* i;

		/* r2 is a route that feeds r1 which somehow feeds base. mark
		   base as being fed by r2
		*/

		rbase->fed_by.insert (r2);

		if (r2 != rbase) {

			/* now recurse, so that we can mark base as being fed by
			   all routes that feed r2
			*/

			trace_terminal (r2, rbase);
		}

	}
}

void
Session::resort_routes (void* src)
{
	/* don't do anything here with signals emitted
	   by Routes while we are being destroyed.
	*/

	if (_state_of_the_state & Deletion) {
		return;
	}

	/* Caller MUST hold the route_lock */

	RouteList::iterator i, j;

	for (i = routes.begin(); i != routes.end(); ++i) {

		(*i)->fed_by.clear ();

		for (j = routes.begin(); j != routes.end(); ++j) {

			if ((*j)->feeds (*i)) {
				(*i)->fed_by.insert (*j);
			} 
		}
	}
	
	for (i = routes.begin(); i != routes.end(); ++i) {
		trace_terminal (*i,* i);
	}

	RouteSorter cmp;
	routes.sort (cmp);
}

AudioTrack*
Session::new_audio_track (int input_channels, int output_channels)
{
	AudioTrack *track;
	char track_name[32];
	unsigned long n = 0;
	string port;

	if (auditioner == 0) {

		/* we delay creating the auditioner till now because
		   it makes its own connections to ports named
		   in the ARDOUR_RC config file. the engine has
		   to be running for this to work.
		*/

		try {
			auditioner = new Auditioner (*this);
		}

		catch (failed_constructor& err) {
			warning << _("cannot create Auditioner: no auditioning of regions possible") << endmsg;
		}
	}
	
	/* count existing audio tracks */

	{
		LockMonitor lm (route_lock, __LINE__, __FILE__);
		for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
			if (dynamic_cast<AudioTrack*>(*i) != 0) {
				if (!(*i)->hidden()) {
					n++;
				}
			}
		}
	}

	/* check for duplicate route names, since we might have pre-existing
	   routes with this name (e.g. create Audio1, Audio2, delete Audio1,
	   save, close,restart,add new route - first named route is now
	   Audio2)
	*/

	do {
		snprintf (track_name, sizeof(track_name), "Audio %lu", n+1);
		if (route_by_name (track_name) == 0) {
			break;
		}
		n++;

	} while (n < ULONG_MAX);

	try {
		track = new AudioTrack (*this, track_name);
		IO& iio = track->input_io();

		for (int x = 0; iio.n_inputs() < (unsigned long) input_channels; ++x) {
			if (Config->get_auto_connect_ports()) {
				port = _engine.get_nth_physical_input (n+x);
			} else {
				port = "";
			}
			if (iio.add_input_port (port, this)) {
				break;
			}
		}

		IO& oio = track->output_io();
		for (int x = 0; oio.n_outputs() < (unsigned long) output_channels; ++x) {
			if (Config->get_auto_connect_ports()) {
				port = _engine.get_nth_physical_output (n+x);
			} else {
				port = "";
			}
			if (oio.add_output_port (port, this)) {
				break;
			}
		}

		track->diskstream_changed.connect (slot (this, &Session::resort_routes));
		add_route (*track);
	}

	catch (failed_constructor &err) {
		error << _("Session: could not create new audio track.") << endmsg;
		return 0;
	}

	return track;
}

Route*
Session::new_audio_route (int input_channels, int output_channels)
{
	Route *bus;
	char bus_name[32];
	unsigned long n = 0;
	string port;

	/* count existing audio busses */

	{
		LockMonitor lm (route_lock, __LINE__, __FILE__);
		for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
			if (dynamic_cast<AudioTrack*>(*i) == 0) {
				if (!(*i)->hidden()) {
					n++;
				}
			}
		}
	}

	do {
		snprintf (bus_name, sizeof(bus_name), "Bus %lu", n+1);
		if (route_by_name (bus_name) == 0) {
			break;
		}
		n++;

	} while (n < ULONG_MAX);

	try {
		bus = new Route (*this, bus_name, -1, -1, -1, -1);
		IO& iio = bus->input_io();

		for (int x = 0; iio.n_inputs() < (unsigned long) input_channels; ++x) {
			if (Config->get_auto_connect_ports()) {
				port = _engine.get_nth_physical_input (n+x);
			} else {
				port = "";
			}
			
			if (iio.add_input_port (port, this)) {
				break;
			}
		}

		IO& oio = bus->input_io();

		for (int x = 0; oio.n_outputs() < (unsigned long) output_channels; ++x) {
			if (Config->get_auto_connect_ports()) {
				port = _engine.get_nth_physical_output (n+x);
			} else {
				port = "";
			}

			if (oio.add_output_port (port, this)) {
				break;
			}
		}

		add_route (*bus);
	}

	catch (failed_constructor &err) {
		error << _("Session: could not create new route.") << endmsg;
		return 0;
	}

	return bus;
}

void
Session::add_route (Route& route)
{
	{ 
		LockMonitor lm (route_lock, __LINE__, __FILE__);
		routes.push_front (&route);
		resort_routes(0);
	}

	route.set_block_size (current_block_size);
	route.solo_changed.connect (bind (slot (*this, &Session::route_solo_changed), &route));
	route.output_changed.connect (slot (*this, &Session::set_worst_physical_output_latency_x));

	set_dirty();
	save_state (_current_snapshot_name);

	EMIT_SIGNAL RouteAdded (route);
}

void
Session::add_diskstream (DiskStream& dstream)
{
	{ 
		LockMonitor lm (diskstream_lock, __LINE__, __FILE__);
		diskstreams.push_back (&dstream);
	}

	/* take a reference to the diskstream, preventing it from
	   ever being deleted until the session itself goes away.
	*/

	dstream.ref();
	dstream.set_block_size (current_block_size);

	dstream.PlaylistChanged.connect (bind (slot (this, &Session::diskstream_playlist_changed), &dstream));
	dstream.playlist()->LengthChanged.connect (bind (slot (this, &Session::playlist_length_changed), dstream.playlist()));

	dstream.prepare ();

	set_dirty();
	save_state (_current_snapshot_name);

	EMIT_SIGNAL DiskStreamAdded (dstream);
}

void
Session::remove_route (Route& route)
{
	{ 	
		LockMonitor lm (route_lock, __LINE__, __FILE__);
		routes.remove (&route);
	}

	set_dirty();

	/* XXX should we disconnect from the Route's signals ? */

	save_state (_current_snapshot_name);

	delete &route;
}	

void
Session::route_solo_changed (void* src, Route* route)
{      
   if (solo_update_disabled) {
      // We know already
      return;
   }

   LockMonitor lm (route_lock, __LINE__, __FILE__);

   set<string> soloed_ports;

   if (route->soloed() && !_solo_latched) {

      // Unsolo all other routes that is related to this route.
      
      for (unsigned int p = 0; p < route->n_outputs(); ++p) {
         Port *port = route->output (p);
         const char **connections = port->get_connections ();
         if (connections) {
            for (int n = 0; connections[n]; ++n) {
               string portname = connections[n];
               strip_portname_for_solo (portname);
               soloed_ports.insert (portname);
            }
            free (connections);
         }
      }      

      for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
	if ((*i)->soloed() && (*i) != route &&
	    ((*i)->mix_group () == 0 ||
	     (*i)->mix_group () != route->mix_group ())) {
            for (unsigned int p = 0; p < (*i)->n_outputs (); ++p) {
               Port *port = (*i)->output (p);
               const char **connections = port->get_connections ();
               if (connections) {
                  for (int n = 0; connections[n]; ++n) {
                     string portname = connections[n];
                     strip_portname_for_solo (portname);
                     if (soloed_ports.find (portname) != soloed_ports.end ())
                     {
		        solo_update_disabled = true;
                        (*i)->set_solo (false, src);
			solo_update_disabled = false;
                        break;
                     }
                  }
                  free (connections);
               }
            }
         }
      }
   }

   update_route_solo_state();

   set_dirty();
}

void
Session::update_route_solo_state()
{
        // Find all soloed ports. Treat external ports as a group with
        // equal solo status.

        set<string> soloed_ports;

        for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
                if ((*i)->soloed()) {
                        for (unsigned int p = 0; p < (*i)->n_outputs(); ++p) {
                                Port *port = (*i)->output (p);
                                const char **connections = port->get_connections ();
                                if (connections) {
                                        for (int n = 0; connections[n]; ++n) {
                                                string portname = connections[n];
						strip_portname_for_solo (portname);
                                                soloed_ports.insert (portname);
                                        }
                                        free (connections);
                                }
                        }
                }
        }
   
        // Set solo_mute on routes based on soloed() and port solo status.

        for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
                if ((*i)->soloed()) {
                        (*i)->set_solo_mute(false);
                } else {
                        bool solo_mute = false;
                        // Route should be muted if it outputs to a soloed port
                        for (unsigned int p = 0; p < (*i)->n_outputs() && !solo_mute; ++p) {
                                Port *port = (*i)->output (p);
                                const char **connections = port->get_connections ();
                                if (connections) {
                                        for (int n = 0; connections[n]; ++n) {
                                                string portname = connections[n];
						strip_portname_for_solo (portname);
                                                if (soloed_ports.find (portname) != soloed_ports.end ())
                                                {
                                                        solo_mute = true;
                                                        break;
                                                }
                                        }
                                        free (connections);
                                }
                        }
                        (*i)->set_solo_mute (solo_mute);
                }
        }
}

void
Session::strip_portname_for_solo(string& portname)
{
	string host = portname.substr (0, portname.find(':'));

	if (host != "ardour") {
		portname = "external";
	} else {
		portname = portname.substr (0, portname.find("/"));
	}
}

Route *
Session::route_by_name (string id)

{
	LockMonitor lm (route_lock, __LINE__, __FILE__);

	for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {
		if ((*i)->name() == id) {
			return* i;
		}
	}

	return 0;
}

void
Session::find_current_end ()
{
	jack_nframes_t max = 0;
	jack_nframes_t me; 

	/* Don't take the diskstream lock. Caller must have other ways to
	   ensure atomicity.
	*/

	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		Playlist* pl = (*i)->playlist();
		if ((me = pl->get_maximum_extent()) > max) {
			max = me;
		}
	}

	if (max != current_end) {
		current_end = max;
		set_dirty();
		EMIT_SIGNAL DurationChanged();
	}
}

DiskStream *
Session::diskstream_by_name (string name)
{
	LockMonitor lm (diskstream_lock, __LINE__, __FILE__);

	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		if ((*i)->name() == name) {
			return* i;
		}
	}

	return 0;
}

DiskStream *
Session::diskstream_by_id (id_t id)
{
	LockMonitor lm (diskstream_lock, __LINE__, __FILE__);

	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		if ((*i)->id() == id) {
			return *i;
		}
	}

	return 0;
}

/* Region management */

int
Session::region_name (string& result, string base, bool newlevel) const

{
	char buf[16];
	string subbase;

	if (base == "") {
		
		LockMonitor lm (region_lock, __LINE__, __FILE__);

		sprintf (buf, "%d", (int)regions.size() + 1);

		
		result = "region.";
		result += buf;

	} else {

		/* XXX this is going to be slow. optimize me later */
		
		if (newlevel) {
			subbase = base;
		} else {
			string::size_type pos;

			if ((pos = base.find_last_of ('-')) == string::npos) {
				pos = base.find_last_of ('.');
			}

			/* pos may be npos, but then we just use entire base */

			subbase = base.substr (0, pos);
		}
		
		bool name_taken = true;
		
		{
			LockMonitor lm (region_lock, __LINE__, __FILE__);
			
			for (int n = 1; n < 5000; ++n) {
				
				result = subbase;
				sprintf (buf, ".%d", n);
				result += buf;
				
				name_taken = false;
				
				for (RegionList::const_iterator i = regions.begin(); i != regions.end(); ++i) {
					if ((*i).second->name() == result) {
						name_taken = true;
						break;
					}
				}
				
				if (!name_taken) {
					break;
				}
			}
		}
			
		if (name_taken) {
			fatal << compose(_("too many regions with names like %1"), base) << endmsg;
			/*NOTREACHED*/
		}
	}

	return 0;
}	

void
Session::add_region (Region* region)
{
	pair<RegionList::key_type, RegionList::mapped_type> entry;

	{ 
		LockMonitor lm (region_lock, __LINE__, __FILE__);

		entry.first = region->id();
		entry.second = region;

		pair<RegionList::iterator,bool> x = regions.insert (entry);

		if (!x.second) {
			/* this happens when regions have been copied between
			   playlists without a new ID being assigned. it
			   might be a dumb idea, but its not actually
			   that much a problem and its not up to the Session
			   to stop it.
			*/
			return;
		}
	}
	
	region->GoingAway.connect (slot (this, &Session::remove_region));
	set_dirty();
	
	EMIT_SIGNAL RegionAdded (region);
}

void
Session::remove_region (Region* region)
{
	RegionList::iterator i;

	{ 
		LockMonitor lm (region_lock, __LINE__, __FILE__);

		if ((i = regions.find (region->id())) != regions.end()) {
			regions.erase (i);
		}
	}

	set_dirty();

	EMIT_SIGNAL RegionRemoved(region);
}

Region *
Session::get_region (id_t id)

{
	RegionList::iterator i;
	Region* region = 0;
	LockMonitor lm (region_lock, __LINE__, __FILE__);

	if ((i = regions.find (id)) != regions.end()) {
		region = (*i).second;
	}

	return region;
}

/* Source Management */

void
Session::add_source (Source* source)
{
	pair<SourceList::key_type, SourceList::mapped_type> entry;

	{
		LockMonitor lm (source_lock, __LINE__, __FILE__);
		entry.first = source->id();
		entry.second = source;
		sources.insert (entry);
	}
	
	source->GoingAway.connect (slot (this, &Session::remove_source));
	set_dirty();
	
	EMIT_SIGNAL SourceAdded (source);
}

void
Session::remove_source (Source* source)

{
	SourceList::iterator i;

	{ 
		LockMonitor lm (source_lock, __LINE__, __FILE__);

		if ((i = sources.find (source->id())) != sources.end()) {
			sources.erase (i);
		}
	}

	set_dirty();

	EMIT_SIGNAL SourceRemoved(source);
}

Source *
Session::get_source (ARDOUR::id_t id)
{
	LockMonitor lm (source_lock, __LINE__, __FILE__);
	SourceList::iterator i;
	Source* source = 0;

	if ((i = sources.find (id)) != sources.end()) {
		source = (*i).second;
	}

	return source;
}

FileSource *
Session::create_file_source (DiskStream& ds)
{
	string spath;
	unsigned int cnt;
	char buf[PATH_MAX+1];
	const unsigned int limit = 10000;
	string legalized;

	buf[0] = '\0';
	legalized = legalize_for_path (ds.name());

	/* find a "version" of the file name that doesn't exist in
	   any of the possible directories.
	*/
	
	for (cnt = 1; cnt <= limit; ++cnt) {

		vector<space_and_path>::iterator i;

		for (i = session_dirs.begin(); i != session_dirs.end(); ++i) {

			spath = (*i).path;
			spath += "/sounds/";
			spath += legalized;
			
			snprintf (buf, sizeof(buf), "%s-%u.wav", spath.c_str(), cnt);

			if (access (buf, F_OK)) {
				break;
			}
		}

		if (i != session_dirs.end()) {
			break;
		}
	}

	if (cnt > limit) {
		error << compose(_("There are already %1 recordings for %2, which I consider too many."), limit, ds.name()) << endmsg;
		throw failed_constructor();
	}

	/* we now have a unique name for the file, but figure out where to
	   actually put it.
	*/

	spath = discover_best_sound_dir ();

	spath += '/';
	spath += legalized;
	snprintf (buf, sizeof(buf), "%s-%u.wav", spath.c_str(), cnt);

	/* this might throw failed_constructor(), which is OK */

	return new FileSource (buf, frame_rate());
}

/* Playlist management */

Playlist *
Session::get_playlist (string name)
{
	Playlist* ret = 0;

	if ((ret = playlist_by_name (name)) == 0) {
		ret = new Playlist (*this, name);
	}

	return ret;
}

Playlist *
Session::playlist_by_name (string name)
{
	LockMonitor lm (playlist_lock, __LINE__, __FILE__);
	for (PlaylistList::iterator i = playlists.begin(); i != playlists.end(); ++i) {
		if ((*i)->name() == name) {
			return* i;
		}
	}
	return 0;
}

void
Session::add_playlist (Playlist* playlist)

{
	if (playlist->hidden()) {
		return;
	}

	{ 
		LockMonitor lm (playlist_lock, __LINE__, __FILE__);
		playlists.insert (playlists.begin(), playlist);
		playlist->ref();
	}

	set_dirty();

	EMIT_SIGNAL PlaylistAdded (playlist);
}

void
Session::remove_playlist (Playlist* playlist)
{
	{ 
		LockMonitor lm (playlist_lock, __LINE__, __FILE__);

		PlaylistList::iterator i = find (playlists.begin(), playlists.end(), playlist);

		if (i != playlists.end()) {
			playlists.erase (i);
			playlist->unref();
		}

	}

	set_dirty();

	EMIT_SIGNAL PlaylistRemoved (playlist);
}

void 
Session::set_audition (Region* r)
{
	pending_audition_region = r;
	post_transport_work = PostTransportWork (post_transport_work | PostTransportAudition);
	schedule_butler_transport_work ();
}

void
Session::non_realtime_set_audition ()
{
	if (pending_audition_region) {
		auditioner->audition_region (*pending_audition_region);
		pending_audition_region = 0;
		EMIT_SIGNAL AuditionActive (true);
	} 
}

void
Session::audition_region (Region& r)
{
	Event* ev = new Event (Event::Audition, Event::Add, Event::Immediate, 0, 0.0);
	ev->set_ptr (&r);
	queue_event (ev);
}

void
Session::cancel_audition ()
{
	if (auditioner->active()) {
		auditioner->cancel_audition ();
		EMIT_SIGNAL AuditionActive (false);
	}
}

bool
Session::RoutePublicOrderSorter::operator() (Route* a, Route* b)
{
	return a->order_key(N_("signal")) < b->order_key(N_("signal"));
}

void
Session::remove_empty_sounds ()
{

	PathScanner scanner;
	string dir;

	dir = sound_dir ();

	vector<string *>* possible_audiofiles = scanner (dir, "\\.wav$", false, true);
	
	for (vector<string *>::iterator i = possible_audiofiles->begin(); i != possible_audiofiles->end(); ++i) {

		if (FileSource::is_empty (*(*i))) {

			unlink ((*i)->c_str());
			
			string peak_path = peak_path_from_audio_path (**i);
			unlink (peak_path.c_str());
		}

		delete* i;
	}

	delete possible_audiofiles;
}

bool
Session::is_auditioning () const
{
	return auditioner->active();
}

int
Session::flatten ()
{
	LockMonitor lm (route_lock, __LINE__, __FILE__);

	for (RouteList::iterator i = routes.begin(); i != routes.end(); ++i) {

		AudioTrack *at;

		if ((at = dynamic_cast<AudioTrack*>(*i)) != 0) {
			if (flatten_one_track (*at, 0, current_end)) {
				return -1;
			}
		}
	}

	return 0;
}

int
Session::flatten_whole_track (AudioTrack& track)
{
	Playlist* playlist = track.disk_stream().playlist();

	if (playlist == 0) {
		return -1;
	}
				
	return flatten_one_track (track, 0, playlist->get_maximum_extent());
}

int
Session::flatten_track (AudioTrack& track, jack_nframes_t start, jack_nframes_t len)
{
	LockMonitor lm (route_lock, __LINE__, __FILE__);
	return flatten_one_track (track, start, len);
}

int
Session::flatten_one_track (AudioTrack& track, jack_nframes_t start, jack_nframes_t len)
{
	int ret = -1;
	Playlist* playlist;
	FileSource* fsource;
	Region* region;
	unsigned long x;
	char buf[PATH_MAX+1];
	string dir;
	Region::SourceList srcs;
	unsigned int nchans;
	
	/* call tree *MUST* hold route_lock */
	
	if ((playlist = track.disk_stream().playlist()) == 0) {
		return -1;
	}

	nchans = track.disk_stream().n_channels();
	
	dir = discover_best_sound_dir ();

	for (unsigned int chan_n=0; chan_n < nchans; ++chan_n) {

		for (x = 0; x < 99999; ++x) {
			snprintf (buf, sizeof(buf), "%s/%s-%d-bounce-%lu.wav", dir.c_str(), playlist->name().c_str(), chan_n, x+1);
			if (access (buf, F_OK) != 0) {
				break;
			}
		}
		
		if (x == 99999) {
			error << compose (_("too many bounced versions of playlist \"%1\""), playlist->name()) << endmsg;
			return -1;
		}
		
		try {
			fsource = new FileSource (buf, frame_rate());
		}
		
		catch (failed_constructor& err) {
			error << compose (_("cannot create new audio file \"%1\" for %2"), buf, track.name()) << endmsg;
			return -1;
		}

		srcs.push_back(fsource);
	}
	
	jack_nframes_t position = start;
	jack_nframes_t this_chunk;
	jack_nframes_t to_do = len;
	
	while (to_do) {
		
		this_chunk = min (to_do, current_block_size);
		
		if (track.export_stuff (_passthru_buffers, start, this_chunk)) {
			goto out;
		}

		unsigned int n=0;
		for (Region::SourceList::iterator src=srcs.begin(); src != srcs.end(); ++src, ++n) {
			if ((*src)->write (_passthru_buffers[n], this_chunk) != this_chunk) {
				goto out;
			}
		}
		
		start += this_chunk;
		to_do -= this_chunk;
	}
	
	time_t now;
	struct tm* xnow;
	time (&now);
	xnow = localtime (&now);

	for (Region::SourceList::iterator src=srcs.begin(); src != srcs.end(); ++src) {
		dynamic_cast<FileSource*>((*src))->update_header (position, *xnow, now);
	}
	

	/* create a new region from all filesources */

	region = new Region (srcs, 0, srcs[0]->length(), playlist->name(), 0, (Region::Flag) (Region::WholeFile|Region::DefaultFlags));
	
	playlist->clear ();
	playlist->add_region (*region, position);

	/* build peakfile for new source */

	for (Region::SourceList::iterator src=srcs.begin(); src != srcs.end(); ++src) {
		dynamic_cast<FileSource*>(*src)->build_peaks ();
	}

	set_dirty();

	ret = 0;

  out:
	if (ret) {
		for (Region::SourceList::iterator src=srcs.begin(); src != srcs.end(); ++src) {
			dynamic_cast<FileSource*>(*src)->mark_for_remove ();
			dynamic_cast<FileSource*>(*src)->unref();
		}
	}

	return ret;
}

string
Session::peak_path_from_audio_path (string audio_path)
{
	/* XXX hardly bombproof! fix me */

	string res = audio_path.substr (0, audio_path.find_last_of ('.'));
	res += ".peak";
	return res;
}

void
Session::catch_up_on_solo ()
{
	/* this is called after set_state() to catch the full solo
	   state, which can't be correctly determined on a per-route
	   basis, but needs the global overview that only the session
	   has.
	*/
        LockMonitor lm (route_lock, __LINE__, __FILE__);
	update_route_solo_state();
}	
		
void
Session::set_all_solo (bool yn)
{
	/* XXX this copy is not safe: the Routes within the list
	   can still be deleted after the Route lock is released.
	*/

	RouteList copy;
	{
		LockMonitor lm (route_lock, __LINE__, __FILE__);
		copy = routes;
	}

	for (RouteList::iterator i = copy.begin(); i != copy.end(); ++i) {
		if (!(*i)->hidden()) {
			(*i)->set_solo (yn, this);
		}
	}

	set_dirty();
}
		
void
Session::set_all_mute (bool yn)
{
	RouteList copy;
	{
		LockMonitor lm (route_lock, __LINE__, __FILE__);
		copy = routes;
	}

	for (RouteList::iterator i = copy.begin(); i != copy.end(); ++i) {
		if (!(*i)->hidden()) {
			(*i)->set_mute (yn, this);
		}
	}

	set_dirty();
}
		
unsigned long
Session::n_diskstreams () const
{
	LockMonitor lm (diskstream_lock, __LINE__, __FILE__);
	unsigned long n = 0;

	for (DiskStreamList::const_iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		if (!(*i)->hidden()) {
			n++;
		}
	}
	return n;
}

void 
Session::foreach_diskstream (void (DiskStream::*func)(void)) 
{
	LockMonitor lm (diskstream_lock, __LINE__, __FILE__);
	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		if (!(*i)->hidden()) {
			((*i)->*func)();
		}
	}
}

void
Session::graph_reordered ()
{
	/* don't do this stuff is we are setting up connections
	   from a set_state() call.
	*/

	if (_state_of_the_state & InitialConnecting) {
		return;
	}

	LockMonitor lm1 (route_lock, __LINE__, __FILE__);
	LockMonitor lm2 (diskstream_lock, __LINE__, __FILE__);

	resort_routes (0);

	/* force all diskstreams to update their capture offset values to 
	   reflect any changes in latencies within the graph.
	*/
	
	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		(*i)->set_capture_offset ();
	}
}

void
Session::record_disenable_all ()
{
	record_enable_change_all (false);
}

void
Session::record_enable_all ()
{
	record_enable_change_all (true);
}

void
Session::record_enable_change_all (bool yn)
{
	LockMonitor lm1 (route_lock, __LINE__, __FILE__);
	LockMonitor lm2 (diskstream_lock, __LINE__, __FILE__);
	
	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		if (!(*i)->hidden() && (*i)->recordable()) {
			(*i)->set_record_enabled (yn, this);
		}
	}

	set_dirty();
}

void
Session::add_redirect (Redirect* redirect)
{
	Send* send;
	Insert* insert;
	PortInsert* port_insert;
	PluginInsert* plugin_insert;

	if ((insert = dynamic_cast<Insert *> (redirect)) != 0) {
		if ((port_insert = dynamic_cast<PortInsert *> (insert)) != 0) {
			_port_inserts.insert (_port_inserts.begin(), port_insert);
			insert_cnt = _port_inserts.size();
		} else if ((plugin_insert = dynamic_cast<PluginInsert *> (insert)) != 0) {
			_plugin_inserts.insert (_plugin_inserts.begin(), plugin_insert);
		} else {
			fatal << _("programming error: unknown type of Insert created!") << endmsg;
			/*NOTREACHED*/
		}
	} else if ((send = dynamic_cast<Send *> (redirect)) != 0) {
		_sends.insert (_sends.begin(), send);
		send_cnt = _sends.size();
	} else {
		fatal << _("programming error: unknown type of Redirect created!") << endmsg;
		/*NOTREACHED*/
	}

	redirect->GoingAway.connect (slot (*this, &Session::remove_redirect));

	set_dirty();
}

void
Session::remove_redirect (Redirect* redirect)
{
	Send* send;
	Insert* insert;
	PortInsert* port_insert;
	PluginInsert* plugin_insert;

	if ((insert = dynamic_cast<Insert *> (redirect)) != 0) {
		if ((port_insert = dynamic_cast<PortInsert *> (insert)) != 0) {
			_port_inserts.remove (port_insert);
			insert_cnt = _port_inserts.size();
		} else if ((plugin_insert = dynamic_cast<PluginInsert *> (insert)) != 0) {
			_plugin_inserts.remove (plugin_insert);
		} else {
			fatal << _("programming error: unknown type of Insert deleted!") << endmsg;
			/*NOTREACHED*/
		}
	} else if ((send = dynamic_cast<Send *> (redirect)) != 0) {
		_sends.remove (send);
		send_cnt = _sends.size();
	} else {
		fatal << _("programming error: unknown type of Redirect deleted!") << endmsg;
		/*NOTREACHED*/
	}

	set_dirty();
}

jack_nframes_t
Session::available_capture_duration ()
{
	const double scale = 4096.0 / sizeof (Sample);
	
	if (_total_free_4k_blocks * scale > (double) max_frames) {
		return max_frames;
	}
	
	return (jack_nframes_t) floor (_total_free_4k_blocks * scale);
}

void
Session::add_connection (ARDOUR::Connection* connection)
{
	{
		LockMonitor (connection_lock, __LINE__, __FILE__);
		_connections.push_back (connection);
	}

	EMIT_SIGNAL ConnectionAdded (connection);

	set_dirty();
}

void
Session::remove_connection (ARDOUR::Connection* connection)
{
	bool removed = false;

	{
		LockMonitor (connection_lock, __LINE__, __FILE__);
		ConnectionList::iterator i = find (_connections.begin(), _connections.end(), connection);
		
		if (i != _connections.end()) {
			_connections.erase (i);
			removed = true;
		}
	}

	if (removed) {
		EMIT_SIGNAL ConnectionRemoved (connection);
	}

	set_dirty();
}

ARDOUR::Connection *
Session::connection_by_name (string name) const
{
	LockMonitor lm (connection_lock, __LINE__, __FILE__);

	for (ConnectionList::const_iterator i = _connections.begin(); i != _connections.end(); ++i) {
		if ((*i)->name() == name) {
			return* i;
		}
	}

	return 0;
}

void
Session::set_edit_mode (EditMode mode)
{
	_edit_mode = mode;
	
	{ 
		LockMonitor lm (playlist_lock, __LINE__, __FILE__);
		
		for (PlaylistList::iterator i = playlists.begin(); i != playlists.end(); ++i) {
			(*i)->set_edit_mode (mode);
		}
	}

	EMIT_SIGNAL ControlChanged (EditingMode);
}

void
Session::tempo_map_changed ()
{
	clear_clicks ();
	set_dirty ();
}

void
Session::ensure_passthru_buffers (unsigned long howmany)
{
	while (howmany > _passthru_buffers.size()) {
		_passthru_buffers.push_back (new Sample[current_block_size]);
	}
}

string
Session::next_send_name ()
{
	char buf[32];
	snprintf (buf, sizeof (buf), "send %lu", send_cnt + 1);
	return buf;
}

string
Session::next_insert_name ()
{
	char buf[32];
	snprintf (buf, sizeof (buf), "insert %lu", insert_cnt + 1);
	return buf;
}

/* Named Selection management */

NamedSelection *
Session::named_selection_by_name (string name)
{
	LockMonitor lm (named_selection_lock, __LINE__, __FILE__);
	for (NamedSelectionList::iterator i = named_selections.begin(); i != named_selections.end(); ++i) {
		if ((*i)->name == name) {
			return* i;
		}
	}
	return 0;
}

void
Session::add_named_selection (NamedSelection* named_selection)
{
	{ 
		LockMonitor lm (named_selection_lock, __LINE__, __FILE__);
		named_selections.insert (named_selections.begin(), named_selection);
	}

	set_dirty();

	EMIT_SIGNAL NamedSelectionAdded ();
}

void
Session::remove_named_selection (NamedSelection* named_selection)
{
	bool removed = false;

	{ 
		LockMonitor lm (named_selection_lock, __LINE__, __FILE__);

		NamedSelectionList::iterator i = find (named_selections.begin(), named_selections.end(), named_selection);

		if (i != named_selections.end()) {
			delete (*i);
			named_selections.erase (i);
			set_dirty();
			removed = true;
		}
	}

	if (removed) {
		EMIT_SIGNAL NamedSelectionRemoved ();
	}
}

void
Session::reset_native_file_format ()
{
	LockMonitor lm1 (route_lock, __LINE__, __FILE__);
	LockMonitor lm2 (diskstream_lock, __LINE__, __FILE__);

	for (DiskStreamList::iterator i = diskstreams.begin(); i != diskstreams.end(); ++i) {
		(*i)->reset_write_source ();
	}
}

bool
Session::route_name_unique (string n) const
{
	LockMonitor lm (route_lock, __LINE__, __FILE__);
	
	for (RouteList::const_iterator i = routes.begin(); i != routes.end(); ++i) {
		if ((*i)->name() == n) {
			return false;
		}
	}
	
	return true;
}

int
Session::read_recent_sessions (RecentSessions& rs)
{
	string path;
	char* envvar;
	
	if ((envvar = getenv ("HOME")) == 0 || strlen (envvar) == 0) {
		return -1;
	}
		
	path = envvar;
	path += "/.ardour/recent";

	ifstream recent (path.c_str());
	
	if (!recent) {
		if (errno != ENOENT) {
			error << compose (_("cannot open recent session file %1 (%2)"), path, strerror (errno)) << endmsg;
			return -1;
		} else {
			return 1;
		}
	}

	while (true) {

		pair<string,string> newpair;

		recent >> newpair.first;

		if (!recent.good()) {
			break;
		}
		
		recent >> newpair.second;

		if (!recent.good()) {
			break;
		}

		rs.insert (newpair);
	}

	return 0;
}

int
Session::write_recent_sessions (RecentSessions& rs)
{
	string path;
	char* envvar;
	
	if ((envvar = getenv ("HOME")) == 0 || strlen (envvar) == 0) {
		return -1;
	}
		
	path = envvar;
	path += "/.ardour/recent";

	ofstream recent (path.c_str());

	if (!recent) {
		return -1;
	}

	for (RecentSessions::iterator i = rs.begin(); i != rs.end(); ++i) {
		recent << (*i).first << ' ' << (*i).second << endl;
	}
	
	return 0;
}
	
int
Session::store_in_recent_sessions ()
{
	RecentSessions rs;

	if (read_recent_sessions (rs) < 0) {
		return -1;
	}

	if (rs.find (_name) == rs.end()) {
		
		pair<string,string> newpair;

		newpair.first = _name;
		newpair.second = _path;
		
		rs.insert (newpair);

		return write_recent_sessions (rs);
	} 

	return 0;
}

