/*
    Copyright (C) 1999-2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: session_time.cc,v 1.6 2003/05/07 18:18:06 pbd Exp $
*/

#include <cmath>
#include <unistd.h>

#include <ardour/timestamps.h>

#include <pbd/error.h>

#include <ardour/ardour.h>
#include <ardour/configuration.h>
#include <ardour/audioengine.h>
#include <ardour/session.h>
#include <ardour/tempo.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

/* BBT TIME*/

void
Session::bbt_time (jack_nframes_t when, BBT_Time& bbt)
{
	_tempo_map->bbt_time (when, bbt);
}

/* SMPTE TIME */


void
Session::set_smpte_offset (jack_nframes_t off)
{
	_smpte_offset = off;
	last_smpte_valid = false;
	EMIT_SIGNAL SMPTEOffsetChanged ();
}

void
Session::smpte_time (jack_nframes_t when, SMPTE_Time& smpte)
{
	if (last_smpte_valid && when == last_smpte_when) {
		smpte = last_smpte;
		return;
	}

	double offset_sample;
	double secs;
	double computed_secs;
	double total_secs;
	long hrs;
	long mins;
	
	if (when < _smpte_offset) {
		offset_sample = ((double) _smpte_offset - when);
		smpte.negative = true;
	} else {
		offset_sample = (double) when - _smpte_offset;
		smpte.negative = false;
	}

	total_secs = offset_sample / frame_rate();
	secs = total_secs;
	hrs =  (long) floor (secs / 3600.0);
	secs -= (hrs * 3600.0);
	mins = (long) floor (secs / 60.0);
	secs -= (mins * 60.0);

	computed_secs = (hrs * 3600.0) + (mins * 60.0) + floor (secs);

	smpte.frames = (long) ((total_secs - computed_secs) * smpte_frames_per_second);
	smpte.seconds = (long) secs;
	smpte.minutes = mins;
	smpte.hours = hrs;

	last_smpte_when = when;
	last_smpte = smpte;
	last_smpte_valid = true;
}

void
Session::smpte_time_subframes (jack_nframes_t when, SMPTE_Time& smpte)
{
	if (last_smpte_valid && when == last_smpte_when) {
		smpte = last_smpte;
		return;
	}

	double offset_sample;
	double secs;
	double computed_secs;
	double total_secs;
	long hrs;
	long mins;
	
	if (when < _smpte_offset) {
		offset_sample = ((double) _smpte_offset - when);
		smpte.negative = true;
	} else {
		offset_sample = (double) when - _smpte_offset;
		smpte.negative = false;
	}

	total_secs = offset_sample / frame_rate();
	secs = total_secs;
	hrs =  (long) floor (secs / 3600.0);
	secs -= (hrs * 3600.0);
	mins = (long) floor (secs / 60.0);
	secs -= (mins * 60.0);

	computed_secs = (hrs * 3600.0) + (mins * 60.0) + floor (secs);

	smpte.frames = (long) ((total_secs - computed_secs) * smpte_frames_per_second);

	/* XXX not sure how to compute this yet */
	
	smpte.subframes = 0;

	smpte.seconds = (long) secs;
	smpte.minutes = mins;
	smpte.hours = hrs;

	last_smpte_when = when;
	last_smpte = smpte;
	last_smpte_valid = true;
}

void
Session::smpte_duration (jack_nframes_t when, SMPTE_Time& smpte) const
{
	double sample;
	double secs;
	double computed_secs;
	double approx_secs;
	long hrs;
	long mins;
	
	sample = when;

	approx_secs = (sample / frame_rate());
	secs = approx_secs;
	hrs =  (long) floor (secs / 3600.0);
	secs -= (hrs * 3600.0);
	mins = (long) floor (secs / 60.0);
	secs -= (mins * 60.0);

	computed_secs = (hrs * 3600.0) + (mins * 60.0) + floor (secs);

	smpte.frames = (long) ((approx_secs - computed_secs) * smpte_frames_per_second);
	smpte.seconds = (long) secs;
	smpte.minutes = mins;
	smpte.hours = hrs;
}

void
Session::smpte_duration_string (char* buf, jack_nframes_t when) const
{
	SMPTE_Time smpte;

	smpte_duration (when, smpte);
	sprintf (buf, "%02ld:%02ld:%02ld:%02ld", smpte.hours, smpte.minutes, smpte.seconds, smpte.frames);
}

void
Session::smpte_time (SMPTE_Time &t)

{
	smpte_time (_transport_frame, t);
}

void
Session::set_jack_timebase ()
{
	jack_transport_info_t t;
	BBT_Time bbt;

	/* frame info */

	t.frame = _transport_frame;

	/* SMPTE info */

	t.smpte_offset = _smpte_offset;
	t.smpte_frame_rate = smpte_frames_per_second;

	t.valid = jack_transport_bits_t (JackTransportPosition|JackTransportSMPTE|JackTransportState);

	/* BBT info */
	
	if (_tempo_map) {

		TempoMap::Metric metric (_tempo_map->metric_at (_transport_frame));
		_tempo_map->bbt_time_with_metric (_transport_frame, bbt, metric);
	
		t.bar = bbt.bars;
		t.beat = bbt.beats;
		t.tick = bbt.ticks;

		// XXX still need to set bar_start_tick

		t.beats_per_bar = metric.meter().beats_per_bar();
		t.beat_type = metric.meter().note_divisor();
		t.ticks_per_beat = Meter::ticks_per_beat;
		t.beats_per_minute = metric.tempo().beats_per_minute();

		t.valid = jack_transport_bits_t (t.valid | JackTransportBBT);
	}

	if (_transport_speed) {

		if (auto_loop) {

			Location* location = _locations.auto_loop_location();

			if (location) {

				t.transport_state = JackTransportLooping;
				t.loop_start = location->start();
				t.loop_end = location->end();
				t.valid = jack_transport_bits_t (t.valid | JackTransportLoop);

			} else {

				t.loop_start = 0;
				t.loop_end = 0;
				t.transport_state = JackTransportRolling;

			}

		} else {

			t.loop_start = 0;
			t.loop_end = 0;
			t.transport_state = JackTransportRolling;

		}

	} else {
		t.transport_state = JackTransportStopped;
	}
		
	_engine.set_transport_info (t);
}

jack_nframes_t
Session::convert_to_frames_at (jack_nframes_t position, AnyTime& any)
{
	double secs;
	
	switch (any.type) {
	case AnyTime::BBT:
		return _tempo_map->frame_time (position, any.bbt);
		break;

	case AnyTime::SMPTE:
		/* XXX need to handle negative values */
		secs = any.smpte.hours * 60 * 60;
		secs += any.smpte.minutes * 60;
		secs += any.smpte.seconds;
		secs += any.smpte.frames / smpte_frames_per_second;
		return _smpte_offset + (jack_nframes_t) floor (secs / frame_rate());
		break;

	case AnyTime::Seconds:
		return (jack_nframes_t) floor (any.seconds / frame_rate());
		break;

	case AnyTime::Frames:
		return any.frames;
		break;
	}

	return any.frames;
}
