/*
    Copyright (C) 2003 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: session_timefx.cc,v 1.11 2003/06/26 19:50:58 essej Exp $
*/

#include <errno.h>

#include <soundtouch/SoundTouch.h>
#include <ardour/session.h>
#include <ardour/region.h>
#include <ardour/filesource.h>

#include "i18n.h"

using namespace ARDOUR;

Region*
Session::tempoize_region (TimeStretchRequest& tsr)
{
	const jack_nframes_t bufsize = 16384;
	gain_t gain_buffer[bufsize];
	Sample buffer[bufsize];
	FileSource* new_file = 0;
	float percentage;
	Region* r = 0;
	jack_nframes_t pos;
	jack_nframes_t this_read = 0;
	string path;
	SoundTouch st;
	string region_name;
	string ident = N_("-TIMEFX-");

	percentage = (float) ((double) tsr.new_length - (double) tsr.region.length()) / ((double) tsr.new_length) * 100.0f;

	/* the soundtouch code wants a *tempo* change percentage, which is of opposite sign
	   to the length change.
	*/
	
	percentage = -percentage;

	st.setSampleRate (frame_rate());
	st.setChannels (1);
	st.setTempoChange (percentage);
	st.setPitchSemiTones (0);
	st.setRateChange (0);
	
	st.setSetting(SETTING_USE_QUICKSEEK, tsr.quick_seek);
	st.setSetting(SETTING_USE_AA_FILTER, tsr.antialias);

	path = path_from_region_name (tsr.region.name(), ident);
	
	if (path.length() == 0) {
		error << compose (_("tempoize: error creating name for new audio file based on %1"), tsr.region.name()) 
		      << endmsg;
		goto out;
	}

	try {
		new_file = new FileSource (path, frame_rate());
	}

	catch (failed_constructor& err) {
		error << compose (_("tempoize: error creating new audio file %1 (%2)"), path, strerror (errno)) 
		      << endmsg;
		goto out;
	}

	pos = 0;
	tsr.progress = 0.0f;

	try {

		while (tsr.running && pos < tsr.region.length()) {
			jack_nframes_t this_time;
			
			this_time = min (bufsize, tsr.region.length() - pos);

			/* read from the master (original) sources for the region, not the ones
			   currently in use, in case its already been subject to timefx.
			*/

			if ((this_read = tsr.region.master_read_at (buffer, buffer, gain_buffer, pos + tsr.region.position(), this_time)) != this_time) {
				error << compose (_("tempoize: error reading data from %1"), new_file->name()) 
				      << endmsg;
				goto out;
			}
			
			pos += this_read;
			
			tsr.progress = ((float) pos) / tsr.region.length();
			
			st.putSamples (buffer, this_read);
			
			while ((this_read = st.receiveSamples (buffer, this_read)) > 0 && tsr.running) {
				if (new_file->write (buffer, this_read) != this_read) {
					error << compose (_("error writing tempo-adjusted data to %1"), new_file->name()) 
					      << endmsg;
					goto out;
				}
			}
		}
		
		if (tsr.running) {
			st.flush ();
		}
		
		while (tsr.running && (this_read = st.receiveSamples (buffer, this_read)) > 0) {
			if (new_file->write (buffer, this_read) != this_read) {
				error << compose (_("error writing tempo-adjusted data to %1"), new_file->name()) 
				      << endmsg;
				goto out;
			}
		}
	}

	catch (soundtouch_exception& err) {
		error << _("timefx code failure. please notify ardour-developers.") << endmsg;
		goto out;
	}

	time_t now;
	struct tm* xnow;
	time (&now);
	xnow = localtime (&now);

	new_file->update_header (0, *xnow, now);

	region_name = region_name_from_path (new_file->name());
	region_name.replace (region_name.find (ident), 8, N_(".t"));

	r = new Region (*new_file, 0, new_file->length(), region_name,
			0, Region::Flag (Region::DefaultFlags | Region::WholeFile));

  out:

	if (new_file) {

		/* if we failed to complete for any reason, mark the new file 
		   for deletion.
		*/

		if ((r == 0 || !tsr.running)) {
			new_file->mark_for_remove ();
		}

		/* release our reference to the file */
		
		new_file->unref ();
	}

	/* if the process was cancelled, delete the region */

	if (!tsr.running) {
		if (r) {
			delete r;
			r = 0;
		} 
	}
	
	return r;
}
